
/* Conduct tests for the presence of gene conversion in
      an aligned DNA or protein sequence file.
   All rights reserved - Stanley Sawyer

   Uses a SCORE allowing mismatches.

   Basic command-line format:
     GENECONV <Seqfile> [Outfile] [configfile[s].cfg] [-Options]

   See documentation for options and other details.

   S. Sawyer - Department of Mathematics
               Washington University in St.Louis
               St. Louis, MO 63130
               email: sawyer@math.wustl.edu
               Web:   http://www.math.wustl.edu/~sawyer
               Office: 314-935-6703  FAX: 314-863-1302

   Parts of source file (line numbers approximate):

     (i) (***AAA) (L62) Main function, ReadSeqFile()

    (ii) (***BBB) (L217)  Display global and local fragments

   (iii) (***CCC) (L531)  Logfile package

*/

#include "vall.h"
#include "vflags.h"
#include "vread.h"
#include "vcalc.h"
#include "vdump.h"

/* Program version etc. are in version.c */

char *Main_AllRightsReserved =
   "All rights reserved - Stanley Sawyer\n";

/* Prototypes for functions in this file */

static void ReadSeqFile(void);

static void openlog_file(void);
static void closelog_file(void);
static int logprintf (const char *control, ...);
static void lognl(void);
void no_seqf(void);   /* If seqfile==NULL */

void closeout(const char *errmess);

/* Prototypes for functions in fragment display package */

static void fixup_frag_structs(void);
static void do_crop(SIMX *sc);
static void showglobfrags(void);

static void showpairs (void);
static void which_frags (int globflag, int simpval);
static int textnote(int len, char *buff, int len2);


/* (***AAA) Main functions */

int main (int argc, char **argv)
 { printf(
    "\n Apply statistical tests for apparent gene conversion events\n"
    "   for aligned DNA or protein sequence data.\n"
    "   %s Vs. %s\n"
    "   See the documentation for details.\n",
          ProgName, ProgVersion);
   printf(
    " Default output file: <Inputfile>.frags   Log file: <Inputfile>.sum\n");

   /* Read command-line file names and flags */
   /*   A <Seqfile> name is essential; this is checked */
   /*   in set_options(). */
   set_options(argc,argv);

   show_commline(NULL,TRUE);
   /* Open the log file */
   if (havelog)  openlog_file();

   /* Make sure that the outfile file names and paths are valid */
   /*   before doing any calculations */
   /* (If the output path is invalid, `testdump()'   */
   /*   generates a fatal error internally.) */
   if (anydump())  testdump();

   /* Build POLYSITE array from sequence data */
   ReadSeqFile();

   set_textprintf(logprintf,lognl);  /* To log file only */
   if (seqtype==SILENT)  show_codon_stats();
   if (showpolyflag)  show_polysites();
   if (npolys<=2)  say_toofewpolys();
   if (numsims>0 || ranflag)
    { char cbuff[44];
      printf ("\nThe starting random number seed is  %lu.\n", wstartseed);
      if (numsims>0)
        printf ("Simulated P-values are based on %s permutations.\n",
           commll(cbuff,numsims)); }

   /* Score_data(TRUE) (in vtcalc.c) constructs fragment arrays */
   /* `Textprintf' is still writing to the log file only */
   score_data(TRUE);
   if (ranflag)
     { bprintf(
       "\nREPLACING THE DATA by a random permutation of itself:\n");
       shuffdata();  rshuffled=TRUE;
       if (showpolyflag) { make_poly_save(FALSE);  show_polysites(); }
       score_data(TRUE); }

   /* Main iteration loop */

   if (numsims>0)
    { char cbuff[44];  int iter;
      printf ("\nPermuting %s times....", commll(cbuff,numsims));
      set_textprintf(bprintf,bnl);     /* For timing information */
      #ifndef UNIX
      timer(TIME_INIT,0,NULL);
      #endif
      for (iter=1; iter<=numsims; iter++)
       { shuffdata();  score_data(FALSE);
         if (irrupt>0 && (iter%irrupt) == 0)
             printf("%d ", iter); }  iter--;
      nl();
      #ifndef UNIX
      timer(TIME_ELAPSED,iter," for simulations");
      #endif
      show_restricts();
      set_textprintf(logprintf,lognl);  /* Only to the log file */
      /* Convert tabled pairwise scores to Pvalues */
      /* (Global scores are handled individually.) */
      convert_htable(); }

   /* Truncate fragment arrays dependent on simulated P-values */
   /* If asked, change sortorder of global frags */
   fixup_frag_structs();

   /* Set npairfrags, bmxlen,gmxlen,pzmxlen, pzovlap, */
   /*   and ghavcirc,pzhavcirc,tothavcirc  flags in cfinx,cfoutseqx,etc. */
   set_totpairparms();
   set_totglobparms();

   if (anydump())
    { textprintf("\n");
      dump_fragments(NULL);   /* This is in vdump.c */
      /* To be safe */
      set_textprintf(logprintf,lognl); }  /* Only to the log file */
   if (numsims==0)
      bprintf ("\nNo permutations done.\n");

   /* Show global LK scores */
   show_vmainpvals();
   if (numsims>0 && show_mmsims)
    { textnl();  show_maxmeansims(TRUE); }

   /* Show global fragments and global P-values, */
   /*   then pairwise fragment lists and pairwise P-values: */
   if (logfragments)
    { if (listglobal>0)  showglobfrags();
      else  textprintf ("\nNo (global) fragments listed.\n");
      if (listpairs && listperpair>0)  showpairs();
      else  textprintf ("\nNo pairwise fragments listed.\n"); }

dump_exit:
   closeout(NULL);
   if (hold_window)  presskey();
   return 0; }

/* End of main() function */


void no_seqf(void)
 { printf("\n %s was called with no Sequence File name.\n", ProgName);
   if (HelperName!=NULL && HelperName[0]!=0)
     printf(
     "\n If you got here by clicking on `%s' and would prefer\n"
     "   a windowed environment, click on `%s' instead.\n",
        ProgName, HelperName);
   printf(
  "\n FORMAT for command-line mode:\n"
  " %s <Seqfile>  [Outputfiles]  [LogFile]  [-optional flags]\n"
  "   for <necessary> and [optional] arguments.\n"
  " See the documentation for the details.\n\n"
  " The default entry ``%s <Seqfile>'' autosenses DNA or protein\n"
  "   sequences, carries out a permutation test with 10,000 permutations,\n"
  "   and writes lists of significant apparent gene conversion events\n"
  "   to ``<Seqfile>.frags''\n",
       ProgName, ProgName);
   presskey();
   exit(1); }


/* Read <Seqfile> into the array  polysite[] */

static void ReadSeqFile(void)
 { OpenSeqFile();    /* Uses seqfile.  Open with error handling. */
   /* Is there group structure or other restrictions? */
   setgroups();
   /* Read and store polymorphic site data in  polysites[] */
   read_polys();
   /* Display various statistics */
   run_comments();

   if (havelog)  printf ("%s log output to `%s'\n",
      (logappend) ? "Appending" : "Writing", logfile);
   else printf ("No log output.\n");

   if (matchchar)
     logprintf("Match (replacement) char in file is %c (%2XH)\n",
        matchchar,matchchar);
   else logprintf("No match (replacement) char in file\n");
   if (use_commchars())
    { if (commchars!=NULL && commchars[0])
        logprintf("Lines beginning with one "
           "of `%s' are comments\n", commchars);
      else logprintf("No initial-line comment flags\n"); }
   if (newnumbersep)
     logprintf("Number separator as in 123%c456\n", numbersep);
       }



/* (***BBB) */
/* Display global and local fragments with P-values */

/* Show global frags and global P-values */

/* Display base range for a segment of polymorphism #s */

char *fragrange(char *gbuff, int start, int finish, char *form)
 { /* typedef struct { int lf,rt,caf; } FRAGOFFS; */
   FRAGOFFS x = offsfrags(start,finish);  int len;
   x.lf += AddOffset;  x.rt += AddOffset;
   len=sprintf(gbuff,form,x.lf);
   gbuff[len++]='-';
   if (x.caf)  sprintf(gbuff+len,"CA%d", x.rt);
   else  sprintf(gbuff+len,form,x.rt);
   return  gbuff; }


/* Outer fragments are displayed for only one sequence from a group */
/*   of size 2 (the information would be identical) */

static void show_gfrag (GFRAG *gg,
     int lastflag, SIMTYPE simtype, int dosims)
 { char gbuff[66];
   /* Display sequence or sequences */
   if (simtype==SIM_INNER || simtype==SIM_OUTERPAIR)
    textprintf (" %-*s/%-*s:",
      mxtag,strtag[gg->seq1], mxtag,strtag[gg->seq2]);
   else if (simtype==SIM_OUTERGROUP) textprintf (" %s%s (%s...):",
      (grouptype==G_GROUPS) ? "GROUP " : "",
      grouplist[gg->group].gname, strtag[grouplist[gg->group].first]);
   else if (grouplist[gg->group].ns==2)  textprintf (" %-*s OR %-*s:",
      mxtag,strtag[gg->seq1], mxtag,strtag[(gg->seq1)+1]);
   else  textprintf (" %-*s:", mxtag,strtag[gg->seq1]);
   /* Show the fragment */
   textprintf ("%s (%s%4.2f, %d,%d,%d,%d, %d,%d)",
     fragrange(gbuff, gg->m.start, gg->m.finish, "%3d"),
     (lastflag && gg->kascore==(gg+1)->kascore) ? "*" : "",
     gg->kascore, gg->m.score,
     gg->m.finish - gg->m.start, gg->m.nmismat,
     (gscale) ? gg->mismatpen: 0, gg->totdifs,npolys);
   if (dosims)
     textprintf (" %6.4fS", (double)*gg->m.grc/numsims);
   textprintf("\n"); }



static void showgfrags (SIMX *sc, int *hflag)
 { if (sc->inmodel)
    { char nbuff[44];  int k, gnum=sc->ngfrags, dosims=sc->dosims;
      GFRAG *gg, *ggend;
      if (*hflag)  *hflag=FALSE;  else textprintf("\n");
      textprintf ("High-scoring %s fragments in `%s' "
         "(*=tie with next):\n", fragname(sc->simtype), seqfile);
      show_restricts();
      gg=sc->gflist;  ggend=gg+gnum;
      for (k=0; gg<ggend; gg++, k++)
        show_gfrag (gg, (k==gnum-1), sc->simtype, dosims);
      textprintf(" A total of %s listed.\n",
         numsay(nbuff, gnum, "fragment", FALSE));
      /* Report on discarded fragments unless none listed */
      if (gnum>0 && maxoverlap>0)
       { char nbuff[88];
         textprintf (" %s discarded.\n",
           numsayc(nbuff,sc->pzovlap,
               "overlapping fragment",TRUE)); }}}



static void showglobfrags(void)
 { int hflag;
   textprintf(
  "\nGlobal list of high-scoring fragments (max: %d, min score: %d),\n"
  "   format: first-last (LKscore, score,numpolys,nummismatches,\n"
  "     mismatch penalty, total sequence differences,numpolys)\n",
      listglobal, minscore);
   if (numsims>0)
    { char nbuff[66];
      textprintf ("   with P-values based on %s:\n",
        numsayc(nbuff, numsims, "simulation", FALSE)); }
   which_frags(TRUE,TRUE);  hflag=TRUE;
   if (innerpairfrags) showgfrags (&cfinx, &hflag);
   if (outerpairfrags) showgfrags (&cfoutpairx, &hflag);
   if (outerseqfrags) showgfrags (&cfoutseqx, &hflag);
   if (outergrpfrags) showgfrags (&cfoutgrpx, &hflag); }




static int *ngr_list;
static int lngr,lxngr;

static void trunc_glob(SIMX *sc)
 { if (sc->inmodel && sc->dosims && sc->ngfrags>0)
    { int nf, nfrags=sc->ngfrags, npairs=sc->npairs;
      GFRAG *glist=sc->gflist;
      if (bcsims) qsort(glist,nfrags, sizeof(GFRAG),sortgsfrag);
      if (nfrags > listglobal+2)  sc->ngfrags=nfrags=listglobal+2;
      for (nf=0;  nf<nfrags;  nf++, glist++)
       { int ngr=*(glist->m.grc);
         if (bcsims)
          { if (ngr==0) ngr=1;  ngr*=npairs;
            if (lngr>=lxngr)
              ngr_list=rexxalloc(ngr_list,sizeof(int), lngr,lxngr+=3000);
            ngr_list[lngr]=ngr;
            glist->m.grc=&ngr_list[lngr];  lngr++; }
         if (havsimglobpval && (double)ngr/numsims > maxsimglobpval)
          { sc->ngfrags=nf;  break; }}}}


static void trunc_pair(SIMX *sc)
 { if (sc->inmodel && sc->dosims)
    { PAIRCOMP *pc,*pcend;
      pc=sc->pairpc;  pcend=pc+(sc->npairs);
      for ( ; pc<pcend; pc++)
       { int nf, pnum=pc->nfrags;
         MFRAG *mfrag=pc->mf;
         for (nf=0;  nf<pnum;  nf++, mfrag++)
           if ((double)*mfrag->grc/numsims > maxsimpairpval)
            { pc->nfrags=nf;  break; }}}}



static void do_fix(SIMX *sc)
 { trunc_glob(sc);
   if (havsimpairpval && numsims>0)  trunc_pair(sc);
   do_crop(sc);
   if (sortgfragsbyseq && sc->ngfrags>0)
     qsort(sc->gflist,sc->ngfrags, sizeof(GFRAG),sortgfbyseqs); }



static void fixup_frag_structs(void)
 { /* Used only if allbc==TRUE */
   if (ngr_list!=NULL) { free(ngr_list);  ngr_list=NULL; }
   lngr=lxngr=0;
   do_fix(&cfinx);  do_fix(&cfoutpairx);
   do_fix(&cfoutseqx);  do_fix(&cfoutgrpx); }



/* Set sc->pcropped and sc->gcropped */
/* Note that trunc_glob() and trunc_pair() have no effect */
/*   unless dosims==TRUE */

static void do_crop(SIMX *sc)
 { if (sc->inmodel)
    { sc->gcropped=sc->pcropped=FALSE;
      if (sc->ngfrags > listglobal)
       { sc->ngfrags=listglobal;  sc->gcropped=TRUE; }
      if (listpairs && listperpair>0)
       { PAIRCOMP *pc, *pcend;
         pc=sc->pairpc;  pcend=pc+(sc->npairs);
         for ( ; pc<pcend; pc++)
          { if (pc->nfrags > listperpair)
             { pc->nfrags=listperpair;  sc->pcropped=TRUE; }}}}}



/* Logfragment package --- for writing fragment information */
/*   to the log file in a condensed form. */

static int show_mfrag(char *buff, MFRAG *mf,
       PAIRCOMP *pc, int lastflag, int dosims)
 { char pbuff[66];
   int len2=sprintf(buff, "  %s(%s%d,%d,%d,%d)",
     fragrange(pbuff, mf->start, mf->finish, "%d"),
     (lastflag && mf->score==(mf+1)->score) ? "*" : "",
     mf->score, mf->finish - mf->start,
     mf->nmismat, (gscale) ? pc->mismatpen : 0);
   if (pc->kavalid)  len2 += sprintf(buff+len2, "%6.4fA",
                       getkapval(mf->score, pc->ll,pc->kmax));
   else  len2 += sprintf(buff+len2, "XA");
   if (dosims)
     len2 += sprintf(buff+len2, "%6.4fS", (double)*mf->grc/numsims);
   return len2; }



/* Only one set of outer fragments is displayed for a group */
/*   of size 2 (the information would be identical otherwise) */

static int showpgroup (PAIRCOMP *pc, SIMTYPE simtype, int dosims)
 { char buff[199], nbuff[33];
   int k,nfrags,nd, len,len2;  MFRAG *mf=pc->mf;
   int i=pc->seq1, j=pc->seq2;
   /* Return if there is nothing to say */
   /* (The heading explains why.) */
   if (pc->nfrags==0)  return 0;

   /* Display sequence or sequences */
   if (simtype==SIM_INNER || simtype==SIM_OUTERPAIR)
     len=textprintf (" %s v %s", strtag[i], strtag[j]);
   else if (simtype==SIM_OUTERGROUP)
     len=textprintf (" %s%s (%s...)",
       (grouptype==G_GROUPS) ? "GROUP " : "",
       grouplist[pc->group].gname,
       strtag[grouplist[pc->group].first]);
   else if (grouplist[pc->group].ns==2)
     len=textprintf (" %s OR %s", strtag[i], strtag[i+1]);
   else len=textprintf (" %s", strtag[i]);

   /* Heading */
   nd=pc->totdifs;  if (simtype!=SIM_INNER)  nd=npolys-nd;
   if (simtype==SIM_INNER || simtype==SIM_OUTERPAIR)
     len+=textprintf ("  (%s diff)", numsay(nbuff, nd, "site", TRUE));
   else if (simtype==SIM_OUTERSEQ)
     len+=textprintf ("  (%s)", numsay(nbuff, nd, "unique site", TRUE));
   else if (simtype==SIM_OUTERGROUP)
     len+=textprintf ("  (%s)",
        numsay(nbuff, nd, "monomorphic site", TRUE));

   /* Now the fragments */
   /* This function called only if listpairs==TRUE */
   if ( (nfrags=pc->nfrags) > listperpair)  nfrags=listperpair;
   for (k=0; k<nfrags; k++, mf++)
    { len2=show_mfrag(buff, mf, pc, (k==nfrags-1), dosims);
      len=textnote(len,buff,len2); }
   if (nfrags>=4)
    { len2=sprintf(buff, "  (%d listed)", nfrags);
      len=textnote(len,buff,len2); }
   len2=sprintf(buff, "  (%s dropped)",
          numsay(nbuff,pc->novlap,"overlap",FALSE));
   textnote(len,buff,len2);
   textprintf("\n");
   return nfrags; }


static int textnote(int len, char *buff, int len2)
 { if ( (len+=len2) > 74)
     { textprintf ("\n %s", buff);  len=len2+1; }
   else  textprintf ("%s",buff);
   return len; }




/* Show pairwise fragments and pairwise P-values */

static void show_ptype (SIMX *sc)
 { if (sc->inmodel)
    { char *what, nbuff[55], bbuff[88];
      char *btag = (blocsize) ? " in that block" : "";
      int fnum, pgroup, dosims=sc->dosims;
      SIMTYPE simtype=sc->simtype;  PAIRCOMP *pc,*pcend;
      switch(simtype) {
        case SIM_INNER:
          what="sites at which the two sequences differ";  break;
        case SIM_OUTERSEQ:
          what="nonunique polymorphic sites";  break;
        case SIM_OUTERPAIR:
          what="sites at which the two sequences match";  break;
        case SIM_OUTERGROUP:
          what="monomorphic sites";
          btag = (blocsize) ? "in that block" : "in that group";  break; }
      /* This function called only if listpairs==TRUE */
      textprintf(
       "\nPairwise %s fragments with scores >= %d (max: %d), format:\n"
       "   (number of %s%s) and\n"
       "   first-last(score,npolys,nummismatches,penalty)%s:\n",
         fragname(simtype), minscore, listperpair, what, btag,
         (sc->dosims) ? " and P-values" : "");
      which_frags(FALSE, sc->dosims);
      pc=sc->pairpc;  pcend=pc+(sc->npairs);
      pgroup=-1;  fnum=0;
      while (pc<pcend)
       { if (grouptype==G_GROUPS && simtype!=SIM_OUTERGROUP
               && pc->group!=pgroup)
          { textprintf ("GROUP %s:\n", grouplist[pc->group].gname);
            pgroup=pc->group; }
         fnum+=showpgroup(pc++,simtype,dosims); }
      if (fnum>0)
       { sprintf(bbuff, "%s fragment", fragname(simtype));
         textprintf("A total of %s listed.\n",
            numsay(nbuff, fnum, bbuff, FALSE)); }
      else
       { textprintf ("No %s fragments have", fragname(simtype));
         if (havsimpairpval)
           textprintf (" pairwise P-value <= %g", maxsimpairpval);
         if (minlength>1)  textprintf ("%s length >= %d",
             (havsimpairpval) ? " and" : "", minlength);
         textprintf("\n"); }}}



static void showpairs (void)
 { /* Inner fragments: */
   show_ptype(&cfinx);
   /* Outer fragments: */
   show_ptype(&cfoutpairx);
   show_ptype(&cfoutseqx);
   show_ptype(&cfoutgrpx); }



static void which_frags (int globflag, int simpval)
 { int xhav;  double pval;
   if (minlength>1)
     textprintf ("   (Listed only if the length >= %d)\n", minlength);
   if (maxoverlap>0)
     textprintf (
      "   (Maximum number of fragments overlapping a first fragment: %d)\n",
            maxoverlap);
   if (havkapairpval)
     textprintf(
    "   (Fragments displayed only if pairwise LK P-value <= %g) )\n",
      maxkapairpval);
   if (havkaglobpval)
     textprintf(
    "   (Fragments displayed only if "
          "Bonferroni-corr. LK P-value <= %g\n", maxkaglobpval);
   if (simpval)
    { if (globflag) { xhav=havsimglobpval;  pval=maxsimglobpval; }
      else { xhav=havsimpairpval;  pval=maxsimpairpval; }
      if (xhav && numsims>0)  textprintf ("   (Listed only if "
            "the simulated P-value <= %g)\n", pval); }}




/* (***CCC) */
/* Open and write to log files */

static FILE *fflog;
static int triedlog_file;

/* Set the log file name and open it */

static void openlog_file(void)
 { triedlog_file=TRUE;
   set_filenames();  /* Try to define `logfile' */
   if (logfile==NULL)  logfile=strxdup("LOGFILE.sum");
   else { char *ofs=logfile;
          logfile=makefnam(".sum",ofs);  if (ofs) free(ofs); }
   fflog = fxopenwrite(logfile, (logappend) ? APPENDMODE : WRITEMODE);
   if (logappend)  fprintf(fflog,"\n\n\n\n");
   logprintf ("LOG FILE for sequence file=%s\n"
       "%s version %s\n",
       (seqfile)?seqfile:"NULL", ProgName, ProgVersion);
   lognl();
   set_textprintf(logprintf,lognl);  /* Write only to the sum file */
   show_commline(NULL,FALSE);
   /* Restore `textprintf' settings */
   reset_textprintf(); }


static void closelog_file(void)
 { if (fflog)
    { fxclose(fflog,logfile);  fflog=NULL; }}


void closeout(const char *errmess)
 { if (havelog && fflog==NULL && triedlog_file==FALSE)
     openlog_file();
   if (errmess!=NULL && errmess[0]!=0)  bprintf("\n%s\n", errmess);
   else { lognl();
          show_alloc((VOID_PRINTFUNP)logprintf); }
   closelog_file(); }



/* Write to the sum file and/or console */

int bprintf (const char *control, ...)
 { int ret;  va_list list;  va_start(list, control);
   ret=vprintf (control, list);
   if (fflog) ret=vfprintf (fflog, control, list);
   va_end(list);
   return ret; }

static int logprintf (const char *control, ...)
 { int ret=0;
   va_list list;  va_start(list, control);
   if (fflog)  ret=vfprintf (fflog, control, list);
   va_end(list);
   return ret; }

void bnl(void)
 { putchar ('\n');
   if (fflog)  fputc('\n', fflog); }

static void lognl(void)
 { if (fflog)  fputc('\n', fflog); }

