
/* Write a random <Seqfile> in FASTA format
     with <nseqs> sequences and <nbases> sites.
   Of these, <npolys> sites are polymorphic.
   No sites have more than 2 nucleotides or proteins.
   All polymorphic sites are ``phylogenetically informative''
   Output file name: ``rand<nseqs><npolys>.fasta"

   Run MAKE2RAND with no options for format

   S. Sawyer */

#define PROG_DATE "November 17, 1999"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>   /* For initializing random number seed */

#define DEFNSEQS 9
#define DEFNPOLY 50
#define DEFNBASES 500

#define TRUE 1
#define FALSE 0

int nseqs=DEFNSEQS, npolys=DEFNPOLY, nbases=DEFNBASES;
int protflag, runflag, nfixed, outgroup;
char *outfile;
unsigned long wseed;

char nucleotides[] = "TCAG";
char proteins[] = "ACDEFGHIKLMNPQRSTVWY";

typedef enum { VAR_FLAG, VAR_INT, VAR_DOUBLE, VAR_ULONG }  VARTYPE;

typedef struct { char tag;  void *vptr;
                 VARTYPE vartype;  char *descr;  int showdef; }  SETFLAGS;

SETFLAGS ss[] = {
 { 'N', &nseqs, VAR_INT, "Number of sequences", TRUE },
 { 'M', &npolys, VAR_INT, "Number of polymorphic sites", TRUE },
 { 'L', &nbases, VAR_INT, "Number of bases (lengths of sequences)", TRUE },
 { 'O', &outgroup, VAR_FLAG,
   "Randomize 20% of fixed sites of LAST sequence (for use as outlier)" },
 { 'P', &protflag, VAR_FLAG, "Proteins, not DNA" },
 { 'R', &runflag, VAR_FLAG,
     "Run the program with default output name" },
 { 'W', &wseed, VAR_ULONG, "Initial random number seed" },
 { 0 }, { 0 } };


void setflags_options(SETFLAGS *ss);
char *setflags(int argc, char *argv[], SETFLAGS *sflag);

void *cxalloc (size_t nitems, size_t size);
void *mxalloc (size_t size);
char *strxdup(const char *p);
void press_exit(void);
FILE *fxopen(const char *path, const char *mode);
int  fxclose(FILE *fp, const char *fname);
char *makefnam (const char *rawfn, const char *tplate);
char *progname (const char *rawfn);

int nrand(int m);
unsigned long setrand(unsigned long  nn);
void shuffchar(char *deck, int size);


int main (int argc, char *argv[])
 { char *baselist, *baselist2;
   char **ranpolys, *ranfsites, *polyorder;
   int i,j,k,m,totb, lenalpha;
   FILE *wf;  unsigned long init_seed;
   printf (
    "Write a random <Seqfile> with <nseqs> sequences\n"
    "   and <npolys> sites.  All polymorphic sites are `informative'.\n"
    "   No sites have more than two nucleotides or amino acids.\n"
    "Format:  %s  [Seqfilename] [-options]   with options\n",
       progname(argv[0]));
   setflags_options(ss);
   printf ("Version %s\n", PROG_DATE);
   printf ("Default output file name:  rand<nseqs><npolys>.fasta\n");

   outfile=setflags(argc,argv,ss);
   if (outfile==NULL && runflag==FALSE)
    { printf ("\nNEED TO ENTER either Seqfilename or -r\n");
      press_exit(); }
   if ( (nfixed=nbases-npolys)<0)
    { printf ("\nCAN'T HAVE FEWER BASES THAN POLYMORPHISMS!\n"
        "  (npolys=%d  nbases=%d)\n", npolys, nbases);  press_exit(); }

   if (nseqs<4 || npolys<4)
    { printf ("Must have at least 4 sequences and 4 polymorphisms !\n"
       "  (nseqs=%d  npolys=%d)\n", nseqs, npolys);  press_exit(); }
   if (outfile==NULL)
    { char nbuff[288];
      sprintf(nbuff, "rand%d%d.fasta", nseqs,npolys);
      outfile=strxdup(nbuff); }
   else  outfile=makefnam(outfile,".fasta");
   printf ("\nOutput file name: %s  (%s)\n", outfile,
     (protflag) ? "proteins" : "nucleotides");

   /* Fill random data */
   init_seed=setrand(wseed);
   printf ("Initial random-number seed: %lu\n", init_seed);

   /* Determine a random data set */
   /*   One random sequence of length `nfixed' of fixed sites */
   /*   `Nseqs' random sequences of length `npolys' */
   /*     of polymorphic sites. */

   baselist = (protflag) ? proteins : nucleotides;
   /* Work with `baselist2', not `proteins' or `nucleotides' */
   baselist2 = strxdup(baselist);
   lenalpha = strlen(baselist2);

   /* Put random bases for fixed sites in `ranfsites' */
   ranfsites = mxalloc(nfixed+3);
   for (k=0; k<nfixed; k++)  ranfsites[k]=baselist[nrand(lenalpha)];

   /* Store random polymorphic sites in `ranpolys' */
   ranpolys = cxalloc(npolys+3, sizeof(char *));
   for (k=0; k<npolys; k++)
    { char b1, b2, *sites;  int n1;
      ranpolys[k]=sites=mxalloc(nseqs+2);
      /* Randomly shuffle the possible bases, so that the first */
      /*   two positions will contain a distinct random pair. */
      shuffchar(baselist2,lenalpha);
      b1=baselist2[0];  b2=baselist2[1];
      /* The number of bases of type b1: Need n1>=2 and n2>=2 */
      n1=2+nrand(nseqs-3);
      /* Assign n1 b1s and n2=nseqs-n1 b2s, independent of order */
      /* `Informative' polymorphisms with exactly two bases */
      /*   are guaranteed. */
      for (i=0; i<n1; i++)  sites[i]=b1;
      for ( ; i<nseqs; i++)  sites[i]=b2;
      /* Shuffle the order of the bases */
      shuffchar(sites,nseqs); }

   /* Determine the positions of the polymorphic sites in the alignment */
   polyorder = mxalloc(nbases+3);
   for (k=0; k<npolys; k++)  polyorder[k]=1;
   shuffchar(polyorder,nbases);

   /* Write a FASTA file */
   wf=fxopen(outfile,"wt");
   for (totb=i=0; i<nseqs; i++)
    { int kp,kf;
      if (i==0) fprintf(wf,
      ">R1  Random (%d seqs  %d polys  %d bases  rseed: %lu)\n",
           nseqs, npolys, nbases, init_seed);
      else  fprintf(wf, ">R%d \n", i+1);
      /* The kth site, kp-th polymorphic site, kf-th fixed site */
      /* Also loop over output file lines */
      k=kp=kf=0;
      do { char lbuff[88];
           for (j=m=0; m<60 && k<nbases; m++)
            { if ( (m%10)==0)  lbuff[j++]=' ';
              if (polyorder[k++])
                lbuff[j++]=ranpolys[kp++][i];
              else
               { char cc;
                 if (outgroup && i==nseqs-1 && nrand(5)==0)
                   cc = baselist[nrand(lenalpha)];
                 else cc = ranfsites[kf];
                 lbuff[j++]=cc;  kf++; }}
           lbuff[j]=0;
           fprintf(wf, "  %s\n", lbuff); }
      while (k<nbases);  totb+=nbases; }

   printf ("%d bases written to %s\n", totb, outfile);
   fxclose(wf,outfile);
   return 0; }




/* Function to display command-line arguments */

static char *boolon(int flag)
 { return (flag) ? "ON" : "OFF"; }


void setflags_options(SETFLAGS *ss)
 { int i;
   for (i=0; ss[i].tag!=0; i++)  if (ss[i].descr!=NULL)
    { char nbuff[100], *app="";
      SETFLAGS s=ss[i];  nbuff[0]=0;
      switch(s.vartype) {
        default:  sprintf(nbuff, "???");  break;
        case VAR_FLAG:
          sprintf(nbuff, "%s", boolon(*(int *)s.vptr));  break;
        case VAR_INT:  app="<num>";
          sprintf(nbuff, "%d", *(int *)s.vptr);  break;
        case VAR_ULONG:  app="<ulnum>";
          sprintf(nbuff, "%lu", *(unsigned long *)s.vptr);  break; }
      if (s.showdef)
        printf ("  -%c%s - %s (default: %s)\n",
            s.tag, app, s.descr, nbuff);
      else
         printf ("  -%c%s - %s\n", s.tag, app, s.descr); }}



static char *dovar(char op, char *p, VARTYPE vartype, void *vptr);

static void badval(char op, char *p, char *bleep)
 { printf("\nIMPROPER -%c<num> FORMAT (p=%s)!!  (%s.)\n",
      op, p, bleep);  press_exit(); }


char *setflags(int argc, char *argv[], SETFLAGS *sflag)
 { char *str=NULL;  int i;
   if (sflag==NULL || sflag[0].tag==0)  return NULL;
   for (i=1; i<argc; i++)
    { char op, *orgp, *p1, *p=argv[i];
      if (*p=='-')  while ( (op=toupper(*++p))!=0)
        { int notfound=TRUE;  SETFLAGS *ss;
          if (op=='-') continue;     /*  --a is OK */
          for (ss=sflag; ss->tag; ss++)  if (toupper(ss->tag)==op)
           { notfound=FALSE;  orgp=p;
             p1=dovar(op,orgp,ss->vartype,ss->vptr);
             if (p1>p) p=p1; }
          if (notfound)
           { printf ("\nUNRECOGNIZED FLAG `%c'!!!\n", op);
             press_exit(); }}
      else if (str==NULL)  str=p;
      else { printf ("More than one string argument!\n"
               "  `%s'  `%s'\n", str, p);  press_exit(); }}
   return str; }


static char *dovar(char op, char *p, VARTYPE vartype, void *vptr)
 { char *orgp=p;  unsigned long xx;
   switch(vartype)
    { default: printf (
         "*** INTERNAL ERROR (SETFLAGS)\n"
         "   Illegal var type `%d' for flag `%c'!\n",
               vartype, op);  press_exit();
      case VAR_FLAG:
         *(int *)vptr = TRUE;  return p;
      case VAR_INT:  case VAR_ULONG:
         if (!isdigit(*++p))  badval(op,orgp,"Expecting a digit");
         xx=*p-'0';
         while(isdigit(*++p))
          { unsigned long xx1=xx;  xx = 10*xx + *p-'0';
            if (xx<xx1)  badval(op,orgp,"Unsigned long overflow"); }
         if (vartype==VAR_ULONG)
           *(unsigned long *)vptr=xx;
         else    /* VAR_INT */
          { unsigned long xx1;  int n=(int)xx;  xx1=n;
            if (xx1!=xx || n<0)  badval(op,orgp,"Int too large");
            *(int *)vptr=n; }
         return --p; }}



/* Allocation package */

static void no_alloc (unsigned long uu, char *func)
 { printf(
    "\nCANNOT ALLOCATE %lu bytes (`%s') from the system.\n",
       uu, func);  press_exit(); }

void *cxalloc (size_t nitems, size_t size)
 { void *vptr;
   unsigned long tsize = (unsigned long)nitems*size;
   if (tsize==0L)  tsize=nitems=size=1;
   if ( (vptr = calloc(nitems,size)) == NULL)
        no_alloc (tsize, "cxalloc");
   return  vptr; }

void *mxalloc (size_t size)
 { void *vptr;
   if (size==0)  size=1;
   if ( (vptr = calloc(1,size)) == NULL)
        no_alloc (size, "mxalloc");
   return  vptr; }

char *strxdup(const char *p)
 { return (p!=NULL) ? strcpy (mxalloc(strlen(p)+1), p)
            : mxalloc(1); }


void press_exit(void)
 { printf ("\nPress `Enter' to continue...");
   getchar();
   printf ("\n");  exit(1); }



/* Open, close, delete files with error handling */

int errmsg (const char *bleep, const char *file)
 { fprintf (stderr, "\n%s %s the file  `%s'\n",
        (isupper(bleep[0])) ? "CANNOT" : "Cannot", bleep, file);
   perror("");  return -1; }

FILE *fxopen(const char *path, const char *mode)
 { FILE *fptr;
   if ( (fptr=fopen(path,mode))==NULL)
    { errmsg("OPEN", path);  press_exit(); }
   return fptr; }

int  fxclose(FILE *fp, const char *fname)
 { int ret=-1;
   if (fp && (ret=fclose(fp)))  errmsg("CLOSE", fname);
   return ret; }


char *makeuc(char *str)
 { char *p;
   if (p=str) while (*p) { *p = toupper(*p);  p++; }
   return str; }



typedef struct { char *drive, *path, *name, *ext; }  PARTFN;

PARTFN partfn (const char *source)
 { int n;  char *p;  PARTFN pf;
   memset(&pf, 0, sizeof(pf));
   if (source==NULL)  source="";
   /* Is there a drive? */
   if (source[0] && source[1]==':')
    { sprintf(pf.drive=mxalloc(3), "%c:", source[0]);
      source += 2; }  else  pf.drive="";
   /* Is there a path? */
   for (n=strlen(source); --n>=0; )
     if (source[n]=='\\' || source[n]=='/')  break;
   if (++n>0)
    { memmove(pf.path = mxalloc(n+1), source, n);
      source += n; }  else  pf.path="";
   /* Is there a file name? */
   if (source[0] && source[0]!='.')
    { if (p=strrchr(source,'.'))
       { memmove(pf.name = mxalloc(p-source+1), source, p-source);
         source=p; }
      else { pf.name = strxdup(source);  source=""; }}
   else  pf.name="";
   /* Is there an extension? */
   pf.ext = (source[0]) ? strxdup(source) : "";
   return  pf; }


char *makefmpf (PARTFN pf)
 { char *name;
   sprintf (name=mxalloc(strlen(pf.drive)+strlen(pf.path)
                +strlen(pf.name)+strlen(pf.ext)+2),
      "%s%s%s%s", pf.drive, pf.path, pf.name, pf.ext);
   return name; }


void freepf (PARTFN pf)
 { if (pf.drive[0])  free(pf.drive);
   if (pf.path[0])  free(pf.path);
   if (pf.name[0])  free(pf.name);
   if (pf.ext[0])  free(pf.ext); }


char *makefnam (const char *rawfn, const char *tplate)
 { char *mname;  PARTFN cpf,tpf, mpf;
   mpf=cpf=partfn(rawfn);  tpf=partfn(tplate);
   if (mpf.drive[0]==0)  mpf.drive=tpf.drive;
   if (mpf.path[0]==0)  mpf.path=tpf.path;
   if (mpf.name[0]==0)  mpf.name=tpf.name;
   if (mpf.ext[0]==0)  mpf.ext=tpf.ext;
   mname=makefmpf(mpf);  freepf(tpf);  freepf(cpf);
   return mname; }


char *progname (const char *rawfn)
 { PARTFN pf=partfn(rawfn);
   makeuc(pf.name);
   return pf.name; }


/* Random number routines

   int nrand(int m)    If m>=0, returns random int in  0,1,2,...,m-1
                       If m<0,  returns random int in -m+1,...,-1,0
   unsigned long setrand(unsigned long  nn)
                       Set the seed at `nn', unless nn==0L.
                       If nn==0L, set from the system timer.

   void shuffchar(char *deck, int size)
                       Do random ``perfect shuffle'' for char array of size
                       of size ``size'' 

*/


#include <stdio.h>
#include <time.h>

#define MULT 0x19660DL

/* Fix added 5-15-97 to avoid problems with 64-bit integers */

#define LRAND (rseed = ((MULT*rseed + 1L) & 0xFFFFFFFFL) )

static unsigned long  rseed = 0x75BE6ACCL;


int nrand(int m)
 { return  (m >= 0)  ?  (int) ((m * (LRAND >> 16)) >> 16)
                  : -(int) ( (-m * (LRAND >> 16)) >> 16); }


unsigned long setrand(unsigned long  nn)
 { int i;
   if (nn)  return  rseed = nn;
   rseed = time(NULL);
   for (i=0; i<20; i++)  LRAND;
   return  LRAND; }


/* Do a perfect shuffle of all elements */

void shuffchar(char *deck, int size)
 { while (--size > 0)
    { int j = nrand(size+1);
      char temp = deck[size];
      deck[size] = deck[j];  deck[j] = temp; }}

