
/* Functions and data for reading input data files
      in either ASF or NEXUS format
   All rights reserved - Stanley Sawyer

   Principal entry points:
       OpenSeqFile()
       read_polys()    (Build polysites[] array)
       CloseSeqFile()

   S. Sawyer - Department of Mathematics
               Washington University in St.Louis
               St. Louis, MO 63130
               email: sawyer@math.wustl.edu
               Office: 314-935-6703  FAX: 314-863-1302

   Parts of source file (line numbers approximate):

    (i) (***AAA) (L324) Open <Seqfile> and set params

   (ii) (***BBB) (L596) String-reading utilities

  (iii) (***CCC) (L643) FASTA and NBRF/PIR format routines

   (iv) (***DDD) (L724) NEXUS routines

    (v) (***EEE) (L1172) Routines for setting groups
                   (grouplist, setgroups() etc.)

   (vi) (***FFF) (L1543) Routines for filling polysites[]
                    (read_polys() and friends)

  (vii) (***GGG) (L1967) Low-level routines for polysites[]
                     (readsite(), readgroup(), init_readvars(), and friends)

 (viii) (***HHH) (L2793) Byte- and line-oriented reading routines

   (ix) (***III) (L2973) Routines for converting polymorphism
          numbers (start,finish) to alignment offsets
                           */


#include "vall.h"
#include "vflags.h"
#include "vread.h"
#include "vcalc.h"
#include "vdump.h"

char *AllRightsReserved_Vread =
   "All rights reserved - Stanley Sawyer";

#define BLOCLEN 2000      /* Allocation block size for `polysites[]' */
#define LINEBAT 2000      /* Batch size of dynamic lineread() buffers */

/* Exportable variables */

char **strtag, **readtag, **strshtag, **seqbase;
STRLIST seqs_notused, seqfcomments, configcomments;
int nseqs, file_nseqs, nbases, blocnum, seqf_open;
int r_polyflag, reread_polymorphisms, numutconv, allctrlnum;
int npairseqs, noutseqs, noutgrps;
int numdelblocs, maxdelblocs, numdelsing, mxtag, gxtag;
int have_silent_multsites, maxsqlinelen;
/* Seqrpoff is used only for Listed Polymorphism format */
int *seqrpoff;

/* X is added in proteins[] to distinguish the 2-fold degen.serines */

static char nucleotides[] = "TCAG",  nucwhat[]="nucleotide";
static char total_nucls[] = "TCAGURYMWSKBDHVNX";
static char tnucwhat[] = "nucleotide(incl.IUB)";
static char proteins[] = "ACDEFGHIKLMNPQRSXTVWY";
static char protwhat[] = "amino-acid";

static char *read_knownchars;
static char *knwhat;

static int numpoly_sites, nummono_sites, nummiss_sites, ntotmiss_sites;
static int numaapoly, nummiss_cods;
static int numsil, num0sil, num2sil, num3sil, num4sil, numisil;
static int num2poly, num3poly, num4poly, numipoly;

/* In sequence data, ignore digits, space characters, parentheses */
/*   of all sorts, and quote characters */

static int isbasechar(int c)
 { return (!isspace(c) && !isdigit(c)
     && strchr("()[]{}<>\"`'",c)==NULL); }

int org_bsize, org_bnum;

POLYSITE  *polysites, *sitend, *polysave;
int  npolys;

/* Local declarations */

static int linenum;

/* Readtag[] and readshowtag[] are used ONLY in readgroup() */
/*   and rdnx_unalmat() for sequence-name checking */
/* If havlabels==FALSE, these are never used. */

static char **readshowtag, *tri_cache[4];
static int tags_set;

static char *next_rpoly;
static int tri_phase, numpxend;
static int nexus_out, unaligned_data;
static char *seqline, *matchbuff, *sitebuff, **rdrow;
static int nread,nxread, matchxlen, linxlen;

static char *ExpPolyForm = "(explicit poly format)";
static char *NexFormErr = "NEXUS FORMAT ERROR: ";

static int is_stitched, stitch_saved;
static POLYSITE first_psite, last_psite;

/* Genetic code */

/* The genetic code for silent-site permutations: */
/* Eukaryote and mammalian-mitochondrial codes */

/* The values of  gflag  in the array  codn[]  below mean

   0  -  Irregular  (i.e,, not regular and not missing data)
   1  -  Regular: All but stop codons and Leucine and Arginine:
                     specifically, all amino acids such that
                  (a) codon positions 1 and 2 are nondegenerate for all
                    codons in that group --- i.e., any change of base
                    changes the amino acid, and
                  (b) codon position 3 is k-fold degenerate (k=0,2,3,4)
   2  -  Error or missing or unrecognized data

   The fields in  signature[4]  give degeneracies at those offsets.

   This array is used only for silent permutations (in ``trisite()'')
      and in some displays.
               */


AACODE codn[66],
   rcodn[] = {

/*TTX*/ {"Phe",'F',1,"002"}, {"Phe",'F',1,"002"},
              {"Leu",'L',0,"202"}, {"Leu",'L',0,"202"},
/*TCX*/ {"Ser1",'S',1,"004"}, {"Ser1",'S',1,"004"},
              {"Ser1",'S',1,"004"}, {"Ser1",'S',1,"004"},
/*TAX*/ {"Tyr",'Y',1,"002"}, {"Tyr",'Y',1,"002"},
              {"???",'?',0,"022"}, {"???",'?',0,"002"},
/*TGX*/ {"Cys",'C',1,"002"}, {"Cys",'C',1,"002"},
              {"???",'?',0,"020"}, {"Trp",'W',1,"000"},

/*CTX*/ {"Leu",'L',0,"004"}, {"Leu",'L',0,"004"},
              {"Leu",'L',0,"204"}, {"Leu",'L',0,"204"},
/*CCX*/ {"Pro",'P',1,"004"}, {"Pro",'P',1,"004"},
              {"Pro",'P',1,"004"}, {"Pro",'P',1,"004"},
/*CAX*/ {"His",'H',1,"002"}, {"His",'H',1,"002"},
              {"Gln",'Q',1,"002"}, {"Gln",'Q',1,"002"},
/*CGX*/ {"Arg",'R',0,"004"}, {"Arg",'R',0,"004"},
              {"Arg",'R',0,"204"}, {"Arg",'R',0,"204"},

/*ATX*/ {"Ile",'I',1,"003"}, {"Ile",'I',1,"003"},
              {"Ile",'I',1,"003"}, {"Met",'M',1,"000"},
/*ACX*/ {"Thr",'T',1,"004"}, {"Thr",'T',1,"004"},
              {"Thr",'T',1,"004"}, {"Thr",'T',1,"004"},
/*AAX*/ {"Asn",'N',1,"002"}, {"Asn",'N',1,"002"},
              {"Lys",'K',1,"002"}, {"Lys",'K',1,"002"},

/* Since silent polymorphisms are codon polymorphisms,
     a codon position with codons of both classes of serines
     (AGT,AGC and TCT,TCA,TCC,TCG) will not generate
     three different silent-site permutations. */

/*AGX*/ {"Ser2",'S',1,"002"}, {"Ser2",'S',1,"002"},
              {"Arg",'R',0,"202"}, {"Arg",'R',0,"202"},

/*GTX*/ {"Val",'V',1,"004"}, {"Val",'V',1,"004"},
              {"Val",'V',1,"004"}, {"Val",'V',1,"004"},
/*GCX*/ {"Ala",'A',1,"004"}, {"Ala",'A',1,"004"},
              {"Ala",'A',1,"004"}, {"Ala",'A',1,"004"},
/*GAX*/ {"Asp",'D',1,"002"}, {"Asp",'D',1,"002"},
              {"Glu",'E',1,"002"}, {"Glu",'E',1,"002"},
/*GGX*/ {"Gly",'G',1,"004"}, {"Gly",'G',1,"004"},
              {"Gly",'G',1,"004"}, {"Gly",'G',1,"004"},

/*ERROR*/ {"!!!",'!',2,"000"}, { 0 } },


/* Now the mammalian mitochondrial genetic code: */

 mtcodn[] = {

/*TTX*/ {"Phe",'F',1,"002"}, {"Phe",'F',1,"002"},
              {"Leu",'L',0,"202"}, {"Leu",'L',0,"202"},
/*TCX*/ {"Ser1",'S',1,"004"}, {"Ser1",'S',1,"004"},
              {"Ser1",'S',1,"004"}, {"Ser1",'S',1,"004"},
/*TAX*/ {"Tyr",'Y',1,"002"}, {"Tyr",'Y',1,"002"},
              {"???",'?',1,"002"}, {"???",'?',1,"002"},
/*TGX*/ {"Cys",'C',1,"002"}, {"Cys",'C',1,"002"},
              {"Trp",'W',1,"002"}, {"Trp",'W',1,"002"},

/*CTX*/ {"Leu",'L',0,"004"}, {"Leu",'L',0,"004"},
              {"Leu",'L',0,"204"}, {"Leu",'L',0,"204"},
/*CCX*/ {"Pro",'P',1,"004"}, {"Pro",'P',1,"004"},
              {"Pro",'P',1,"004"}, {"Pro",'P',1,"004"},
/*CAX*/ {"His",'H',1,"002"}, {"His",'H',1,"002"},
              {"Gln",'Q',1,"002"}, {"Gln",'Q',1,"002"},
/*CGX*/ {"Arg",'R',1,"004"}, {"Arg",'R',1,"004"},
              {"Arg",'R',1,"004"}, {"Arg",'R',1,"004"},

/*ATX*/ {"Ile",'I',1,"002"}, {"Ile",'I',1,"002"},
              {"Met",'M',1,"002"}, {"Met",'M',1,"002"},
/*ACX*/ {"Thr",'T',1,"004"}, {"Thr",'T',1,"004"},
              {"Thr",'T',1,"004"}, {"Thr",'T',1,"004"},
/*AAX*/ {"Asn",'N',1,"002"}, {"Asn",'N',1,"002"},
              {"Lys",'K',1,"002"}, {"Lys",'K',1,"002"},

/* Since silent polymorphisms are codon polymorphisms,
     a codon position with codons of both classes of serines
     (AGT,AGC and TCT,TCA,TCC,TCG) will not generate
     three different silent-site permutations. */

/*AGX*/ {"Ser2",'S',1,"002"}, {"Ser2",'S',1,"002"},
              {"???",'?',1,"002"}, {"???",'?',1,"002"},

/*GTX*/ {"Val",'V',1,"004"}, {"Val",'V',1,"004"},
              {"Val",'V',1,"004"}, {"Val",'V',1,"004"},
/*GCX*/ {"Ala",'A',1,"004"}, {"Ala",'A',1,"004"},
              {"Ala",'A',1,"004"}, {"Ala",'A',1,"004"},
/*GAX*/ {"Asp",'D',1,"002"}, {"Asp",'D',1,"002"},
              {"Glu",'E',1,"002"}, {"Glu",'E',1,"002"},
/*GGX*/ {"Gly",'G',1,"004"}, {"Gly",'G',1,"004"},
              {"Gly",'G',1,"004"}, {"Gly",'G',1,"004"},

/*ERROR*/ {"!!!",'!',2,"000"}, { 0 } };



/* Input file format */
typedef enum { F_NONE, F_ASF, F_ASFLISTPOLY, F_NEXUS, F_CLUSTAL,
     F_PHYLIP, F_FASTA, F_NBRF_PIR }  FFORMAT;

static FFORMAT fformat=F_NONE;


/* Prototypes for local functions */

/* Functions for reading <Seqfile> and friends */

/* Functions for opening <Seqfile> */

static FFORMAT read_fftype(void);
static void sqf_rewind(long offset);
static void say_fewseqs(char *bleep, int ns);
static const char *parse_dcomm(const char *str);
static void init_rpoly(void);
static void set_slabels(int nseqs);
static void init_rg_arrays(int nseqs);
static int readgroup(void);
static void init_stag(int nseqs);

/* Functions for reading lines and bytes */
static char *lineread(int allow_parse);
static void pushline(char *p);
static void formerr(void);
static char *Show_lbuff(void);
static void reset_sqf_vals(void);
static void reset_readbuff(void);

static int fbuffc(void);
static int unfbufc(int c);

/* FASTA/NBRF_PIR-reading package */
static void read_fasta_nbrf(void);
static void fin_unal_seqs(void);

/* Nexus-reading package */
static void read_nexus_parms(void);
static char *getfirtoken(void);
static void nx_enderr(void);
static void readnum (int *ip);
static void readchar(char *ch);
static void readword(char **strp);
static void nx_comments(STRLISTP strlp);

/* Functions for grouplist[] */
static void show_wlist(void);
static void parse_seqlist(STRLIST list, char flag);
static void show_skipped_seqs (void);
static void set_bloc_err (int ns);
static SEQNAME make_sqn(int r);
static void build_groups(void);

static void sort_groups(void);

/* Functions for reading polysites[] */
static int rewrite_stag(void);
static int rewrite_sqbases(void);
static char *poly_rewrite(char *oldsites);
static void maybe_stitch(int nseqs);
static void save_stitch(int nseqs);

static void storebaseval (int c);
static void reset_baseval (void);

/* Functions for reading sites */
static char *readsite(int *offptr);
static void dosite (char *site,int nseqs, int offset, int mcnt, int polyflag);
static void trisite (char *feedin,int nseqs, int offset,
       int mcnt, int polyflag);
static void addrun (char *site,int nseqs, int offset, int polyflag);
static void init_sqbase_read(void);
static void init_poly_arrays(void);
static SEQTYPE get_sqtype(void);

static char *readpolysite(int *offptr);
static AACODE *getcode(char *p);
static void addpolysite (char *site,int nseqs,
        int firstoff, int lastoff, int classflag);
static char *addstitch (char *state, char *site, int nseqs);
static char newlett (int initflag);
static int havmis(char *site, int nseqs);

static int skip_range (int offset);
static void set_first_last(void);


/* (***AAA) */
/* Open <Seqfile> */

static char *lrfname, *lrsfname;
static FILE *lrff;

/* <Seqfile>s are read in binary mode to handle */
/*    DOS, MAC, and UNIX text files */

/* NOTE: Formats that can be opened and read:              */
/*       Format:          Opened by         Sites read by  */
/*   FASTA/NBRF_PIR       read_fasta_nbrf   same           */
/*   PHYLIP INTERLEAVED:  OpenSeqFile       readgroup      */
/*   NEXUS:                                                */
/*       Character        read_nxdat        readgroup      */
/*       Data             read_nxdat        readgroup      */
/*       Unaligned        read_nxdat        rdnx_unalmat   */
/*   ASF:                                                  */
/*       Interleaved      OpenSeqFile       readgroup      */
/*       ListedPoly       init_rpoly        readpolysite   */
/*                                                         */

void OpenSeqFile(void)
 { char *rs, sbuff[66];  const char *pc;
   /* Open <Seqfile> in binary mode for DOS,UNIX, or MacIntosh files. */
   /* If no path is given in seqfile, and if `seqfile' is not in */
   /*   the default directory, then look also along `filepath', */
   /* If not in the current directory, put the full path in `seqfixed_dir' */
   lrff=fxopenread(seqfile,&lrfname,&seqfdir,&seqfixed_dir,FALSE);
   lrsfname=strxdup(shortname(sbuff,lrfname));
   seqf_open=TRUE;

   /* Where to go in a crash to return system resources */
   err_clear=TRUE;  err_clear_funp=CloseSeqFile;
   /* Initialize variables */
   init_readvars();
   /* Set some variables depending on `seqtype', such as */
   /*   the recognized characters (`read_knownchars') etc. */
   set_kchars();

   /* Is `seqfile' an ASF, NEXUS, CLUSTAL, FASTA, NBRF_PIR, etc. file? */
   /* Check for #NEXUS , CLUSTAL, OR  an initial > at offsets */
   /*   0, 128, and 256 . */
   /* WARNING: This can misclassify files with stray > characters */
   /*   unless  -nomac or -nomacfork  is used. */
   fformat=read_fftype();
   /* Get_ffname() uses fformat */
   textprintf("\n%s format sequence file `%s'\n",
      get_ffname(), lrfname);
   /* Gapchar: Used only for */
   /*   (i) If matchchar==gapchar or misschar, set matchchar=0  */
   /*   (ii) In FASTA/NBRF format, include as data (otherwise ignore) */
   /* Misschar: Used only for */
   /*   (i) If matchchar==gapchar or misschar, set matchchar=0  */
   /*   (ii) As a fill character in unaligned formats. */
   /*        For this purpose, NEXUS and FASTA/NBRF formats use '?' */
   /*          if misschar==0 */
   /* Commchars:  Currently used ONLY in ASF format */
   switch(fformat) {
     case F_FASTA:  case F_NBRF_PIR:
        /* Read the sequences in the alignment */
        matchchar=0;  unaligned_data=havlabels=TRUE;
        read_fasta_nbrf();  break;
     case F_NEXUS:
        /* For unaligned NEXUS data, */
        /*   read the sequences in the alignment */
        /* For aligned (interleaved) NEXUS data, */
        /*   read size data and perhaps tag names */
        /* NEXUS format says the defaults are no matchchar, */
        /*   no gapchar, misschar=='?', and labels. */
        /* Changes to either MUST BE DECLARED in the NEXUS file */
        gapchar=matchchar=0;  misschar='?';  havlabels=TRUE;
        read_nexus_parms();
        if (nseqs==0)
           ErrHandler("%sNo sequences declared or found!",NexFormErr);
        break;
     case F_CLUSTAL:
        matchchar=0;  gapchar='-';  havlabels=TRUE;
        /* Read the description line, NOT allowing for parsed comments */
        /* CLUSTAL files have NO size line */
        if ( (rs=lineread(FALSE))==NULL)  formerr();
        /* Initial spaces are kept when reading CLUSTAL format */
        while(isspace(*rs)) rs++;
        textprintf("\nSequence file `%s'.  Heading:\n"
           " %s\n", lrfname, rs);
        addstr(&seqfcomments,rs);   /* Save for output file */
        break;
     case F_PHYLIP:
        matchchar=0;  havlabels=TRUE;
        /* The size line should be the first line */
        if ( (rs=lineread(TRUE))==NULL)  formerr();
        if (sscanf(rs, "%d %d", &nseqs, &nbases) < 1)
          ErrHandler ("CANNOT READ THE SIZE LINE of the PHYLIP file%s `%s'\n"
            "Line read: `%s'",
            (strlen(lrfname)>20) ? "\n " : "", lrsfname, rs);
        break;
     default:  /* Assume ASF format */
        /* Read the description line, checking for parsed comments BEFORE */
        /*   the description line but NOT IN the description line. */
        if ( (rs=lineread(TRUE))==NULL)  formerr();
        textprintf("\nSequence file `%s'.  Heading:\n"
           " %s\n", lrfname, rs);
        addstr(&seqfcomments,rs);   /* Save for output file */
        /* Read the size line, checking for any parsed comments */
        /*   BEFORE the size line.  (The size line is not a comment.) */
        if ( (rs=lineread(TRUE))==NULL)  formerr();
        /* Init_readvars() sets nbases=-1 */
        if (sscanf(rs, "%d %d %d", &nseqs, &nbases, &r_polyflag) < 1)
          ErrHandler ("CANNOT READ THE SIZE LINE of the ASF file%s `%s'\n"
            "Line read: `%s'",
            (strlen(lrfname)>20) ? "\n " : "", lrsfname, rs);
        if (r_polyflag)  fformat=F_ASFLISTPOLY;
        /* Check if the size line has any parsed commands: */
        if ( (pc=parse_dcomm(rs))!=NULL)
         { char *rbuff=mxalloc(strlen(pc)+100);
           sprintf(rbuff, "From size-data line: %s", pc);
           addstr(&seqfcomments,rbuff);  free(rbuff); }
        /* Read another line  to make sure that all parsed comments */
        /*   are read before the sequence data */
        /* No further comments will be parsed for commands. */
        if ( (rs=lineread(TRUE))==NULL)  formerr();
        pushline(rs); }

   /* Nseqs should be known now except for these formats */
   if (nseqs==0 && !(fformat==F_NEXUS || fformat==F_CLUSTAL))
     ErrHandler("%s format SEQUENCE FILE `%s' has no sequences !!\n",
        get_ffname(), lrfname);

   /* Make sure that `matchchar' is not interpreted foolishly */
   if (matchchar==gapchar || matchchar==misschar)  matchchar=0;

   if (nseqs>0)
    { textprintf("%d sequence%s   ", nseqs, (nseqs==1) ? "" : "s");
      file_nseqs=nseqs;
      if (nbases < 0)  textprintf("Number of bases not specified.\n");
      else  textprintf("%d bases.\n", nbases);
      if (nseqs<=2)  say_fewseqs("", nseqs); }

   /* Keep sequences in memory in these cases */
   saveseqs = (saveseqs || annotate || showunal
       || dumpjseqs || seqtype==AUTO);

   if (fformat==F_ASFLISTPOLY)  init_rpoly();  /* Read tag names */
   else if (!unaligned_data)   /* Including all interleaved files */
    { /* Read the first group so that  strtag[]  will be valid. */
      init_rg_arrays(nseqs);
      if (readgroup()<0)
        ErrHandler("CANNOT READ the first group "
           "of sequences in %s!\n%s", lrsfname, Show_lbuff()); }
   /* Nseqs is now known for all formats */
   sitebuff = mxalloc(nseqs+3); }


/* What is the <Seqfile> file format ? */

static FFORMAT ff_test_off(long int fof)
 { char *rs;
   if (fof>0L)  sqf_rewind(fof);
   if ( (rs=getfirtoken())!=NULL)
    { if (!stricmp(rs,"#NEXUS"))  return F_NEXUS;
      else if (!strcmp(rs,"CLUSTAL"))
       { sqf_rewind(fof);  return F_CLUSTAL; }
      else if (rs[0]=='>')
       { sqf_rewind(fof);  /* Start reading data here */
         return  (strlen(rs)>=4 && rs[3]==';')  ? F_NBRF_PIR : F_FASTA; }}
   return F_NONE; }


static FFORMAT read_fftype(void)
 { FFORMAT ff;
   if (phylip)  return F_PHYLIP;
   if ( (ff=ff_test_off(0L))!=F_NONE)  return ff;
   if (nomacfork==FALSE)
    { if ( (ff=ff_test_off(128L))!=F_NONE
            || (ff=ff_test_off(256L))!=F_NONE)  return ff; }
   /* Default assumes ASF format with no Mac resource fork */
   sqf_rewind(0L);
   return  F_ASF; }


char *get_ffname(void)
 { switch(fformat) {
     default:  return "UNKNOWN";
     case F_ASF:          return "ASF";
     case F_ASFLISTPOLY:  return "ASF listed polymorphism";
     case F_NEXUS:        return "NEXUS";
     case F_CLUSTAL:      return "CLUSTAL";
     case F_PHYLIP:       return "PHYLIP INTERLEAVED";
     case F_FASTA:        return "FASTA";
     case F_NBRF_PIR:     return "NBRF/PIR"; }}


int use_commchars(void)
 { return (fformat==F_ASF || fformat==F_ASFLISTPOLY); }


static void kill_linebuffs(void)
 { if (seqline) { free(seqline);  seqline=NULL; }
   if (matchbuff) { free(matchbuff);  matchbuff=NULL; }
   kill_nstr(&rdrow,nseqs);
   matchxlen=linxlen=0; }


void CloseSeqFile(void)
 { if (lrff)
    { fxclose(lrff,lrfname);  lrff=NULL;  reset_sqf_vals(); }
   seqf_open=FALSE;
   kill_linebuffs();
   kill_nstr(&readshowtag,nseqs); }


char *set_offtypes (void)
 { return (seqtype==PROTEIN) ? proteins : total_nucls; }


static void sqf_rewind(long offset)
 { rewind(lrff);  reset_sqf_vals();
   if (offset>0L)  fseek(lrff,offset,SEEK_SET); }


static void say_fewseqs(char *bleep, int ns)
 { if (ns<=1)
     ErrHandler("%sToo few sequences (n=%d)!\n", bleep, ns);
   else if (include_monosites==FALSE)
     ErrHandler("%sToo few sequences (n=%d)!\n"
   "You must have at least three sequences for polymorphisms to be\n"
   "informative about possible gene conversion events.\n",
        bleep, ns); }


void say_toofewpolys(void)
 { if (npolys<=0) ErrHandler("NO POLYMORPHISMS:  CANNOT ANALYZE !!");
   else
    { char nbuff[88];
      if (npolys==1) strcpy(nbuff,"Only one polymorphism");
      else numsay(nbuff,npolys, "polymorphism",TRUE);
      ErrHandler ("%s: Too few to analyze!", nbuff); }}


static void rp_flag_warn(char *bleep)
 { WarnHandler("THE `LISTED POLYMORPHISM' FORMAT CANNOT BE USED\n"
      "  %s\nSee the documentation for more details.\n", bleep); }


static void init_rpoly(void)
 { textprintf("LISTED POLYMORPHISM FORMAT.\n");
   if (seqtype==SILENT)
    { rp_flag_warn("with the silent site flag `R'.");
      seqtype=ALL; }
   if (annotate || dumpjseqs || showunal)
    { rp_flag_warn("with `annotate', `dumpjseqs', or `showunal'.");
      annotate=dumpjseqs=showunal=FALSE; }

   init_stag(nseqs);
   if (havlabels==FALSE)  set_slabels(nseqs);
   else
    { int i=0;  char *rs;
      /* Try to read `nseqs' sequence names in a row */
      while (i<nseqs && (rs=lineread(FALSE))!=NULL)
       { char *nextp;
         do { rs=parse_unquote(rs,&nextp,NULL);
              strtag[i] = gettag(rs,&readtag[i],NULL,'S');  i++; }
         while (i<nseqs && (rs=nextp)!=NULL); }
      if (i<nseqs)
        ErrHandler("CANNOT READ %d SEQUENCE NAMES %s !!\n%s",
          nseqs, ExpPolyForm, Show_lbuff()); }
   tags_set=TRUE;
   if (saveseqs)  seqbase=cxalloc(nseqs+3,sizeof(char *));
   nread=nxread=0;   /* To be safe */

   /* Initialize the configuration-reading routine */
   next_rpoly=NULL; }



/* (***BBB) */
/* String-reading utilities */

static const char *pcommfirst;

static const char *striastr(const char *str, const char *match)
 { const char *p;  size_t mlen=strlen(match);
   /* Count a match ONLY at the first character or if the */
   /*   preceding character is blank. */
   for (p=str; *p!=0; p++)
     if (!strincmp(p,match,mlen) && (p==str || isspace(p[-1])))
       { if (pcommfirst==NULL || p<pcommfirst)  pcommfirst=p;
         return p+mlen; }
   return NULL; }


static char *dist_str (const char *p)
 { char *rs, *pbuff=strxdup(p);
   rs=parse_unquote(pbuff,NULL,NULL);
   while(isspace(*rs)) rs++;  rs=strxdup(rs);  free(pbuff);
   return rs; }


/* Called only for ASF format and if use_commchars()==TRUE */
/*   (Current ASF format only.) */

static const char *parse_dcomm(const char *str)
 { const char *p;  pcommfirst=NULL;
   while(isspace(*str)) str++;
   /* Several options can be set on one line, */
   /*   so can't use ``else if''s  */
   if ( (p=striastr(str,"commchar="))!=NULL
       || (p=striastr(str,"commchars="))!=NULL)  commchars=dist_str(p);
   if ( (p=striastr(str,"seqtype="))!=NULL)
    { read_sqtype(p);  checkflags();  set_kchars(); }
   if ( (p=striastr(str,"matchchar="))!=NULL)  matchchar=set_vchar(p);
   if ( (p=striastr(str,"gapchar="))!=NULL)  gapchar=set_vchar(p);
   if ( (p=striastr(str,"misschar="))!=NULL)  misschar=set_vchar(p);
   if ( (p=striastr(str,"numbersep="))!=NULL)  numbersep=set_vchar(p);
   if ( (p=striastr(str,"nolabels"))!=NULL)
    { if (*p==0 || isspace(*p))  havlabels=FALSE;
      else if (*p=='=')  read_nflag(p+1,&havlabels,NULL); }
   if (matchchar==gapchar || matchchar==misschar)  matchchar=0;
   return pcommfirst; }


static void formerr(void)
 { ErrHandler(Show_lbuff()); }



/* (***CCC) */
/* Read sequences from a FASTA or NBRF/PIR format file */

static void r_fasta_err (const char *bleep, int ns)
 { textprintf ("FASTA/NBRF FORMAT:  %s (Seq #%d)\n", bleep,ns+1);
   formerr(); }

/* NOTE: These formats have sequence comments, not file comments */

static void read_fasta_nbrf(void)
 { char *p, *rs;  int ns=0, nxseqs=0;
   char c, xmchar=(misschar!=0) ? misschar : '?';
   char *linbuff=NULL;  int un,uxn=0, umax=-1;
   for(;;)   /* Loop over sequences */
    { /* Read the first line */
      if ( (rs=lineread(FALSE))==NULL || rs[0]!='>')
        r_fasta_err ("Sequence line error", ns);
      if (fformat==F_NBRF_PIR)
       { if (strlen(rs)<4 || rs[3]!=';')  r_fasta_err("Missing ;",ns);
         rs+=3; }
      /* Get the sequence name */
      p=parse_unquote(rs+1,&rs,NULL);
      if (p==NULL || *p==0)  r_fasta_err("Missing sequence name",ns);
      if (ns>=nxseqs)
       { nxseqs+=20;
         strtag=rexxalloc(strtag,sizeof(char *), ns,nxseqs);
         readtag=rexxalloc(readtag,sizeof(char *), ns,nxseqs);
         seqbase=rexxalloc(seqbase,sizeof(char **), ns,nxseqs); }
      strtag[ns] = gettag(p,&readtag[ns],NULL,'S');
      /* Read NBRF/PIR format sequence comment */
      if (fformat==F_NBRF_PIR && (rs=lineread(FALSE))==NULL)
        r_fasta_err("End of file at comment",ns);
      /* Save the first comment for the output file */
      if (ns==0 && rs!=NULL && rs[0]!=0)
       { addstr(&seqfcomments, "First sequence comment:");
         addstr(&seqfcomments, rs); }
      /* Now read the sequence */
      un=0;
      if ( (rs=lineread(FALSE))!=NULL)
       { do { if (rs[0]=='>')  break;   /* The next sequence */
              while ( (c=*rs++)!=0)
                /* All other characters than these are ignored */
                if (isalpha(c) || c=='-' || c=='*' || c==gapchar)
                 { if (c=='*' && fformat==F_NBRF_PIR)  continue;
                   if (un>=uxn)
                     linbuff=rexxalloc(linbuff,1, un,uxn+=LINEBAT);
                   linbuff[un++]=c; }}
         while ( (rs=lineread(FALSE))!=NULL); }
      /* Finish up and save the sequence */
      textprintf ("Read `%s':  length=%d\n", strtag[ns], un);
      linbuff[un]=0;
      /* Preliminary padding */
      if (umax<un)  umax=un;
      if (un<umax)  memset(linbuff+un, xmchar, umax-un);
      seqbase[ns++]=strxdup(linbuff);
      if (rs!=NULL)  pushline(rs);
      else break; }
   nseqs=ns;  nread=nbases=umax;
   textprintf ("%d bases read.\n", nbases);
   fin_unal_seqs();
   if (linbuff)  free(linbuff);
   }


static void fin_unal_seqs(void)
 { char xmchar;  int i;  tags_set=TRUE;
   CloseSeqFile();  err_clear=FALSE;  err_clear_funp=NULL;
   /* Even out the sequences with `misschar' if they have */
   /*  have different lengths, and convert sequences to upper case */
   xmchar=(misschar!=0) ? misschar : '?';
   for (i=0; i<nseqs; i++)
    { size_t len;
      if ( (len=strlen(seqbase[i]))<nbases)
       { seqbase[i]=rexxalloc(seqbase[i],1, len,nbases);
         memset(seqbase[i]+len, xmchar, nbases-len); }
      makeuc(seqbase[i]); }}




/* (***DDD) */
/* Read data from a NEXUS file, */
/*   and initialize interleaved formats for readgroup()  */

static void NexusErr(char *bleep)
 { ErrHandler ("%s%s\n%s", NexFormErr, bleep, Show_lbuff()); }

/* If this list includes `#', it will interfere with */
/*   reading the #NEXUS token */

static int is_nxpunct(char c)
 { return  (strchr("(){}/\\,;:=*'\"+-<>", c)!=NULL); }

/* Return a token from any text file */
/* Returns NULL if overflow and oflowp==NULL */

#define FIRTOKLEN 10

static char *getfirtoken(void)
 { int c,n;  static char wbuff[FIRTOKLEN+3];
   while ( (c=fbuffc())>=0 && isspace(c))  ;
   if (c<0)  return NULL;  n=0;
   do  wbuff[n++]=c;
   while(n<FIRTOKLEN && (c=fbuffc())>=0 && !isspace(c));
   wbuff[n++]=0;  wbuff[n]=0;
   if (ShowNexusToks)  printf (" TOKEN:%s ", wbuff);
   return  wbuff; }

static char *save_nxtok, save_ispun;

static void pushnxtok (char *p, char ispun)
 { save_nxtok=p;  save_ispun=ispun; }


/* Following the NEXUS specifications,    */
/*   the following function should return ONE OF   */
/*   (i) A word,                          */
/*   (ii) A single punctuation character,  OR */
/*   (iii) NULL, for EOF    */
/*                                            */
/*   In the NEXUS specification, the definition for a word is:  */
/*
 ``Word.---Any string of text characters bounded by whitespace or
punctuation is a NEXUS word. Unless the word is delimited by single
quotes, the string cannot contain whitespace or punctuation. If the first
character of a word is a single quote, then the word ends with the next
single quote (unless that single quote is in a pair of consecutive single
quotes; if so, then the word ends at the first unpaired single quote). Any
character, including punctuation and whitespace, may be contained within a
quoted word. A word cannot consist of only whitespace and punctuation.

 ``Underscores are considered equivalent to blank spaces, except that
underscores are dark characters, and blank spaces are whitespace. Thus, a
program encountering B._zephyrum and 'B. zephyrum' should judge them to be
identical.

 ``Any doubled single quotes within a quoted word should be converted to
single quotes within the program. Thus, 'John''s' would be treated
internally within a program as John's. Of course, if the program writes
out such a word to a NEXUS file, the quote should be redoubled, and quotes
should be written around the word: 'John''s'. A string of dark characters
is broken into several words by whitespace and any punctuation other than
brackets, unless the string is surrounded by quotes. Thus,
[\i]Bembidion_velox[\p]_Linn consists of one word, as does '  A single
word', whereas two()words consists of four tokens: the word two, two
punctuation characters, and the word words.''

  The NEXUS specification defines ``Whitespace'' as one of tab (0x09),
``blank'' (0x20), a ``newline character'' (0x0D, 0x0A, or 0x0D 0x0A
combination), or ASCII 0x00-0x06 inclusive.


     The definition of NEXUS word is more than a little fuzzy. A NEXUS word
is not defined; it is only said that certain strings are NEXUS words. One
reading of the definition would say that the string, ``seqfile='No Way'''
consists of three NEXUS tokens (seqfile='No, Way, and ') instead of being
one token, since the single quotes do not delimit the entire word.
However, since the first sentence of the description is not worded to be
all-inclusive, it may not be excluded that this defines a single NEXUS
token.

     The definition seems to exclude double quotes ("), but this may be an
oversight. Again, the definition is not worded to be all-inclusive.

     The definition does allow you to begin a string with a single quote by
entering (for example) '''John''s other office' .

     Again, the definition above is unclear, but may be consistent with
(for example) the strings    seqfile="No Way"  and
 seqfile=N"o wa"y   both being read as   seqfile=No Way, as well
 as   seqfile=N"o W"a"y Atall"  as   seqfile=No Way Atall

     The NT3.5.1 command-line parser does both of the above, but
 treats a single quote as a normal black-space character.

*/



/* NOTE: The following unquoting logic appears in 4 places: */
/*   (i) The following, for reading NEXUS tokens */
/*   (ii) In parse_unquote() in vutil.c  for parsing strings */
/*   (iii) In readtok() in vsetopts.c for reading CFG files */
/*   (iv) In getname2() in vrtab.c for parsing pairs */
/*          of sequence names */

/* Here *punctptr==0 UNLESS the token is a single, unquoted, punctuation */
/*   character, in which case *punctptr is the same as that character */
/*   (and is also the first character of the returned token). */

static char *getnxtoken(char *punctptr)
 { char ispun;  int i,c,n,qc;
   static char wbuff[MAXTOKSIZE+7];
   if (save_nxtok)
    { char *p=save_nxtok;  save_nxtok=NULL;
      if (punctptr) *punctptr=save_ispun;  save_ispun=0;
      if (ShowNexusToks)  printf (" RETRIEVED TOKEN:%s ", p);
      return p; }
   ispun=0;  if (punctptr) *punctptr=0;
   /* Skip over initial spaces */
   while ( (c=fbuffc())>=0 && isspace(c))  ;
   if (c<0)  return NULL;
   /* Initialize at beginning of word and no quotefield */
   n=0;  qc=-1;
   do { /* First, check for a quote character */
        if (c=='"' || c=='\'')
         { if (c==qc)    /* Quotefield is on */
            { /* If the next c==qc, doubled quotes, so fall through */
              /*   and display one qc */
              /* If the next c!=qc, end quotefield and save new c */
              if ( (c=fbuffc())!=qc)
               { if (c<0) break;
                 unfbufc(c);  qc=-1;  continue; }}  /* End match */
           /* Fall through if c is a doubled qc character */
           /*   or else is the other qc character */
           else if (qc==-1) { qc=c;  continue; }}  /* Begin match */
        else if (qc==-1)   /* Not qc, quotefield is off */
         { if (isspace(c))  break;      /* End of token */
           /* Push back a closing punctuation character, then end token */
           else if (is_nxpunct(c))
            { if (n==0)    /* Single punctuation character */
               { wbuff[n++]=c;  ispun=c;
                 if (punctptr) *punctptr=c;  break; }
              else         /* Terminating punctuation character */
               { unfbufc(c);  break; }}}
        else if (c==0x0A)  /* Quotefield is on, EOL */
         { textprintf ("WARNING (line #%d, token so far=%s)\n"
           "  LINE END in the middle of a token within quotes!!!\n"
           "  Ending the token at this point.\n",
             linenum, wbuff);  break; }
        /* Now store or try to store the character */
        if (n<MAXTOKSIZE)  wbuff[n++]=c;
        else { textprintf ("OVERFULL NEXUS TOKEN! "
          "(size %d, line #%d)\n"
          "  Breaking into two tokens.\n", MAXTOKSIZE, linenum);
               unfbufc(c);  break; }}
   while( (c=fbuffc())>=0);
   /* A NEXUS token cannot consist entirely of */
   /*   quoted spaces and punctuation */
   if (ispun==FALSE)
    { for (c=' ', i=n; --i>=0; )
       { if (is_nxpunct(wbuff[i]))  c=wbuff[i];
         else if (!isspace(wbuff[i]))  break; }
      if (i<0)   /* All spaces and punctuation */
       { wbuff[0]=c;  wbuff[1]=0;  if (punctptr) *punctptr=c; }}
   wbuff[n++]=0;  wbuff[n]=0;
   if (ShowNexusToks)  printf (" TOKEN:%s ", wbuff);
   return  wbuff; }






static char *wait_until_nxtoken(char *mat)
 { char *p;
   while ( (p=getnxtoken(NULL))!=NULL)
    { if (!stricmp(p,mat))  return p; }
   return NULL; }


static char *wait_until_eithertoken(char *m1, char *m2)
 { char *p;
   while ( (p=getnxtoken(NULL))!=NULL)
    { if (!stricmp(p,m1) || !stricmp(p,m2))  return p; }
   return NULL; }



char *nx_datatype;
static int interleave;

/* Read NEXUS blocks */
static void read_nxtaxa(void);
static void read_nxdat(int allow_newtaxa, int unaligned, char *bleep);

/* Read NEXUS commands */
static void rdnx_taxlabels(void);
static void rdnx_dimens(int allow_newtaxa, char *bleep);
static void rdnx_format(void), rdnx_unalmat(void);

/* Read NEXUS file */
static void read_nexus_parms(void)
 { /* Look for  begin <name>;        */
   /*   for  data, taxa, characters, and/or unaligned blocks. */
   /* For unaligned data,  read_nxdat() reads the entire dataset. */
   /* For interleaved data,  read_nxdat() reads up to and */
   /*   including the `matrix' command, and initializes for */
   /*   calling `readgroup()'. */
   for(;;)
    { char *p;
      /* This returns the NEXT token after `begin' */
      if (wait_until_nxtoken("begin")==NULL
        || (p=getnxtoken(NULL))==NULL)  return;
      if (!stricmp(p,"taxa"))  { read_nxtaxa();  continue; }
      else if (!stricmp(p,"characters") || !stricmp(p,"character"))
       { read_nxdat(FALSE,FALSE,"character");  return; }
      else if (!stricmp(p,"data"))
       { read_nxdat(TRUE,FALSE,"data");  return; }
      else if (!stricmp(p,"unaligned"))
       { read_nxdat(FALSE,TRUE,"unaligned");  return; }
      if (wait_until_eithertoken("end","endblock")==NULL)  return; }}



/* Read a NEXUS taxa block */

static void read_nxtaxa(void)
 { char ispun, *p;
   while ( (p=getnxtoken(NULL))!=NULL)
    { if (!stricmp(p,"dimensions"))
       { if (nseqs>0)  NexusErr ("(taxa) dimension statement\n"
           "  when number of sequences HAS ALREADY BEEN DEFINED");
         while ( (p=getnxtoken(&ispun))!=NULL && ispun!=';')
          { if (!stricmp(p,"ntax"))  readnum(&nseqs); }}
      else if (!stricmp(p,"taxlabels"))  rdnx_taxlabels();
      else if (!stricmp(p,"begin"))  nx_enderr();
      else if (!stricmp(p,"end") || !stricmp(p,"endblock"))  return; }}


/* Read a NEXUS taxlabels command */

static void rdnx_taxlabels(void)
 { char ispun, *rs;  int n;
   if (nseqs<=0)  NexusErr ("(taxlabels) the number of sequences\n"
         "  has NOT been defined!");
   if (strtag!=NULL)  NexusErr ("(taxlabels) the sequences\n"
         "  have ALREADY been defined !!");
   strtag = cxalloc(nseqs+2,sizeof(char *));
   readtag = cxalloc(nseqs+2,sizeof(char *));
   readshowtag = cxalloc(nseqs+2,sizeof(char *));
   n=0;
   while ( (rs=getnxtoken(&ispun))!=NULL && ispun!=';')
    { if (n<nseqs)
        strtag[n]=gettag(rs,&readtag[n],&readshowtag[n],'S');
      n++; }
   if (n!=nseqs)  ErrHandler ("%s(taxlabels) %d taxa claimed,\n"
         "  %s taxon names found.\n%s", NexFormErr, nseqs, n, Show_lbuff());
   if (ShowNexusToks)  printf ("%d taxon names read.\n", nseqs);
   tags_set=TRUE; }



/* Read a NEXUS dimensions command */

static void rdnx_dimens(int allow_newtaxa, char *bleep)
 { char ispun, *p;  int ntaxa=-1;
   while ( (p=getnxtoken(&ispun))!=NULL && ispun!=';')
    { if (!stricmp(p,"newtaxa"))  allow_newtaxa=TRUE;
      else if (!stricmp(p,"ntax"))
       { readnum(&ntaxa);
         if (nseqs>0 && nseqs!=ntaxa)
           ErrHandler (
      "%s(dimensions) (L%03d) Number of sequences defined before as %d\n"
      "  is redefined as %d (nbases=%d)\n%s",
             NexFormErr, linenum, nseqs, ntaxa, nbases, Show_lbuff());
         nseqs=ntaxa; }
      else if (!stricmp(p,"nchar"))  readnum(&nbases); }
   if (ShowNexusToks)
     printf ("\nNEXUS dimension statement:  "
        "nseqs=%d  ntaxa=%d  nbases=%d\n", nseqs,ntaxa,nbases);
   if (allow_newtaxa==FALSE && ntaxa>0)
      textprintf (
       "\nWARNING: Taxa are defined in a NEXUS %s block\n"
       "without a `newtaxa' statement or a previous `taxa' block.\n"
       "(This is a technical NEXUS format error.)\n\n", bleep); }




/* Read a NEXUS format command */
/* NEXUS format says no default gapchar but default misschar='?' */

static void rdnx_format(void)
 { char ispun, *p;
   while ( (p=getnxtoken(&ispun))!=NULL && ispun!=';')
    { if (!stricmp(p,"datatype"))
       { readword(&nx_datatype);
         if (!stricmp(nx_datatype,"protein"))  seqtype=PROTEIN;
         else if (!stricmp(nx_datatype,"dna") && seqtype!=SILENT)
           seqtype=ALL; }
      else if (!stricmp(p,"missing"))  readchar(&misschar);
      else if (!stricmp(p,"gap"))  readchar(&gapchar);
      else if (!stricmp(p,"matchchar"))  readchar(&matchchar);
      else if (!stricmp(p,"nolabels"))  havlabels=FALSE;
      else if (!stricmp(p,"interleave"))  interleave=TRUE; }
   if (matchchar==gapchar || matchchar==misschar)  matchchar=0; }




/* Read a NEXUS data or characters block */
/* Exit in a state ready to call readgroup() */

static void read_nxdat(int allow_newtaxa, int unaligned, char *bleep)
 { char *p;  unaligned_data=unaligned;
   while ( (p=getnxtoken(NULL))!=NULL)
    { if (!stricmp(p,"comment") || !stricmp(p,"comments"))
        nx_comments(&seqfcomments);
      else if (!stricmp(p,"dimensions"))
        rdnx_dimens(allow_newtaxa,bleep);
      else if (!stricmp(p,"format"))  rdnx_format();
      else if (!stricmp(p,"taxlabels"))  rdnx_taxlabels();
      else if (!stricmp(p,"matrix"))
       { if (unaligned_data) rdnx_unalmat();  return; }
      else if (!stricmp(p,"begin"))  nx_enderr();
      else if (!stricmp(p,"end") || !stricmp(p,"endblock"))  break; }
   NexusErr( (unaligned) ?
     "Unaligned data block\n  WITHOUT a `matrix' statement"  :
     "Data or character block\n  WITHOUT a `matrix' statement"); }




/* Read a NEXUS unaligned data matrix */

static void rdnx_unalmat(void)
 { char *linbuff=NULL;
   char xmchar = (misschar) ? misschar : '?';
   int c,ns=0, un,uxn=0, umax=-1;
   if (nseqs<=0)   NexusErr ("(Unaligned) Matrix command\n"
     "  when the number of sequences is UNDEFINED");
   /* Initialize sequence arrays if not done already */
   if (strtag==NULL)  strtag=cxalloc(nseqs+2,sizeof(char *));
   if (readtag==NULL)  readtag=cxalloc(nseqs+2,sizeof(char *));
   if (seqbase==NULL)  seqbase=cxalloc(nseqs+2,sizeof(char *));
   if (havlabels==FALSE && tags_set==FALSE)
    { set_slabels(nseqs);  tags_set=TRUE; }
   do { char ispun, *rs, q1b[MAXNAMEBUFF];
        /* Unless labels are suppressed, get the label or sequence */
        /*  tag name for sequence #n. Note that embedded spaces or */
        /*  punctuation must be quoted. */
        if (havlabels)
         { if ( (rs=getnxtoken(&ispun))==NULL || ispun==';')  break;
           if (ns<nseqs)
            { if (tags_set==FALSE)     /* No tag names defined yet */
               { if (strtag[ns])
                   ErrHandler("%s(Unaligned) Sequence #%d:\n"
                     "  Tag string `%s' is already defined!\n%s",
                            NexFormErr, ns+1, strtag[ns], Show_lbuff());
                 strtag[ns]=gettag(rs,&readtag[ns],NULL,'S'); }
              else if (striucmp(rs,readtag[ns]))
                ErrHandler("%s(Unaligned) Sequence #%d (%s):\n"
                  "  should be `%s', found `%s'\n%s",
                    NexFormErr, ns+1, strtag[ns],
                    readshowtag[ns], doquote(q1b,rs), Show_lbuff()); }
           else ErrHandler (
             "%s(Unaligned) %d seqs predicted, reading #%d!\n%s",
                NexFormErr, nseqs, ns+1, Show_lbuff()); }
        /* Read data for sequence #n */
        /* Since this code uses fbuffc() with fformat==F_NEXUS, it ignores */
        /*   text within matched []s */
        un=0;
        while ( (c=fbuffc())>=0 && c!=',' && c!=';') if (isbasechar(c))
         { if (un>=uxn)
             linbuff=rexxalloc(linbuff,1, un,uxn+=LINEBAT);
           linbuff[un++]=c; }
        linbuff[un]=0;   /* Finished a sequence */
        if (umax<un)  umax=un;
        if (ns<nseqs)
         { /* Minimize later realloc's */
           if (un<umax)  memset(linbuff+un, xmchar, umax-un);
           seqbase[ns]=strxdup(linbuff); }
        ns++; }
   /* After reading data, c==-1, ',', or ';' */
   /* Continue if ',', otherwise fall through */
   while (c==',');
   if (ns!=nseqs)
     ErrHandler ("%s(Unaligned) %d seqs predicted, found %d\n%s",
        NexFormErr, nseqs, ns, Show_lbuff());
   nread=nbases=umax;
   textprintf ("%d bases read.\n", nbases);
   fin_unal_seqs();
   if (linbuff)  free(linbuff); }


static void nx_enderr(void)
 { NexusErr("2nd `begin' in a block:\n"
    "  Probable missing end or endblock!!!!"); }


/* Read a number, word, or character */
/* Recall that getnxtoken() strips off enclosing quotes */

static void readnum (int *ip)
 { char ispun, *p;
   if ( (p=getnxtoken(&ispun))!=NULL)
    { if (ispun=='=' && (p=getnxtoken(&ispun))!=NULL
            && sscanf(p,"%d",ip)==1) return;
      /* If p==NULL, pushnxtok() clears save_nxtok */
      pushnxtok(p,ispun); }}


static void readchar(char *cp)
 { char ispun, *p;
   if ( (p=getnxtoken(&ispun))!=NULL)
    { if (ispun=='=' && (p=getnxtoken(&ispun))!=NULL)
       { if (*p==';' || *p==',')  *cp=0;
         else { *cp=p[0];  return; }}
      pushnxtok(p,ispun); }}


/* Used only to parse `datatype=....' */

static void readword(char **strp)
 { char ispun, *p;
   if ( (p=getnxtoken(&ispun))!=NULL)
    { if (ispun=='=' && (p=getnxtoken(&ispun))!=NULL)
       { if (*p==';' || *p==',')  *strp=NULL;
         else
           { if (use_underscores)  unspace_str(p);
             *strp=strxdup(p);  return; }}
      pushnxtok(p,ispun); }}


static void nx_comments(STRLISTP strlp)
 { char ispun, *p, *q;
   /* By the NEXUS specification, a token cannot contain */
   /*   embedded EOL characters */
   while ( (p=getnxtoken(&ispun))!=NULL && ispun!=';')
    { /* Skip initial and trailing spaces */
      while (isspace(*p)) p++;
      q=p+strlen(p);  while(q>p && isspace(*--q)) *q=0;
      /* Now add to *strlp */
      addstr(strlp,p); }}




/* (***EEE) Routines for grouplist[] */

/* Set  grouplist[]  to indicate the sequences used         */
/* Uses  seq_list, seq_skip_list, strtag, grouplist, etc.     */

static char *keeplist, *newsites;
static STRLIST warnlist;

static void reset_wlist(void)
 { kill_strlist(&warnlist); }


void setgroups (void)
 { int i, rn;
   org_bsize=org_bnum=0;
   /* Clear the missing-sequence warning list */
   reset_wlist();
   /* Use seq_list and seq_skip_list to fill keeplist[] */
   /*   if specific sequence names are listed/proscribed */
   keeplist=mxalloc(nseqs+2);
   if (seq_list.s==NULL)  for (i=0; i<nseqs; i++)  keeplist[i]=1;
   parse_seqlist(seq_list,1);
   parse_seqlist(seq_skip_list,0);

   /* Construct grouplist[] */
   /* If grouptype==G_BLOCS, also sets blocnum, org_bsize, org_bnum */
   /* If grouptype==G_GROUPS, sequences not in listed groups */
   /*   are ignored. */
   switch(grouptype) {
     default: build_simple_group(TRUE);  break;
     case G_BLOCS:  build_bloc_groups(TRUE);  break;
     case G_GROUPS:  build_groups();  break; }

   /* Called only if G_SIMPLE or G_BLOCS */
   if (warnlist.n>0)
    { show_wlist();  reset_wlist(); }

   /* How many sequences are left? */
   for (rn=i=0; i<nseqs; i++)  if (keeplist[i])  rn++;
   if (rn<nseqs)  show_skipped_seqs();  /* Uses keeplist[] */
   if (rn<=2)
    { char *bleep = (grouptype==G_GROUPS)
         ? "AFTER RESTRICTING SEQUENCES USING THE GROUP STRUCTURE:\n"
         : "AFTER RESTRICTING SEQUENCES: ";
      say_fewseqs(bleep, rn); }
   /* Done with keeplist[], so de-allocate */
   free(keeplist);  keeplist=NULL;

   if (sortgfragsbyseq || sortgfragsboth)  sortnames=TRUE;
   if (sortnames)  sort_groups();
   if (grouptype==G_BLOCS)  set_bloc_names();

   /* Fill  offs,first,last   fields  in grouplist[] */
   /*  and calculate  npairseqs, noutseqs, noutgrps, mxtag, ... */
   set_groupvars();
       }



static void show_wlist(void)
 { char nbuff[199], abuff[44], sbuff[66];
   sprintf(nbuff, "WARNING: %s NOT FOUND:  ",
     numsay(abuff,warnlist.n, "prescribed sequence",TRUE));
   /* Warnlist.s has names in quoted (truncated) form. */
   toklist(nbuff, warnlist.s,warnlist.n);
   show_all_seqnames("Sequences in file:  ");
   numsay(abuff,warnlist.n, "sequence",TRUE);
   ( (notfound_nonfatal) ? WarnHandler : ErrHandler )
     ("%s NOT FOUND in %s!\n", abuff, shortname(sbuff,seqfile)); }


/* The strings in list.s are from either parse_unquote() */
/*   or readtok(), and so are unquoted. */
/* Sequence name comparisons use `striucmp()', which does not */
/*   distinguish between spaces and underscores. */
/* NOTE: Feature to recognize `S33' if the tag name is `33' */
/*   is no longer supported. */

static void parse_seqlist(STRLIST list, char flag)
 { int i,k;  char *p;
   if (list.s==NULL || list.n==0)  return;
   /* Loop over words in the list */
   for (k=0; k<list.n; k++)  if ( (p=list.s[k])!=NULL && p[0]!=0)
    { /* Loop over known sequences */
      for (i=0; i<nseqs; i++)  if (!striucmp(p,readtag[i]))  break;
      if (i<nseqs)  keeplist[i]=flag;
      else if (flag!=0)  addstr(&warnlist, p); }}



static void show_skipped_seqs (void)
 { int i,n;  char **skstr=cxalloc(nseqs+3,sizeof(char *));
   for (n=i=0; i<nseqs; i++)
     if (keeplist[i]==0)  skstr[n++]=strtag[i];
   if (n>0)
    { char nbuff[88];
      sprintf (nbuff, "SKIPPING SEQUENCE%s: ", (n>1) ? "S" : "");
      toklist(nbuff, skstr,n); }
   free(skstr); }


/* Functions for build_simple_group() etc. */

/* Set and test blocnum */

static void set_bloc_err (int ns)
 { ErrHandler("ILLEGAL BLOC SIZE %d!  Must properly divide %d !",
     blocsize, ns); }


/* Fill a template SEQNAME from strtag[] and readtag[] */

static SEQNAME make_sqn(int r)
 { SEQNAME sq;  memset(&sq,0,sizeof(sq));
   sq.name=strxdup(strtag[r]);
   sq.rname=strxdup(readtag[r]);
   sq.orgoffs=r;
   return sq; }



/* Called from IDM_GROUPS in vswin.c if group structure is changed */
/* Only called if grouptype is G_BLOCS or G_SIMPLE */
/* Make_sqn(r) reads from strtag[] and readtag[], */
/*   which are assumed to be valid */

void new_bloc_groups(void)
 { int i,j,r;
   kill_groups();
   if (blocsize==0)
    { GROUPLIST gg;
      grouptype=G_SIMPLE;  memset(&gg,0,sizeof(gg));
      gg.gname=strxdup("All");  gg.grname=strxdup("All");
      gg.dostat=TRUE;
      gg.ns=nseqs;  gg.sqn=cxalloc(nseqs+2,sizeof(SEQNAME));
      for (r=0; r<nseqs; r++)  gg.sqn[r]=make_sqn(r);
      grouplist=cxalloc(2,sizeof(grouplist[0]));
      grouplist[0]=gg;  ngroups=1;
      blocnum=blocsize=org_bsize=org_bnum=0; }
   else
    { if (blocsize<=1 || (blocnum=nseqs/blocsize)<=1
           || blocsize*blocnum!=nseqs)  set_bloc_err(nseqs);
      grouptype=G_BLOCS;  ngroups=blocnum;
      grouplist=cxalloc(blocnum+2,sizeof(grouplist[0]));
      for (r=i=0; i<blocnum; i++)
       { GROUPLIST gg;  memset(&gg,0,sizeof(gg));
         gg.gname=gg.grname="***";  gg.dostat=TRUE;
         gg.ns=blocsize;  gg.sqn=cxalloc(blocsize+2, sizeof(SEQNAME));
         for (j=0; j<blocsize; j++, r++)  gg.sqn[j]=make_sqn(r);
         grouplist[i]=gg; }
      org_bsize=blocsize;  org_bnum=blocnum; }
   if (sortgfragsbyseq || sortgfragsboth)  sortnames=TRUE;
   if (sortnames)  sort_groups();
   if (grouptype==G_BLOCS)  set_bloc_names();
   /* Calculate npairseqs,noutseqs, etc. */
   set_groupvars(); }



/* Called by setgroups() if grouptype===G_SIMPLE */
/* Make_sqn(r) reads from strtag[] and readtag[], */
/*   which are assumed to be valid */

void build_simple_group(int usekeep)
 { int r,rn;  GROUPLIST gg;
   /* Set a grouplist[] with a single group */
   kill_groups();   /* Destroy any previous grouplist[] */
   blocsize=0;  memset(&gg,0,sizeof(gg));
   grouplist=cxalloc(2,sizeof(gg));  ngroups=1;
   gg.gname=strxdup("All");  gg.grname=strxdup("All");
   gg.dostat=TRUE;
   if (usekeep)
    { for (rn=r=0; r<nseqs; r++)  if (keeplist[r])  rn++;
      gg.sqn=cxalloc(rn+2,sizeof(SEQNAME));
      for (r=0; r<nseqs; r++)  if (keeplist[r])
        gg.sqn[gg.ns++] = make_sqn(r); }
   else
    { gg.sqn=cxalloc(nseqs+2,sizeof(SEQNAME));
      for (r=0; r<nseqs; r++)
        gg.sqn[gg.ns++] = make_sqn(r); }
   grouplist[0]=gg; }



/* Called by setgroups() if grouptype===G_BLOCS */
/* Make_sqn(r) reads from strtag[] and readtag[], */
/*   which are assumed to be valid */

void build_bloc_groups(int usekeep)
 { int i,j,k,r,rn;  SEQNAME *sq;
   /* Find how many sequences to use */
   if (usekeep) { for (rn=i=0; i<nseqs; i++)  if (keeplist[i])  rn++; }
   else rn=nseqs;
   /* Define blocnum from blocsize */
   if (blocsize<=1 || (blocnum=rn/blocsize)<=1
        || blocsize*blocnum!=rn)  set_bloc_err(rn);
   org_bsize=blocsize;  org_bnum=blocnum;

   /* Fill sequences from strtag[] */
   sq=cxalloc(rn+2,sizeof(SEQNAME));
   for (r=i=0; i<blocnum; i++)
     for (j=0; j<blocsize; j++, r++)
      { if (usekeep) { while(keeplist[r]==0)  r++; }
        if (homolflag)  sq[j*blocnum+i] = make_sqn(r);
        else  sq[i*blocsize+j] = make_sqn(r); }
   if (homolflag)
    { int bb=blocsize;  blocsize=blocnum;  blocnum=bb; }

   kill_groups();   /* Destroy any previous grouplist[] */
   ngroups=blocnum;
   grouplist=cxalloc(ngroups+2,sizeof(grouplist[0]));

   for (k=i=0; i<blocnum; i++)
    { GROUPLIST gg;  memset(&gg,0,sizeof(gg));
      gg.gname=gg.grname="***";  gg.dostat=TRUE;  gg.ns=blocsize;
      gg.sqn=cxalloc(blocsize+2, sizeof(SEQNAME));
      for (j=0; j<blocsize; j++)  gg.sqn[j]=sq[k++];
      grouplist[i]=gg; }
   free(sq); }



/* Used by build_groups(), which is called by setgroups() */
/*   if grouptype==G_GROUPS */
/* Here `sq.name' is quoted and truncated, */
/*   while `sq.rname' is as was read. */

static int warn_seq_not_found (SEQNAME sq)
 { int r;
   for (r=0; r<nseqs; r++)
     if (!striucmp(sq.rname,readtag[r]))
      { textprintf ("%s is PROSCRIBED by a -seq_skip "
           "or -seq_list command !!\n", sq.name);  return TRUE; }
   addstr(&warnlist, sq.name);
   return FALSE; }


static void say_no_strt (int nmiss, int npro, int g2)
 { if (nmiss>1)
     textprintf ("%d sequences (in grouplist) are NOT in %s%s\n",
       nmiss, seqfile,
       (npro>0) ?
         "\n  (or are proscribed by -seq_skip command(s))" : "");
   else
    { textprintf ("One sequence (in group %s) ", grouplist[g2].gname);
      if (npro>0)
        textprintf ("\n  is PROSCRIBED by a -seq_skip command.\n");
      else  textprintf ("is NOT in %s.\n", seqfile); }}



/* Called by setgroups() if grouptype==G_GROUPS */

static void build_groups(void)
 { int r,g, g2=-1;
   int num_miss=0, num_proscr=0;
   char *matched=mxalloc(nseqs+2);  /* Start with no matches */
   /* Fill grouplist[] from perm_group[], which was filled */
   /*   from command-line arguments in vsetopts.c */
   /* Deallocates any previous grouplist[].  */
   init_grouplist();
   /* Match names in  SEQNAME *sqn  with  readtag[] (strtag[])  */
   /*   to the extent possible. */
   /* Set orgoffs field, which is initialized at -1 */
   for (g=0; g<ngroups; g++)
    { int j,r, ns=grouplist[g].ns;
      SEQNAME *sqn=grouplist[g].sqn;
      for (j=0; j<ns; j++)
       { char *rname=sqn[j].rname;
         for (r=0; r<nseqs; r++)
           if (keeplist[r] && !striucmp(rname,readtag[r]))
             { sqn[j].orgoffs=r;  matched[r]=1;  break; }
         if (r>=nseqs)
          { num_miss++;  g2=g;
            if (warn_seq_not_found(sqn[j]))  num_proscr++; }}}
   /* If there are any sequences that cannot be matched, */
   /*   complain and then clear the warning list. */
   if (warnlist.n>0)
    { show_wlist();  reset_wlist(); }

   /* Condense any unmatched entries in grouplist[].sqn (orgoffs==-1) */
   if (num_miss>0)
    { say_no_strt(num_miss,num_proscr,g2);
      for (g=0; g<ngroups; g++)
       { int j,k, ns=grouplist[g].ns;
         SEQNAME *sqn=grouplist[g].sqn;
         for (k=j=0; j<ns; j++)
          { if (sqn[j].orgoffs<0)
             { free(sqn[j].name);  free(sqn[j].rname); }
            else { if (k<j)  sqn[k]=sqn[j];  k++; }}
         if (k<ns)
          { grouplist[g].ns=k;
            memset(sqn+k, 0, (ns-k)*sizeof(SEQNAME)); }}}

   /* Are there any unmatched sequences in strtag[]? */
   /* Note that keeplist[r]==0 prevents matches. */
   /* Remainder list is used only in show_seqnames() in vdump.c */
   kill_strlist(&seqs_notused);   /* To be safe */
   for (r=0; r<nseqs; r++)  if (matched[r]==0)
    { addstr(&seqs_notused,strtag[r]);  keeplist[r]=0; }

   /* Mark groups that are too small to compute statistics */
   for (g=0; g<ngroups; g++)
    { if (grouplist[g].ns<2)  grouplist[g].dostat=FALSE; }

   free(matched); }



static int seqn_sort (const void *p, const void *q)
 { int x;
   if ( (x=striucmp(((SEQNAME *)p)->rname, ((SEQNAME *)q)->rname))!=0)
           return x;
   if ( (x=strcmp(((SEQNAME *)p)->rname, ((SEQNAME *)q)->rname))!=0)
           return x;
   if ( (x=((SEQNAME *)p)->orgoffs - ((SEQNAME *)q)->orgoffs)!=0)
           return x;
   return 0; }


static int groupn_sort (const void *p, const void *q)
 { int x;  const GROUPLIST *p1=p, *q1=q;
   /* Put groups with statistics first */
   if ( (x=q1->dostat - p1->dostat)!=0)  return x;
   /* Now sort on name */
   if ( (x=striucmp(p1->grname, q1->grname))!=0)  return x;
   if ( (x=strcmp(p1->grname, q1->grname))!=0)  return x;
   if (p1->ns>0 && q1->ns>0
     && (x=strcmp(p1->sqn[0].rname,q1->sqn[0].rname))!=0)  return x;
   if ( (x=p1->ord - q1->ord)!=0)  return x;
   return 0; }


/* Sort grouplist[] by groupname then sequence names */

static void sort_groups(void)
 { int i;
   for (i=0; i<ngroups; i++)
    { grouplist[i].ord=i;
      qsort(grouplist[i].sqn, grouplist[i].ns,
         sizeof(SEQNAME), seqn_sort); }
   qsort(grouplist, ngroups,
      sizeof(GROUPLIST), groupn_sort); }


/* Set the bloc names equal to the name of */
/*   the first sequence in that block */

void set_bloc_names(void)
 { int i;
   for (i=0; i<ngroups; i++)
    { grouplist[i].gname = strxdup(grouplist[i].sqn[0].name);
      grouplist[i].grname = strxdup(grouplist[i].sqn[0].rname); }}


/* Fill  offs,first,last   fields  in grouplist[] */
/*  and calculate  npairseqs, noutseqs, and noutgrps */

void set_groupvars(void)
 { int i,j,r,x;
   npairseqs=noutseqs=noutgrps=mxtag=gxtag=0;
   for (r=i=0; i<ngroups; i++)
    { int ns=grouplist[i].ns;  SEQNAME *sqn=grouplist[i].sqn;
      if ( (x=strlen(grouplist[i].gname)) > gxtag)  gxtag=x;
      grouplist[i].first=r;
      for (j=0; j<ns; j++, r++)
        if ( (x=strlen(sqn[j].name)) > mxtag)  mxtag=x;
      grouplist[i].last=r;
      if (grouplist[i].dostat)
       { npairseqs += ((ns*(ns-1))/2);  noutgrps++;
         if (ns>2)  noutseqs+=ns;
         else if (ns==2)  noutseqs++; }}}




/* (***FFF) Fill the polysites[] array */

void read_polys(void)
 { char *rs, nbuff[88];
   int offset, nrsites, rnseqs=nseqs, needreorder=FALSE;
   /* UNLESS read_polys() is reading from sequences stored in a  */
   /*   seqbase[] array, DETECT if sites have to be reordered due to */
   /*   -seq_skip, -group, -sortname, or etc. commands. */
   /* Set needreorder==TRUE if reordering will be necessary. */
   if (reread_polymorphisms)  init_sqbase_read();
   else
    { int g,rn;
      for (rn=g=0; g<ngroups; g++)
       { int j, ns=grouplist[g].ns;  SEQNAME *sq=grouplist[g].sqn;
         for (j=0; j<ns; j++, rn++)
           if (sq[j].orgoffs!=rn)  needreorder=TRUE; }
      if (rn!=nseqs)  { needreorder=TRUE;  rnseqs=rn; }
      if (needreorder)  newsites=mxalloc(nseqs+2); }
   /* Else set parameters to read from seqbase[] */

   /* Reset (destroy previous) polysites[] and polysave[] arrays */
   init_poly_arrays();

   if (saveseqs && reread_polymorphisms==FALSE)
    { if (unaligned_data==FALSE)
       /* Read the rest of the sequence data into seqbase[] */
       { if (r_polyflag)  while(readpolysite(&offset)!=NULL)  ;
         else  while (readgroup()>=0)  ; }
      if (seqtype==AUTO)  seqtype=get_sqtype();
      init_sqbase_read(); }
   if (showsites)
    { textprintf("\nSite data:\n");  nrsites=0; }

   while ( (rs=readsite(&offset)) != NULL)
    { /* Are there any restrictions on the offset? */
      if ( (base_range.s || base_skip_range.s)
          && skip_range(offset))  continue;
      /* Should the bases from any sequences be dropped? */
      /* Should they be written in a different order? */
      if (needreorder)  rs=poly_rewrite(rs);
      /* If requested, show the sites */
      if (showsites)
        textprintf ("#%2d. Offs=%3d  Sites=%*.*s\n",
          ++nrsites, offset, rnseqs,rnseqs,rs);
      /* Is rs[] polymorphic and does it have any missing data? */
      /* Note that sites with all missing data can */
      /*   still be polymorphic. */
    { int i, mcnt=0, allmiss=TRUE, polyflag=FALSE;
      for (i=0; i<rnseqs; i++)
       { if (rs[i]=='U' && seqtype!=PROTEIN)
          { rs[i]='T';  numutconv++; }
         storebaseval(rs[i]);
         if (strchr(read_knownchars,rs[i]))  allmiss=FALSE;
         else  mcnt++;
         if (rs[i]!=rs[0])  polyflag=TRUE; }
      /* Update some counts. */
      if (allmiss)  ntotmiss_sites++;   /* All bases unrecognized */
      if (mcnt) nummiss_sites++;        /* Some bases unrecognized */
      else if (polyflag) numpoly_sites++;  /* Only if ALL bases recognized */
      else  nummono_sites++;            /* Only if no missing data */

      /* All sites reach here, including monomorphic sites */
      /* Trisite() needs sites with indels to avoid phase errors. */
      /* Dosite() needs sites with indels to manage runs properly. */
      /*   (Sites with one or more indels continue a TD_BLOCS run.) */
      if (seqtype==SILENT)  trisite(rs,rnseqs, offset, mcnt, polyflag);
      else  dosite (rs,rnseqs, offset, mcnt, polyflag); }}

   /* Display a warning if a partial codon is left */
   if (seqtype==SILENT)  trisite(NULL,rnseqs, offset,0,0);

   /* Free newsites[] and sitebuff[] */
   if (newsites) { free(newsites);  newsites=NULL; }
   if (sitebuff) { free(sitebuff);  sitebuff=NULL; }
   /* Find first_offset and last_offset, taking base_range */
   /*   and base_skip_range into account. */
   set_first_last();

   /* If TD_BLOCS, flush the `addrun()' queue for sites with indels */
   /* If TD_BLOCS and CIRCULAR, may need to stitch together the */
   /*   first and last polymorphisms */
   if (treat_dels==TD_BLOCS)
    { addrun(NULL,rnseqs, 0,0);
      save_stitch(rnseqs);
      if (circular)  maybe_stitch(rnseqs); }

   if (npolys<=2)  say_toofewpolys();

   /* If need be, rewrite strtag[] and seqbase[] and redefine nseqs */
   /*   if there are fewer sequences or if the sequences are */
   /*   in a different order */
   if (needreorder)
    { int ns=rewrite_stag();
      if (saveseqs)  ns=rewrite_sqbases();
      nseqs=ns; }

   if (reread_polymorphisms)
      printf ("%s re-read.\n", numsay(nbuff,npolys, "polymorphism",TRUE));

   if (showpolyflag)  make_poly_save(TRUE);
   if (lrfname) { free(lrfname);  lrfname=NULL; }
   if (lrsfname) { free(lrsfname);  lrsfname=NULL; }
       }  /* End of read_polys() */


/* Reorder a Seqfile column in `grouplist[]' order */
/* This assumes newsites!=NULL */
/* Bases are included from `NonCompute' sequences, */
/*   but not from `Remainder' (seqs_notused) sequences. */

static char *poly_rewrite(char *oldsites)
 { int r,g;
   for (r=g=0; g<ngroups; g++)
    { int j, ns=grouplist[g].ns;  SEQNAME *sq=grouplist[g].sqn;
      for (j=0; j<ns; j++)
        newsites[r++]=oldsites[sq[j].orgoffs]; }
   return newsites; }



/* Rewrite strtag[],readtag[] from grouplist[] and old strtag,readtag */

static int rewrite_stag(void)
 { int i,j,r, rn;  char **newtags, **newrdtags;
   for (rn=i=0; i<ngroups; i++)  rn += grouplist[i].ns;
   newtags=cxalloc(rn+2,sizeof(char *));
   newrdtags=cxalloc(rn+2,sizeof(char *));
   for (r=i=0; i<ngroups; i++)
    { int ns=grouplist[i].ns;  SEQNAME *sq=grouplist[i].sqn;
      for (j=0; j<ns; j++, r++)
       { int k=sq[j].orgoffs;
         newtags[r]=strxdup(strtag[k]);
         newrdtags[r]=strxdup(readtag[k]); }}
   destroy_stag();  strtag=newtags;  readtag=newrdtags;
   return rn; }


/* Rewrite seqbase[] from grouplist[] */
/* If r_polyflag==TRUE (listed polymorphism format), note that */
/*   `seqrpoff[]' needn't be rewritten since it has offset data only. */

static int rewrite_sqbases(void)
 { int i,j,r, rn;  char **newbases, *useseqs;
   for (rn=i=0; i<ngroups; i++)  rn += grouplist[i].ns;
   if (seqbase==NULL)  return rn;   /* To be safe */
   newbases=cxalloc(rn+2,sizeof(char *));
   useseqs=mxalloc(nseqs+2);       /* Initially 0 */
   for (r=i=0; i<ngroups; i++)
    { int ns=grouplist[i].ns;  SEQNAME *sq=grouplist[i].sqn;
      for (j=0; j<ns; j++)
       { newbases[r++]=seqbase[sq[j].orgoffs];
         useseqs[sq[j].orgoffs]=1; }}
   for (i=0; i<nseqs; i++)  if (useseqs[i]==0)  free(seqbase[i]);
   free(useseqs);  free(seqbase);  seqbase=newbases;
   return rn; }



void make_poly_save(int resize)
 { if (showpolyflag)
    { if (resize)
        polysave=rexxalloc(polysave,sizeof(POLYSITE), 0,npolys);
      memmove(polysave,polysites, npolys*sizeof(POLYSITE)); }}


/* Stitch together polymorphic blocks if the first */
/*   and last polymorphisms contain indels */

static void save_stitch(int nseqs)
 { if (npolys>1 && polysites[0].loca==1
        && havmis(polysites[0].site,nseqs)
        && polysites[npolys-1].locb==nbases
        && havmis(polysites[npolys-1].site,nseqs)
        && stitch_saved==FALSE)
    { first_psite=polysites[0];
      first_psite.site=strxdup(first_psite.site);
      last_psite=polysites[npolys-1];
      last_psite.site=strxdup(last_psite.site);
      stitch_saved=TRUE; }}



static void maybe_stitch(int nseqs)
 { if (npolys>2 && circular && is_stitched==FALSE
        && polysites[0].loca==1
        && polysites[npolys-1].locb==nbases)
    { char *fir=polysites[0].site, *las=polysites[npolys-1].site;
      if (havmis(fir,nseqs) && havmis(las,nseqs))
       { addstitch(fir,las,nseqs);  npolys--;  sitend--;
         if (polysites[0].locb>polysites[0].loca)  numdelblocs--;
         else numdelsing--;
         if (polysites[npolys].locb>polysites[npolys].loca)  numdelblocs--;
         else numdelsing--;
         numdelblocs++;  free(las);
         polysites[0].loca = polysites[npolys].loca - nbases;
         memset(&polysites[npolys],0,sizeof(polysites[0]));
         is_stitched=TRUE; }}
   if (npolys<=2)  say_toofewpolys(); }


void restitch(void)
 { if (circular)  maybe_stitch(nseqs);
   else if (stitch_saved && is_stitched)
    { free(polysites[0].site);
      polysites[0]=first_psite;  numdelblocs--;
      polysites[0].site=strxdup(polysites[0].site);
      if (first_psite.locb>first_psite.loca)  numdelblocs++;
      else numdelsing++;
      polysites[npolys]=last_psite;
      polysites[npolys].site=strxdup(polysites[npolys].site);
      if (last_psite.locb>last_psite.loca)  numdelblocs++;
      else numdelsing++;
      npolys++;  sitend++;  is_stitched=FALSE; }}




void run_comments(void)
 { textprintf("\n");
   show_features();
   show_seqnames();
   textprintf("\n");
   gscale_comment();

   show_restricts();
   show_basedata();

   if (saveseqs && showunal)
    { nl();  show_unalens(BPRINTF,BPRINTF); }}



void show_restricts(void)
 { if (base_range.s || base_skip_range.s)
    { char nbuff[88];  int len, n=1, nx=nbases;
      if (base_range.s) { n=base_range.n;  nx=base_range.nx; }
      len=sprintf(nbuff, "%d-%d", n,nx);
      if (base_skip_range.s)
        sprintf (nbuff+len, " EXCLUDING %d-%d",
          base_skip_range.n, base_skip_range.nx);
      textprintf ("BASE RANGE %s\n", nbuff); }
   if (include_monosites)
     textprintf("USING MONOMORPHIC AS WELL AS POLYMORPHIC %s.\n",
       (seqtype==SILENT) ? "CODON POSITIONS" : "SITES"); }



void set_kchars(void)
 { if (seqtype==PROTEIN)
    { read_knownchars=proteins;  knwhat=protwhat; }
   else if (use_iubnucls)
    { read_knownchars=total_nucls;  knwhat=tnucwhat; }
   else { read_knownchars=nucleotides;  knwhat=nucwhat; }
   if (seqtype==SILENT)
    { treat_dels=TD_SKIP;  setdna(mitcodes); }}


char *set_pchars(void)
 { if (seqtype==PROTEIN) return proteins;
   else if (use_iubnucls) return total_nucls;
   else return "TCAGU"; }



#define MX_CHVAL 256

static int baseval[MX_CHVAL+2], misover;

static void reset_baseval(void)
 { memset(baseval,0,sizeof(baseval));  misover=0; }


static void storebaseval (int c)
 { if (c<0 || c>=MX_CHVAL)  misover++;
   else  baseval[c]++; }


void show_basedata()
 { char *bleep, nbuff[99], cbuff[22];
   int i, nmiss, n=strlen(read_knownchars);
   TOKLIST clist,dlist, mvlist;
   memset(&clist,0,sizeof(clist));
   memset(&dlist,0,sizeof(dlist));
   memset(&mvlist,0,sizeof(mvlist));
   /* Collect recognized values in clist (if PROTEIN: also dlist) */
   for (i=0; i<n; i++)
    { int c=read_knownchars[i];
      sprintf(nbuff,"%c(%s)",c,commll(cbuff,baseval[c]));
      addtokl(&clist, nbuff, TOK_VOLATILE);
      /* Both read_knownchars[] and amtag[] are upper case */
      if (seqtype==PROTEIN)
       { char *str="???";  int j;
         /* Handle the two serine classes separately */
         if (c=='S')  str="Ser1";
         else if (c=='X')  str="Ser2";
         else for (j=0; j<64; j++) if (c==codn[j].amtag)
               { str=codn[j].aname;  break; }
         sprintf(nbuff,"%s(%s)",str,commll(cbuff,baseval[c]));
         addtokl(&dlist, nbuff, TOK_VOLATILE); }}
   /* Collect unrecognized values in mvlist */
   for (nmiss=i=0; i<MX_CHVAL; i++)
     if (strchr(read_knownchars,i)==NULL && baseval[i]>0)
      { sprintf(nbuff,"%c(%02XH)(n=%s)",i,i,commll(cbuff,baseval[i]));
        addtokl(&mvlist, nbuff, TOK_VOLATILE);  nmiss+=baseval[i]; }
   if (misover>0)
    { sprintf(nbuff,"Other(n=%s)",commll(cbuff,misover));
      addtokl(&mvlist, nbuff, TOK_VOLATILE);  nmiss+=misover; }
   /* Add total unrecognized values to clist (if PROTEIN: also dlist) */
   if (nmiss>0)
    { sprintf(nbuff,"Other(%s)",commll(cbuff,nmiss));
      addtokl(&clist, nbuff, TOK_VOLATILE);
      if (seqtype==PROTEIN)
       { sprintf(nbuff,"Other(%s)",commll(cbuff,nmiss));
         addtokl(&dlist, nbuff, TOK_VOLATILE); }}
   /* If ALL or SILENT, add the CG/AT ratio */
   if (seqtype==ALL || seqtype==SILENT)
    { long nat=baseval['A']+baseval['T'];
      long ncg=baseval['C']+baseval['G'];
      if (nat+ncg>0L)
       { sprintf(nbuff, "CG/TCAG ratio: %.2f%%", (100.0*ncg)/(nat+ncg));
         addtokl(&clist, nbuff, TOK_VOLATILE); }}

   /* Finall, display the recognized values... */
   sprintf(nbuff, "%c%s codes in data:  ",
      toupper(knwhat[0]), knwhat+1);
   ftoklistex(nbuff, &clist, 2,NULL,0, TOK_FREE);
   if (seqtype==PROTEIN)
    { tokl_string_sort(dlist);
      ftoklistex("In terms of 3-letter codes:  ",
        &dlist, 2,NULL,0, TOK_FREE); }

   if (numutconv>0)
     textprintf ("%s converted to T%s.\n",
       numsayc(nbuff,numutconv,"U(nucleotide)",TRUE),
       (numutconv==1)?"":"s");

   /* Show missing or unrecognized values */
   if (nmiss==0)
     textprintf ("No non-%s characters in data.\n", knwhat);
   else
    { sprintf(nbuff, "Non-%s characters in data:  ", knwhat);
      ftoklistex(nbuff, &mvlist, 2,NULL,0, TOK_FREE);
      textprintf ("%s with missing data or unrecognized characters.\n",
        numsay(nbuff, nummiss_sites, "site", TRUE));
      if (treat_dels==TD_BLOCS)
       { textprintf ("(These characters are treated as indels for the\n");
         textprintf ("   purpose of creating indel blocks.)\n");  }
      textprintf ("%s with all gap or unrecognized characters.\n",
        numsay(nbuff, ntotmiss_sites, "site", TRUE)); }

   /* Other descriptive data */
   textprintf ("%d polymorphic and %d monomorphic sites with no indels.\n",
     numpoly_sites, nummono_sites);
   if (seqtype==SILENT)
    { textprintf("%s with indels or missing data.\n",
        numsay(nbuff, nummiss_cods, "codon position", TRUE));
      textprintf("%s with no indels.\n",
        numsay(nbuff, numaapoly,
          "amino-acid polymorphic codon position", TRUE)); }
   else
    { textprintf ("%s with indels.\n",
        numsay(nbuff, numdelsing, "single polymorphic site", TRUE));
      textprintf ("%s with more than one site%s\n",
        numsay(nbuff, numdelblocs, "indel block", TRUE),
          (numdelblocs==0) ? "." : "");
      if (numdelblocs>0)
        textprintf ("  (each block is treated as "
             "a single polymorphism).\n"); }
   if (mitcodes)
     textprintf ("USING (MAMMALIAN) MITOCHONDRIAL GENETIC CODES.\n");

   switch(seqtype) {
     default:  bleep="polymorphism";  break;
     case SILENT:  bleep="silent codon polymorphism";  break;
     case PROTEIN: bleep="amino-acid polymorphism";  break; }
   textprintf ("%s%s.\n",
      numsay(nbuff,npolys, bleep,TRUE),
        (numsims>0) ? " to be permuted" : "");

   if (nummiss_sites)
    { bleep=NULL;
      if (seqtype==SILENT)
          bleep="CODON POSITIONS with unknown "
             "or missing data are excluded";
      else if (treat_dels==TD_SKIP)
          bleep="SITES WITH UNKNOWN OR MISSING DATA are excluded";
      else if (treat_dels==TD_SITES)
          bleep="EACH SITE WITH DELETED DATA is "
             "counted as a single partition";
      if (bleep)  textprintf ("%s.\n", bleep); }}



void show_codon_stats(void)
 { int k, oldtag;  AACODE *cc;
   int  sumsil = num0sil + num2sil + num3sil + num4sil + numisil;
   TOKLIST alist;  memset(&alist,0,sizeof(alist));
   textprintf ("The silent sites at %d amino-acid monomorphic "
      "codon positions are\n", sumsil);
   textprintf ("   0-degen (%d)  2-degen (%d)  3-degen (%d)"
      "  4-degen (%d)  irreg (%d)\n",
        num0sil, num2sil, num3sil, num4sil, numisil);
   /* Note: When called from vdump.c, each textprintf() line */
   /*   is prepended by `# '  */
   textprintf (
     "The silent polymorphic sites from these codon positions are\n");
   textprintf ("   0-degen (0)  2-degen (%d)  3-degen (%d)"
     "  4-degen (%d)  irreg (%d)\n",
        num2poly, num3poly, num4poly, numipoly);
   oldtag=-1;
   for (cc=codn, k=0; k<64; k++, cc++)
     if (cc->gflag && !strcmp(cc->signature,"003"))
       { if (cc->amtag!=oldtag)
          { addtokl(&alist, getstr("(%3.3s)",cc->aname), TOK_FREE);
            oldtag=cc->amtag; }
         addtokl(&alist, getstr("%s(%d)",codonval(k),cc->codoncnt),
            TOK_FREE); }
   ftoklistex("Codon usage for 3-fold degenerate codons: ",
     &alist, 1,NULL,0, TOK_FREE); }




/* (***GGG) */
/* Site-oriented reading routines */

/* Return a site from <Seqfile> */
/* For rpolysite, interleaved, or non interleaved formats */

static int nleft, nrun, k0ff, rgroup, roffset;

void init_readvars(void)
 { int i;
   destroy_stag();
   reset_sqf_vals();  reset_baseval();
   reset_textprintf();  dumpx_close("???");  reset_wlist();
   dumpflag=tothavcirc=FALSE;
   kill_strlist(&seqs_notused);
   kill_strlist(&seqfcomments);
   if (configcomments.n>0)  seqfcomments=configcomments;
   save_nxtok=NULL;  save_ispun=0;
   /* Reset buffers */
   reset_readbuff();
   if (seqrpoff) { free(seqrpoff);  seqrpoff=NULL; }
   kill_nstr(&seqbase,nseqs);
   if (sitebuff) { free(sitebuff);  sitebuff=NULL; }
   if (newsites) { free(newsites);  newsites=NULL; }
   kill_linebuffs();
   if (keeplist) { free(keeplist);  keeplist=NULL; }

   nseqs=file_nseqs=blocnum=0;  nbases=-1;
   npairseqs=noutseqs=noutgrps=numutconv=0;
   nrun=nread=nxread=nleft=k0ff=roffset=maxsqlinelen=0;
   rgroup=org_bsize=org_bnum=allctrlnum=0;
   numsil=num0sil=num2sil=num3sil=num4sil=numisil=0;
   numpoly_sites=nummono_sites=nummiss_sites=ntotmiss_sites=0;
   num2poly=num3poly=num4poly=numipoly=0;
   tri_phase=mxtag=gxtag=numdelblocs=maxdelblocs=0;
   for (i=0; i<4; i++)  if (tri_cache[i])
    { free(tri_cache[i]);  tri_cache[i]=NULL; }

   reread_polymorphisms=tags_set=nexus_out=FALSE;  fformat=F_NONE;
   r_polyflag=unaligned_data=FALSE;
   is_stitched=stitch_saved=have_silent_multsites=FALSE;
   next_rpoly=NULL;
   if (first_psite.site)  free(first_psite.site);
   memset(&first_psite,0,sizeof(first_psite));
   if (last_psite.site)  free(last_psite.site);
   memset(&last_psite,0,sizeof(last_psite));
   init_poly_arrays();
   #ifdef VSHOW
   /* Reset re-entrant command-line settings */
   set_base_vars();
   #endif
         }


static void init_sqbase_read(void)
 { unaligned_data=TRUE;  nrun=roffset=0;
   if (seqbase==NULL)
     ErrHandler("INITIALIZATION to read from stored sequences\n"
       "  when there are no stored sequences!\n");
   if (sitebuff==NULL)  sitebuff=mxalloc(nseqs+3);
   set_kchars(); }


static void init_poly_arrays(void)
 { if (polysites)
    { int i=npolys;  while (--i>=0)  free(polysites[i].site);
      free(polysites);  polysites=NULL; }
   sitend=NULL;  npolys=numpxend=0;
   if (polysave) { free(polysave);  polysave=NULL; }}


/* NOTE: Sequence files in three NEXUS formats and two ASF formats */
/*   can be read.  See the note before OpenSeqFile(). */

static char *readsite(int *offptr)
 { int i;
   /* Init_sq_read() sets  unaligned_data=TRUE */
   if (unaligned_data)    /* Read site from seqbase[] */
    { if (nrun<nread)
       { if (r_polyflag) *offptr = seqrpoff[roffset++];
         else  *offptr = ++roffset;
         for (i=0; i<nseqs; i++)
           sitebuff[i]=seqbase[i][nrun];  nrun++;
         return sitebuff; }
      /* No more bases to read: Destroy seqbase[] if saveseqs==FALSE. */
      if (saveseqs==FALSE)  kill_nstr(&seqbase,nseqs);
      return NULL; }
   else if (r_polyflag)  return readpolysite(offptr);
   else /* All other cases use readgroup() */
    { *offptr = ++roffset;
      /* The `while' skips empty groups, and also debits `nleft' */
      /*   for the current site */
      while (--nleft < 0)
       { if (readgroup()<0)  { *offptr=-1;  return NULL; }}
      for (i=0; i<nseqs; i++)
        sitebuff[i]=rdrow[i][k0ff];  k0ff++;
      return sitebuff; }}



/* Read a group of bases from <Seqfile>: */

static void set_slabels(int nseqs)
 { char sb[44];  int i;
   /* To be safe */
   if (strtag==NULL)  strtag=cxalloc(nseqs+2,sizeof(char *));
   if (readtag==NULL)  readtag=cxalloc(nseqs+2,sizeof(char *));
   for (i=0; i<nseqs; i++)
    { sprintf(sb,"S%d",i+1);
      strtag[i]=strxdup(sb);
      readtag[i]=strxdup(sb); }}


static void init_rg_arrays(int nseqs)
 { if (nseqs>0)
    { init_stag(nseqs);
      if (havlabels==FALSE && tags_set==FALSE)
       { set_slabels(nseqs);  tags_set=TRUE; }
      if (saveseqs)  seqbase=cxalloc(nseqs+3,sizeof(char *));
      else  rdrow=cxalloc(nseqs+3,sizeof(char *)); }
   /* To be safe */
   nread=nxread=rgroup=linxlen=matchxlen=0; }


static int is_blanklike(char *rbuff)
 { int i=0;
   while (i<3 && isspace(rbuff[i]))  i++;
   return (i>=3 || rbuff[i]==0); }


/* Batch length for sequences if saveseqs==TRUE: */

#define SEQBBAT 5000

static int readgroup(void)
 { char c, *rs, *rbuff;
   int i,j,k, nseqx, linelength;
   int matchlen, matchflag, copy_match;
   int no_nseqs = (nseqs==0);
   int old_nxread=nxread, old_linxlen=linxlen;
   char q1b[MAXNAMEBUFF];

   /* Mathchlen and matchflag are used only when there is */
   /*   an explicit consensus line (with tag=="!") . */
   matchlen=nseqx=0;  matchflag=FALSE;
   rgroup++;   /* Group number */

   if (fformat==F_NEXUS &&
     (nexus_out || (interleave==FALSE && rgroup>1)))  goto allread;

   /* In CLUSTAL format, must manually skip initially-blank */
   /*   lines before a group */
   if (fformat==F_CLUSTAL)
    { do { if ( (rbuff=lineread(FALSE))==NULL)  break; }
      while (is_blanklike(rbuff)) ;
      pushline(rbuff); }

   /* Read the sequences in group #rgroup: */
   for (i=0; i<nseqs || no_nseqs; i++)  {

     /* Read the next line, unless end of group */
     if ( (rbuff=lineread(FALSE))==NULL
         || nexus_out || (fformat==F_NEXUS && rbuff[0]==';')
         || (fformat==F_CLUSTAL && is_blanklike(rbuff)) )
       { if (i==0)  goto allread;
         if (no_nseqs)  break;  /* End of first group */
         ErrHandler("DID NOT FIND sequence `%s' (sequence #%d)\n"
           "  in group #%d of `%s'\n%s",
             (strtag[i]) ? strtag[i] : "???",
                 i+1, rgroup, lrsfname, Show_lbuff()); }

     /* In this format, there may still be a few initial spaces */
     if (fformat==F_CLUSTAL)  while(isspace(*rbuff))  rbuff++;

     /* If the sequence tag is "!", then this is a consensus line */
     /*   and not a sequence data line. */

     if ( (fformat==F_ASF || fformat==F_NEXUS)
        && rbuff[0]=='!' && (isspace(rbuff[1]) || rbuff[1]==0))
      { if (i>0)  ErrHandler(
          "THE CONSENSUS LINE `!' (if one exists)\n"
          "  MUST BE THE FIRST LINE of a group!\n%s", Show_lbuff());
        if (rbuff[1]==0)  { i=-1;  continue; }
        if (matchchar==0)  ErrHandler(
          "CONSENSUS LINE `!' FOUND BUT MATCHCHAR IS NOT DEFINED!!\n%s",
              Show_lbuff());
        /* Copy the consensus line to matchbuff[] */
        makeuc(rbuff+=2);   /* Convert to upper case */
        k=j=0;
        while ( (c=rbuff[j++])!=0)  if (isbasechar(c))
         { if (c==matchchar)
             ErrHandler("YOU CANNOT HAVE THE MATCHCHAR `%c' "
                 "IN A CONSENSUS LINE!\n%s", matchchar, Show_lbuff());
           if (k>=matchxlen)
             matchbuff=rexxalloc(matchbuff,1, k,matchxlen+=LINEBAT);
           matchbuff[k++]=c; }
        if (k>0) { matchlen=k;  matchbuff[k]=0;  matchflag=TRUE; }
        i=-1;  continue; }   /* Re-read line #1 */

     /* Not a consensus line */
     rs=rbuff;
     if (no_nseqs && i>=nseqx)
      { nseqx+=20;
        strtag=rexxalloc(strtag,sizeof(char *), i,nseqx);
        readtag=rexxalloc(readtag,sizeof(char *), i,nseqx);
        readshowtag=rexxalloc(readshowtag,sizeof(char *), i,nseqx);
        if (saveseqs)
          seqbase=rexxalloc(seqbase,sizeof(char *), i,nseqx);
        else  rdrow=rexxalloc(rdrow,sizeof(char *), i,nseqx); }

     /* If there are labels, find and separate out the sequence name */
     if (havlabels)
      { int openflag;  char *cbuff=NULL, *sname=rbuff;
        char *p, p1buff[33], p2buff[33];
        /* In PHYLIP format, the first 10 characters of the line */
        /*   are reserved for the sequence name, which may include */
        /*   unquoted blanks.  */
        if (phylip)
         { sname=memmove(p1buff,rbuff,10);  p1buff[10]=0;
           p=p1buff+strlen(p1buff);
           while(p>p1buff && isspace(*--p))  *p=0;
           for (p=p1buff; *p!=0; p++)
            if (isspace(*p)) /* Embedded spaces */
             { if (use_underscores)  unspace_str(sname);
               else sname=doquote(p2buff,p1buff);  break; }
           rs = (strlen(rbuff)<=10) ? NULL : rbuff+10; }
        else
          parse_unquote(rbuff,&rs,&openflag);
        /* Allow for a mildly corrupted file with a line break */
        /*   immediately after the sequence name */
        /* Since this may be common, complain only if the line ends */
        /*   inside open quotation marks, and make that fatal. */
        if (rs==NULL)
         { if (openflag)
             ErrHandler("LINE END WITHIN OPEN QUOTES INSIDE"
               " A SEQUENCE NAME in aligned data !!\n"
               "Line #%d, sequence name so far: `%s'", linenum,
               doquote(q1b,sname));
           /* Save the sequence name and read the next line */
           sname=cbuff=strxdup(sname);
           if ( (rs=rbuff=lineread(FALSE))==NULL)
             ErrHandler("FILE ENDED just after reading "
               "sequence name `%s' !!", doquote(q1b,sname)); }

        /* Read or check the tag string, if there are tag strings */
        if (tags_set==FALSE)  /* First group ? */
         { if (strtag[i]!=NULL)
             ErrHandler(
               "*** INTERNAL ERROR: Tag string `%s' already defined\n"
               "in group #%d for sequence #%d of `%s'\n%s",
                  strtag[i], rgroup, i+1, lrsfname, Show_lbuff());
           strtag[i]=gettag(sname, &readtag[i], &readshowtag[i], 'S'); }
        else if (striucmp(sname,readtag[i]))  /* Succeeding groups ? */
          /* Sequence tag name mismatch */
          /* Readshowtag[i] is truncated */
          ErrHandler(
            "EXPECTING SEQUENCE `%s' (sequence #%d) in group %d of `%s':\n"
            "  Expected `%s' for tag, found `%s'\n%s",
             strtag[i], i+1, rgroup, lrsfname, readshowtag[i],
             doquote(q1b,sname), Show_lbuff());
        /* In case `sname' was allocated and not in the read buffer */
        if (cbuff!=NULL)  free(cbuff); }

     /* Read the data in line #i and group #rgroup: */
     /* First, capitalize the sequence data after the sequence name. */
     makeuc(rs);  k=0;
     /* Will this line also be copied to matchbuff[] ? */
     copy_match = (i==0 && matchchar!=0 && matchflag==FALSE);
  restart_short_line:
     while ( (c=*rs++)!=0) if (isbasechar(c))
      { if (fformat==F_NEXUS && (c==',' || c==';'))
         { if (c==';')  nexus_out=TRUE;  break; }
        /* If line#1 will be the consensus line, copy c to matchbuff */
        if (copy_match)
         { if (c==matchchar)
             ErrHandler("YOU CANNOT HAVE THE MATCHCHAR '%c' IN "
                     "THE FIRST LINE OF A GROUP\n"
               "  unless there is a consensus line !!\n%s",
                    matchchar, Show_lbuff());
           if (k>=matchxlen)
             matchbuff=rexxalloc(matchbuff,1, k,matchxlen+=LINEBAT);
           matchbuff[k]=c; }
        /* Is this character beyond the end of the first line? */
        /* If so, the file may be corrupted and `c' may be the */
        /*   first character of the next sequence name. */
        if (i>0 && k>=linelength)
         { textprintf (
        "WARNING: Data is TOO LONG for sequence #%d `%s' (line #%d)\n"
        "Introducing a line break and continuing.\n",
              i+1, readshowtag[i], linenum);
           pushline(--rs);  break; }
        /* Is c a matchchar? */
        /* If i>0 && matchflag==FALSE, matchbuff has length linelength */
        if (c==matchchar)  /* Either i>0 or matchflag==TRUE */
         { if (matchflag && k>=matchlen)
             ErrHandler("THE MATCHCHAR '%c' OCCURS BEYOND THE END of\n"
             "  an existing consensus line!\n%s",
                    matchchar, Show_lbuff());
           c=matchbuff[k]; }
        /* Finally, save the character */
        if (k>=linxlen)   /* This implies line #1 */
         { linxlen+=LINEBAT;
           seqline=rexxalloc(seqline,1, k,linxlen); }
        seqline[k++]=c; }

     /* Is this line SHORTER than the first line? */
     /* Perhaps it was cut by a word processor or mailing program. */
     if (c==0 && i>0 && k<linelength)
      { textprintf (
     "WARNING: Data is TOO SHORT for sequence #%d `%s' (line #%d)\n"
     "Reading the next line for the rest of the bases.\n",
           i+1, readshowtag[i], linenum);
        if ( (rs=rbuff=lineread(FALSE))==NULL)
          ErrHandler("FILE ENDED after a short line for "
            "sequence `%s' (line #%d) !!", strtag[i], linenum);
        makeuc(rs);
        goto  restart_short_line; }

     /* End of line #i in group #rgroup */
     if (i==0)  linelength=k;    /* Set linelength for this group */
     else if (k!=linelength)
       ErrHandler(
       "UNEQUAL LINE LENGTHS in group #%d of `%s'!!!\n"
       "  Sequence #1: `%s': %d bases.\n"
       "  Sequence #%d: `%s': %d bases%s.\n%s",
           rgroup, lrsfname, readshowtag[0], linelength,
             i+1, readshowtag[i], k,
               (k<linelength) ? "" : " (at least)", Show_lbuff());

     /* Now copy the data line somewhere */
     seqline[k]=0;
     if (saveseqs)
      { if (nread+linelength>=old_nxread)
         { if (nxread==old_nxread)
             nxread += ( (linelength<SEQBBAT) ? SEQBBAT : linelength+10 );
           seqbase[i]=rexxalloc(seqbase[i],1, nread,nxread); }
        memmove(&seqbase[i][nread], seqline, linelength+1); }
     else  /* Keep rdrow[i] the same length as seqline[] */
      { if (linxlen>old_linxlen)
          rdrow[i]=rexxalloc(rdrow[i],1, 0,linxlen);
        memmove(rdrow[i], seqline, linelength+1); }}

/* (Above) pick up another sequence */
/* (Below) have read an entire group, so tidy up and return */

   k0ff=0;  nread+=linelength;  tags_set=TRUE;
   /* Define nseqs and file_nseqs if CLUSTAL format */
   if (no_nseqs)  file_nseqs=nseqs=i;
   /* Turn off labels if PHYLIP INTERLEAVED format */
   if (fformat==F_PHYLIP)  havlabels=FALSE;
   return nleft=linelength;

allread:
  if (nbases != nread)
   { if (nbases < 0)
       textprintf("%d sites read from each of %d sequences in `%s'.\n",
          nread, nseqs, lrfname);
     else  textprintf("DISCREPENCY:  %d bases promised in `%s',"
            " found (and using) %d.\n",
               nbases, lrfname, nread);
     nbases=nread; }
  if (saveseqs)
   { for (i=0; i<nseqs; i++)
       seqbase[i]=rexalloc(seqbase[i],nread+5); }
  CloseSeqFile();  err_clear=FALSE;  err_clear_funp=NULL;
  return  RET_ERROR; }



static SEQTYPE get_sqtype(void)
 { int i,j, nnuc=0,ntot=0;
   if (seqbase==NULL)
     ErrHandler("Get_sqtype: No saved sequences !!");
   for (i=0; i<nseqs; i++)
    { char *sb=seqbase[i];
      for (j=0; j<nread; j++)  if (isupper(sb[j]))
       { if (strchr("TCAGUN",sb[j]))  nnuc++;  ntot++; }}
   if (ntot==0)
    { WarnHandler("Warning: No recognized bases !!");
      return PROTEIN; }
   else if (nnuc>=(int)(0.85*ntot))  return ALL;
   else return PROTEIN; }




static char *getw_rpoly(int *endptr)
 { char *rs, *nextp;  *endptr=FALSE;
   if (next_rpoly==NULL
     && (next_rpoly=lineread(FALSE))==NULL)  return NULL;
   rs=parse_unquote(next_rpoly,&nextp,endptr);
   next_rpoly=nextp;
   return rs; }



static char *readpolysite(int *offptr)
 { char c, *rp, *p, *q, *site, q1b[MAXNAMEBUFF];
   int i, offset, openqflag;

   /* Read data for the next site */
   if ( (rp=getw_rpoly(&openqflag))==NULL || openqflag)  goto allread;

   /* Read the offset */
   if ( (offset = atoi(rp))<=0)
     ErrHandler("UNKNOWN OR OUT-OF-RANGE OFFSET AT %s %s\n%s",
          rp, ExpPolyForm, Show_lbuff());
   if (nbases>=0 && offset > nbases)
     ErrHandler("%d BASES CLAIMED, BUT OFFSET %d (%s) FOUND %s\n%s",
          nbases, offset, doquote(q1b,rp), ExpPolyForm, Show_lbuff());
   if (offset <= roffset)
     ErrHandler("OUT-OF-ORDER OR REPEATED OFFSET %s\n"
       "Last offset: %d   This offset: %d\n%s",
            ExpPolyForm, roffset, offset, Show_lbuff());

   /* Now read the bases */
   if ( (site=getw_rpoly(&openqflag))==NULL || openqflag)
    { if (openqflag) goto allread;
      ErrHandler("NO SITES FOUND for offset %d %s\n%s",
        offset, ExpPolyForm, Show_lbuff()); }

   /* Rewrite `site' without digits and parentheses */
   for (q=p=site; *p; p++)
     if (isbasechar(*p)) { if (q<p) *q=*p;  q++; }  *q=0;
   if (strlen(site)!=nseqs)
     ErrHandler("POLYMORPHISM AT %d DOES NOT HAVE %d BASES %s\n%s",
        offset, nseqs, ExpPolyForm, Show_lbuff());
   /* `Matchchar' characters match the first base */
   if ( (c=*site)==matchchar)
     ErrHandler("ILLEGAL INITIAL CHARACTER `%c' %s\n%s",
        matchchar, ExpPolyForm, Show_lbuff());
   for (p=site+1; *p; p++)  if (*p==matchchar) *p = c;

   *offptr = roffset = offset;    /* Return the site */
   makeuc(site);
   if (saveseqs)
    { if (nread>=nxread)
       { nxread+=SEQBBAT;
         for (i=0; i<nseqs; i++)
           seqbase[i]=rexxalloc(seqbase[i],1, nread,nxread);
         seqrpoff=rexxalloc(seqrpoff,sizeof(int), nread,nxread); }
      for (i=0; i<nseqs; i++)  seqbase[i][nread]=site[i];
      seqrpoff[nread++]=offset; }

   return site;

allread:
   if (openqflag)
     ErrHandler("LINE ENDS IN OPEN QUOTES (site offset=%d, %s)!!\n%s",
       offset, ExpPolyForm, Show_lbuff());
   if (nbases < 0)  nbases = roffset;  /* Best guess */
   CloseSeqFile();  err_clear=FALSE;  err_clear_funp=NULL;
   return  NULL; }



static void init_stag(int nseqs)
 { /* Strtag[] might already be filled */
   /*   from a NEXUS taxlabels statement, */
   /*   and nseqs may not yet be known in CLUSTAL format */
   if (nseqs>0 && strtag==NULL)
    { strtag = cxalloc(nseqs+2,sizeof(char *));
      readtag = cxalloc(nseqs+2,sizeof(char *));
      readshowtag = cxalloc(nseqs+2,sizeof(char *)); }}


void destroy_stag(void)
 { kill_nstr(&strtag,nseqs);
   kill_nstr(&readtag,nseqs);
   kill_nstr(&readshowtag,nseqs);
   kill_nstr(&strshtag,nseqs); }



/* Dosite(), trisite(), and friends */

/* If `site' is polymorphic, add it to the array `polysites'.
   If `site' is monomorphic, skip it unless use_mono...=TRUE
   If `site' has missing or unrecognized data:
     if seqtype==SILENT, `trisite()' will throw out that codon position
     if seqtype!=SILENT,
        if treat_dels==TD_SKIP, skip that site.
        if treat_dels==TD_BLOCS, treat a run of sites, each with some
           unrecognized data, as a single partition for `polysites'
        if treat_dels==TD_SITES, treat each site with missing data
           as a single partition for `polysites'

          */


/* Dosite(): Called unless seqtype==SILENT */

/* If treat_dels==TD_BLOCS: */
/*   If a site has missing data, add it to the queue and return.  */
/*   If there is no missing data, first flush any possible queue */
/*   If offset!=lastoffset+1, the queue is flushed within addrun() */

/* If treat_dels==TD_SKIP: */
/*   If a site has missing data, then return without storing the site */

/* If the site is polymorphic (or monomorphic and use_mono...=TRUE), */
/*   then store it otherwise */

/* Polymorphic sites with all missing data are still polymorphic. */
/*   Note that a monomorphic site can have all missing data. */

static void dosite (char *site,int nseqs, int offset, int mcnt, int polyflag)
 { if (treat_dels==TD_BLOCS)
    { if (mcnt)
        { addrun(site,nseqs, offset,polyflag);  return; }
      /* No missing sites, so flush the queue, then store the site */
      /*   as a single polymorphism. */
      /* Beyond this point, no missing data if TD_BLOCS */
      addrun(NULL,nseqs, 0,0); }
   if (mcnt>0 && polyflag)  numdelsing++;
   if ( (polyflag || include_monosites)
     && (mcnt==0 || treat_dels==TD_SITES) )
       addpolysite(site,nseqs, offset,offset,-1); }



/* Called ONLY if seqtype==SILENT */

/* Given a stream of sites in `feedin', store only polymorphisms
     from amino-acid monomorphic codon positions.
     This requires a look-ahead feature.
     This function is codon-oriented, not site-oriented */

/* If include_monosites=TRUE, then */
/*   monomorphic codon positions are also stored */

/* NOTE: The most appropriate way to handle silent-site polymorphisms
     is by considering codons, not sites.
   The code below accomplishes this by storing polymorphic sites
     for regular silent codon positions and by using ``addstitch()''
     to stitch together site polymorphisms at silent codon positions
     that have more than one silent polymorphic site.  */

static void trisite (char *feedin,int nseqs, int offset,
      int mcnt, int polyflag)
 { char *q1, *q2, *q3;  int i, classflag, amino_poly;
   static int mscnt[3], polyf[3];
   AACODE *aa;
   if (tri_cache[0]==NULL)
    { for(i=0; i<3; i++)  tri_cache[i] = mxalloc(nseqs+2); }
   if (feedin==NULL)
    { if (tri_phase>0)  textprintf ("THE FILE `%s' ENDED "
        "with a partial codon left!\n",  lrfname);  return; }
   memmove (tri_cache[tri_phase], feedin, nseqs);
   mscnt[tri_phase]=mcnt;
   polyf[tri_phase]=(polyflag) ? 1 : 0;
   if (++tri_phase < 3)  return;  tri_phase = 0;   /* Set for next codon */

   /* `trisite()' has now collected a complete codon position, */
   /*    and must now figure out what to do with it: */
   /* The first step is to exit if the codon position */
   /*    has any missing data. */

   if (mscnt[0] || mscnt[1] || mscnt[2])  { nummiss_cods++;  return; }
   /* Next, find the amino acids for those codons */
   q1=tri_cache[0];  q2=tri_cache[1];  q3=tri_cache[2];
   amino_poly=FALSE;
   for (i=0; i<nseqs; i++)
    { static char amino, codon[4];   /* Collect a codon in codon[] */
      codon[0] = *q1++;  codon[1] = *q2++;  codon[2] = *q3++;
      (aa = getcode(codon))->codoncnt++;
      /* This should not happen */
      if (aa->gflag==2)  { nummiss_cods++;  return; }
      if (i==0)  amino = aa->amtag;
      else if (aa->amtag != amino)  amino_poly = TRUE; }

   /* Ignore amino-acid polymorphic positions */
   if (amino_poly)  { numaapoly++;  return; }
   numsil++;
   if (aa->gflag)   /* Set classflag */
     { classflag = aa->signature[2] - '0';
       if (d3flag && classflag==3)  classflag = 2;
       switch(classflag) {
         case 0:  num0sil++;  break;
         case 2:  num2sil++;  break;
         case 3:  num3sil++;  break;
         case 4:  num4sil++;  break; }}
   else  { classflag = -1;  numisil++; }

   /* If the codon position is polymorphic, store one polymorphic site. */

   /* If an irregular codon position has multiple polymorphisms, */
   /*   stitching together a single codon-wise polymorphism */
   /*   is more conservative against codon bias. */

   /* The following assumes that polyf[i]==0 or 1 */
   /* Site these are codon polymorphisms, the offsets should be */
   /*   for the codon position. */

   if (polyf[0] || polyf[1] || polyf[2] || include_monosites)
    { int ic=2;  /* Default polymorphic site is 3rd position */
      if (polyf[0]+polyf[1]+polyf[2] >= 2)
       { have_silent_multsites=TRUE;
         if (polyf[0])  addstitch(tri_cache[2], tri_cache[0], nseqs);
         if (polyf[1])  addstitch(tri_cache[2], tri_cache[1], nseqs); }
      /* Else exactly one site (or none) is polymorphic */
      else if (polyf[1])  ic=1;
      else if (polyf[0])  ic=0;
      addpolysite(tri_cache[ic], nseqs, offset-2,offset, classflag); }}




static int val(int c)
 { switch (toupper(c))
    { case 'T': case 'U':  return 0;
      case 'C':  return 1;
      case 'A':  return 2;
      case 'G':  return 3; }
   return -100; }


static AACODE *getcode(char *p)
 { int i, off=0;
   for (i=0; i<3; i++)  off = 4*off + val(*p++);
   return  codn + ( (off >= 0)  ?  off : 64); }


char *codonval(int i)
 { static char cbuff[5];  int j;
   for (j=3; --j>=0; )
    { cbuff[j] = nucleotides[i&3];  i/=4; }
   return cbuff; }


void setdna(int mit)
 { memmove(codn, (mit) ? mtcodn : rcodn, sizeof(codn)); }


/* Add a polymorphic site to  polysites[] */

/* This is the only function that adds to polysites[] */

static void addpolysite (char *site,int nseqs,
        int firstoff, int lastoff, int classflag)
 { if (simpflag || nosimpflag)  /* Screen for simple/multiple polys */
     { char fc=site[0], sc;
       int i, fcnt=1, scnt=0, multflag=FALSE;
       for (i=1; i<nseqs; i++)
        { if (site[i]==fc)  fcnt++;
          else if (scnt==0)  { sc=site[i];  scnt=1; }
          else if (site[i]==sc)  scnt++;
          else  { multflag=TRUE;  break; }}
       if (fcnt < nseqs-1 && scnt < nseqs-1)  multflag=TRUE;
       if (multflag) { if (simpflag)  return; }
       else if (nosimpflag)  return; }

   /* Store the site */
   if (npolys >= numpxend)
    { numpxend+=BLOCLEN;
      if (sizeof(int)<=2 && (long)(numpxend+2)*sizeof(POLYSITE)>65535L)
        ErrHandler( "*** Polymorphism array overflow!\n"
         "You are limited to data with <= %ld polymorphisms\n"
         "  if your computer program uses 16-bit integers and arrays.\n"
         "\nUSE A 32-bit VERSION OF THIS PROGRAM!!!!!\n",
             65535L/sizeof(POLYSITE) - 2);
      polysites = rexxalloc(polysites,sizeof(POLYSITE), npolys,numpxend+2);
      sitend = polysites+npolys; }

   sitend->loca=firstoff;  sitend->locb=lastoff;
   switch (sitend->degen = classflag)  {
     default:  /* Unrecognized classflag; treat as irregular */
               sitend->degen = -1;  /* Now fall thru to case -1 */
     case -1:  numipoly++;  break;
     case  2:  num2poly++;  break;
     case  3:  num3poly++;  break;
     case  4:  num4poly++;  break; }

   memmove ( (sitend++)->site = mxalloc(nseqs+1), site, nseqs);
   npolys++; }



/* Combine runs of sites containing missing (or unrecognized)
     characters as a single partition.
   Only called if  treat_dels==TD_BLOCS
   A possible run is `flushed' to  addpolysite() by  addrun(NULL,.....)
   when
     (i)   A site without missing bases
              is encountered in  dosite()
     (ii)  The offset jumps by 2 or more
             (the sites between are assumed monomorphic)
     (iii) All sites have been read

   A site with any unrecognized characters adds to the queue.
   If a site is monomorphic for an unrecognized character,
     an existing run is extended without making the partition finer.
   Such a site does not begin a run.

*/

static void addrun (char *site,int nseqs, int offset, int polyflag)
 { static int multisite, firstoff, lastoff, haverun;
   static char *state;  static int nxseqs;
   if (nseqs>nxseqs)
      state=rexxalloc(state,1, 0,nxseqs=nseqs+4);
   if (site==NULL)
    { if (haverun)        /* Flush and reset */
       { addpolysite(state,nseqs, firstoff,lastoff,-1);
         /* Update some counts */
         if (multisite)
          { int x; numdelblocs++;
            if ( (x=lastoff-firstoff+1)>maxdelblocs)  maxdelblocs=x; }
         else numdelsing++; }
      multisite=haverun=FALSE;
      return; }
   if (haverun)
    { /* Add to a pre-existing run. */
      /* Continue beyond this block if */
      /*   (i) A new bloc needs to be started  or */
      /*   (ii) The site is polymorphic and the current partition */
      /*          needs to be stitched */
      if (offset!=lastoff+1)  addrun(NULL,nseqs, 0,0);
      /* The following extends a bloc with monomorphic missing data */
      else if (polyflag==FALSE) { lastoff=offset;  return; }}
   if (haverun==FALSE)   /* First site in a possible run */
    { if (polyflag || include_monosites)
       { memmove(state,site,nseqs);  multisite=FALSE;
         firstoff=lastoff=offset;  haverun=TRUE; }}
   else                  /* Successive site in a run */
    { lastoff=offset;  multisite=TRUE;
      addstitch(state,site,nseqs); }}


/* Combine `state' and `site' into a finer partition: */

static char *addstitch (char *state, char *site, int nseqs)
 { int i,j;  static char *newstate;  static int nxseqs;
   if (nseqs>nxseqs)
      newstate=rexxalloc(newstate,1, 0,nxseqs=nseqs+4);
   newlett(0);        /* Reset `newlett' */
   for (i=0; i<nseqs; i++)
    { for (j=0; j<i; j++)
        if (state[j]==state[i] && site[j]==site[i])
          { newstate[i]=newstate[j];  break; }
      if (j>=i)  newstate[i] = newlett(1); }
   return memmove(state, newstate, nseqs); }


static char newlett (int initflag)
 { static char ch;
   if (initflag==0)  return ch=0;
   if (ch==0)  return  ch='1';
   if (isdigit(ch))  return (ch=='9') ? ch='A' : ++ch;
   if (isupper(ch))  return (ch=='Z') ? ch='a' : ++ch;
   return ++ch; }


static int havmis(char *site, int nseqs)
 { int i;
   for (i=0; i<nseqs; i++)
    { if (strchr(read_knownchars,site[i])==NULL)  return TRUE; }
   return FALSE; }



/* (***HHH) */
/* Lower-level byte and line-reading package for sequence files */
/*  (All formats, including NEXUS) */

/* Byte-level routines for reading NEXUS distinguished comments */

/* typedef struct { char *s; int n,nx; }  STRBUFF, *STRBUFFP; */

static STRBUFF nxcomm;
static int nxcflag, ncbrak;

static void nx_clearcomm(void)
 { if (nxcomm.s)  free(nxcomm.s);
   memset(&nxcomm,0,sizeof(nxcomm));
   nxcflag=FALSE;  ncbrak=0; }

static void nx_opencomm(void)
 { nx_clearcomm();  nxcflag=TRUE;  ncbrak=1; }

static void nx_storecomm(void)
 { if (nxcomm.s!=NULL && nxcomm.n>0)
    { /* Strip open and closing spaces */
      char *q, *p=nxcomm.s;
      while(isspace(*p)) p++;
      q=p+strlen(p);  while(q>p && isspace(*--q)) *q=0;
      /* Don't store empty comments */
      if (p[0]!=0) addstr(&seqfcomments,p);
      /* Reset string buffer */
      memset(nxcomm.s, 0, nxcomm.n);
      nxcomm.n=0; }}

static int nx_addcomm(int c)
 { /* Store as separate strings if EOLs */
   if (c==0x0A) { nx_storecomm();  return c; }
   if (nxcomm.n>=nxcomm.nx)
     nxcomm.s=rexxalloc(nxcomm.s,1, nxcomm.n, nxcomm.nx+=50);
   return nxcomm.s[nxcomm.n++]=c; }

static void nx_closecomm(void)
 { /* Store string as comment */
   nx_storecomm();
   nx_clearcomm(); }



/* Main part of byte-oriented read package */

static int nbrak, lrsavec=-1;
static size_t lrnleft;
static char *lr_saveline;

static void reset_sqf_vals(void)
 { lrsavec=-1;  linenum=0;
   nbrak=lrnleft=0;  lr_saveline=NULL;
   nx_clearcomm(); }

static int unfbufc(int c)
 { return  lrsavec=c; }


#define LOWSIZE 3000      /* Buffer size for buffered fread() */

static int rawfgc(void)
 { static char *rp, rbuff[LOWSIZE+7];
   if (lrsavec>=0)
    { int c=lrsavec;  lrsavec=-1;  return c; }
   if (lrnleft==0)
    { rp=rbuff;
      if ( (lrnleft=fread(rbuff,1,LOWSIZE,lrff))==0)  return -1; }
   lrnleft--;
   return  *rp++; }


static int lowfgc(void)
 { int c;
   if ( (c=rawfgc())==0x0D)
    { if ( (c=rawfgc())!=0x0A)  unfbufc(c);
      linenum++;  return 0x0A; }
   if (c==0x0A)
    { linenum++;  return 0x0A; }
   return c; }


static int fbuffc(void)
 { int c;
   if ( (c=lowfgc())<0)  return -1;
   if (c==0)  c=1;        /* Change 0 in file to control A */
   if (fformat==F_NEXUS)  /* Check for NEXUS comments */
    { do { if (c=='[')
            { nbrak++;  ncbrak++;
              if (!no_nexus_dcomm)
               { if (nxcflag)  nx_addcomm(c);
                 else if ( (c=lowfgc())=='!')  nx_opencomm();
                 else unfbufc(c); }}
           else if (c==']')
            { if (nbrak==0)
               { textprintf ("UNMATCHED ']' in a NEXUS file!  "
                   "Ignored. (Line #%d)\n", linenum);
                 if (nxcflag) nx_addcomm(c); }
              else
               { nbrak--;  ncbrak--;
                 if (nxcflag)
                  { if (ncbrak>0)  nx_addcomm(c);
                    else nx_closecomm(); }}}
           else if (nbrak==0)  break;  /* Return c; */
           else if (nxcflag)  nx_addcomm(c); }
      while ( (c=lowfgc())>=0); }
   return c; }


/* Line-oriented part of read package */

static char *lrbuff;
static int lrxblen;

static void reset_readbuff(void)
 { if (lrbuff) free(lrbuff);  lrbuff=NULL;  lrxblen=0; }

static void pushline(char *str)
 { lr_saveline=str; }

static char *lineread(int allow_parse)
 { char *rs;
   if (lr_saveline!=NULL)
    { char *rs=lr_saveline;  lr_saveline=NULL;
      if (showlinesflag)  printf ("AGAIN(%s) `%s'\n", lrfname, rs);
      return rs; }
   if (lrfname==NULL || lrff==NULL)
     ErrHandler(
       "INTERNAL ERROR:  Attempt to read from `lineread()' before\n"
       "a file has been opened or after it has been closed!!!\n"
       "Line number: %d\n%s", linenum, Show_lbuff());
   /* This is needed if the first line read is a blank line . */
   if (lrbuff==NULL)  lrbuff=mxalloc( (lrxblen=LINEBAT)+3);

   for(;;)  /* Read lines */
    { int c,n=0;
      /* Fbuffc() returns 0x0A for EOL for UNIX,DOS, or Mac format */
      if ( (c=fbuffc())<0)  return NULL;
      do { if (c==0x0A)  break;
           if (n>=lrxblen)
             lrbuff=rexxalloc(lrbuff,1, n,lrxblen+=LINEBAT);
           lrbuff[n++]=c; }
      while ( (c=fbuffc())>=0);

      lrbuff[n]=0;
      if (n>maxsqlinelen)  maxsqlinelen=n;
      if (c<0) linenum++;   /* If there is no final EOL character */

      /* Remove white space after */
      while (n>0 && isspace(lrbuff[--n]))  lrbuff[n]=0;
      /* In most formats, remove white space before */
      rs=lrbuff;
      /* CLUSTAL format needs blank lines and lines with initial spaces */
      if (fformat!=F_CLUSTAL) { while(isspace(*rs))  rs++; }
      if (showlinesflag)
        printf ("%03d(%s) `%s'\n", linenum, lrfname, rs);
      if (fformat==F_CLUSTAL)  return lrbuff;
      /* Skip blank lines if not CLUSTAL format */
      if (*rs==0)  continue;
      /* If format supports comment characters, */
      /*   ignore lines beginning with a comment char */
      /*   (Currently ASF format only) */
      if (use_commchars() && commchars!=NULL
           && strchr(commchars,*rs)!=NULL)
       { /* Is it a DISTINGUISHED comment line?  */
         /*   Then save and parse, allowing (for example)  */
         /*   #! Matchchar=.  or  #!Gapchar=-    */
         if (rs[1]=='!' && allow_parse)
          { rs+=2;  while(isspace(*rs)) rs++;
            parse_dcomm(rs);
            addstr(&seqfcomments,rs); }}
      else break; }
   return  rs; }


static char *Show_lbuff(void)
 { static char fbuff[800];  int len, i;
   len=sprintf(fbuff,
     "\nFormat error in line #%d of `%s'\n", linenum, lrsfname);
   if (lrbuff!=NULL)
    { len+=sprintf(fbuff+len, "Line buffer = `%.120s'\n= `",  lrbuff);
      for (i=0; i<15; i++)
        len+=sprintf(fbuff+len, "%02X ", lrbuff[i]);
      sprintf (fbuff+len, "...'\n"); }
   return fbuff; }


/* (***III)  */
/* Functions for censoring offsets and converting polymorphism */
/*   numbers (start,finish) to alignment offsets */

static int skip_range (int offset)
 { if (base_range.s)
    { if (!(base_range.n<=offset && offset<=base_range.nx))
        return TRUE; }
   if (base_skip_range.s)
    { if (base_skip_range.n<=offset
               && offset<=base_skip_range.nx)  return TRUE; }
   return FALSE; }


/* Set first_offset and last_offset, taking base_range */
/*   and base_skip_range into account. */

static int first_offset, last_offset;

static void set_first_last(void)
 { first_offset=1;  last_offset=nbases;
   if (base_range.s!=NULL)
    { if (base_range.n>1)  first_offset=base_range.n;
      if (base_range.nx<nbases)  last_offset=base_range.nx; }
   if (base_skip_range.s!=NULL)
    { /* If base_skip_range brackets `first_offset' */
      if (base_skip_range.n<=first_offset
        && base_skip_range.nx>=first_offset)
         first_offset=base_skip_range.nx+1;
      /* If base_skip_range brackets `last_offset' */
      if (base_skip_range.nx>=last_offset
        && base_skip_range.n<=last_offset)
         last_offset=base_skip_range.n-1; }}


/* Minimum and maximum base offsets for a polymorphism # */

static int lfoffset(int r);
static int rtoffset(int r);

/* typedef struct { int lf,rt,caf, lfn,rtf; } FRAGOFFS;  */

FRAGOFFS offsfrags (int start, int finish)
 { FRAGOFFS x;  memset(&x,0,sizeof(x));
   x.lf=lfoffset(start);  x.rt=rtoffset(finish);
   /* Convert offsets in range -nbase < x < 2*nbases to (1,nbases) */
   while(x.lf<=0) { x.lf+=nbases;  x.caf=TRUE; }
   while(x.rt>nbases) { x.rt-=nbases;  x.caf=TRUE; }
   return x; }


int offsetlen (int rstart, int rfin)
 { return rtoffset(rfin) - lfoffset(rstart) + 1; }


int lfoffset(int r)
 { int k=0, lf;
   if (circular==FALSE)
    { lf = (r>0) ? polysites[r-1].locb+1 : first_offset; }
   else
    { /* Scale polymorphism number to 0<r<=npolys, */
      /*   so that 0<=r-1<npolys */
      while (r>npolys) { r-=npolys;  k++; }
      while (r<=0) { r+=npolys;  k--; }
      lf = polysites[r-1].locb+1; }
   /* Don't say fragment = 50-100  if  base_skip_range = 50-80 */
   if (base_skip_range.s!=NULL && lf==base_skip_range.n)
      lf=base_skip_range.nx+1;
   return lf + k*nbases; }


int rtoffset(int r)
 { int k=0, lf;
   if (circular==FALSE)
    { lf = (r<npolys) ? polysites[r].loca-1 : last_offset; }
   else
    { /* Scale polymorphism number to 0=<r<npolys */
      while (r>=npolys) { r-=npolys;  k++; }
      while (r<0) { r+=npolys;  k--; }
      lf = polysites[r].loca-1; }
   /* Don't say fragment = 50-100  if  base_skip_range = 70-100 */
   if (base_skip_range.s!=NULL && lf==base_skip_range.nx)
      lf=base_skip_range.n-1;
   return lf + k*nbases; }

