
/* Read and translate command-line arguments,
     including CFG files
   All rights reserved - Stanley Sawyer

   Entry points:
       set_options(argc,argv)
       show_commline(const char *bleep, int spaceflag)

   S. Sawyer - Department of Mathematics
               Washington University in St.Louis
               St. Louis, MO 63130
               email: sawyer@math.wustl.edu
               Office: 314-935-6703  FAX: 314-863-1302

   Parts of source file (line numbers approximate):

   (***AAA) (L202) Setflags() and friends

   (***BBB) (L1328) Readconfig() and friends

*/

#include "vall.h"
#include "vflags.h"
#include "vread.h"

char *AllRightsReserved_Vsetfl =
   "All rights reserved - Stanley Sawyer";

#define DEFCOMMCHARS "#;%$["    /* Default ASF file line-comment chars */
#define DEFNUMBERSEP ','        /* Default number separator */

#define LISTGLOBAL      2000    /* Default show this many global fragments */
#define DEFLISTPERPAIR  2000    /* This many per pair */

#define MAXSIMGLOBPVAL   0.05
#define MAXSIMPAIRPVAL   0.05   /* Default values */
#define MAXKAGLOBPVAL    0.05
#define MAXKAPAIRPVAL    0.05

#define GENPOLYONLY  "Gen_Poly"   /* Group for polymorphisms only */

#ifdef VSHOW
#define DEFSIM 0         /* Default number of simulations */
#else
#define DEFSIM 10000
#endif /* VSHOW */

#ifdef UNIX
#define IRRUPT 0         /* Default display iteration count this often */
#else /* if DOS */
#define IRRUPT 500       /* Default display iteration count this often */
#endif /* UNIX */

/* Various defaults: */

#define MINLENGTH 1      /* Default minimum length for fragment arrays */
#define MINNPOLY 2       /* Default minimum npolys for fragment arrays */
#define MINSCORE 2       /* Default minimum score for fragment arrays */
#define MAXOVERLAP 1     /* Default max no of overlapping fragments */

#define JSEQLEN 10       /* List this many sites near an insertion */

/* Magic constant for CFG file */

#define CONFIG_MAGIC   "#GCONV_CONFIG"
#define CONFIG_MAGIC_BUFFLEN  20

#define LOBVFFSIZE (MAXTOKSIZE+1000)
                          /* Low-level CFG buffer size */

/* Structures and functions for testmatch and listonly */

typedef struct { char *sq1,*sq2; } SEQPAIR;

typedef struct { char *title;  SEQPAIR *list;
                 int n,nx; }  SEQPAIRLIST, *SEQPAIRLISTP;

static SEQPAIRLIST
  testprlist = { "BLAST-like parameter display list:  "},
  listonlylist = { "List only these pairs/sequences:  " };


/* Variables set at the command line */

char *seqfile, *outfile, *logfile, *seqfdir;

G_GROUPTYPE grouptype=G_SIMPLE;
int save_argc, seqfixed_dir;
char **save_argv;

/* Basic command-line variables with a non-zero default */

char *commchars=DEFCOMMCHARS;
char matchchar, gapchar, misschar, numbersep=DEFNUMBERSEP;
int numsims=DEFSIM, irrupt=IRRUPT, jseqlen=JSEQLEN;

int listglobal=LISTGLOBAL, listpairs=FALSE, listperpair=DEFLISTPERPAIR;
int minlength=MINLENGTH, minnpoly=MINNPOLY, minscore=MINSCORE;
int maxoverlap=MAXOVERLAP;
SEQTYPE seqtype=AUTO;
TREAT_DELS treat_dels=TD_BLOCS;
double maxsimpairpval = MAXSIMPAIRPVAL;
double maxsimglobpval = MAXSIMGLOBPVAL;
double maxkapairpval  = MAXKAPAIRPVAL;
double maxkaglobpval  = MAXKAGLOBPVAL;
int havsimpairpval=TRUE, havsimglobpval=TRUE, bcsims;

int havkapairpval, havkaglobpval, TabKAPvals;
int setkapairpval, setkaglobpval;

int innerpairfrags=TRUE, outersims=TRUE;
int outerseqfrags=TRUE, outerpairfrags, outergrpfrags;
int pairplot, altcolor;

/* Primarily graphics variables */
/* Vertnumstep returns to defvertnumstep */
/*   unless keepvertscale=TRUE */
/* These are defined here so that the same configuration files */
/*   can be used */
int fancy_output;
int vertnumstep=4, defvertnumstep=4, vertroundstep, bcvertscale;
static int vertnumflag;
double vertminscale, vertmaxscale, vertstepscale;
int havminscale, havmaxscale, havstepscale;
int keepvertscale, vertskipstep, ylabels_oneside;
int addhashrules, skiptext, nocolor;

/* Used in reading tab files */
int havMinDispPvalue;
double minDispPvalue=1e-40;

/* Verbose variables */

int gscale, circular, allowindelstrat, havelog=TRUE;
int showunal, havlabels=TRUE, use_iubnucls, phylip;
int nomacfork, no_nexus_dcomm;
int dumpfrags, dumptab, dumpdif, dumpjseqs, dumpplot;
int dumpexit, use_underscores=TRUE, commline_shown;
int sortnames, sortgfragsbyseq, sortgfragsboth;
static int saydump, cl_outfile;
int annotate, saveseqs, notfound_nonfatal;
int show_mmsims, logfragments;
int expformat, widecols, AddOffset, truncate_pvals;
int showbcpwkapvals, showlog10pvals;
unsigned long wstartseed;
static unsigned long wseed;

/* Lists */

STRLIST seq_list, seq_skip_list, filepath, cfgnames;
STRBUFF base_range, base_skip_range;

/* Other single-char variables */

int nodegenflag, d3flag, simpflag, nosimpflag, mitcodes;
int blocsize, homolflag;
int ranflag, hold_window, logappend, include_monosites;
int showpolyflag, showblast, showfragdata;
static int showcftoks;
int ShowNexusToks, showlinesflag, showsites, showtabfrags;

/* Global variables that are not set from the command line */

int rshuffled, tabfileflag;

/* Called in vswin.c:  return resources if fatal error */
int err_clear;
void (*err_clear_funp)(void);

/* Variables saying that a flag has been set */

int newnumbersep, newstartseed, newlglobal;
int commline_sayorg;

#ifdef VSHOW

/* Functions and variables needed for re-entrant versions */
/*   of GENECONV */
/* These variables are NOT used in GENECONV or VTDS */
/* These variables are set just after set_options() */

static char *defcommchars;
static char defmatchchar, defgapchar, defmisschar, defnumbersep;
static int defnumsims, defirrupt, defjseqlen, defhavlabels, defnomacfork;
static int defminlength, defminnpoly, defminscore, defmaxoverlap;
static SEQTYPE defseqtype;
static TREAT_DELS def_treat_dels;
static double defmaxsimpairpval, defmaxsimglobpval;
static double defmaxkapairpval, defmaxkaglobpval;
static int defhavsimpairpval, defhavsimglobpval;
static int defhavkapairpval, defhavkaglobpval;

/* Reset the above variables */

void set_base_vars(void)
 { commchars=defcommchars;
   matchchar=defmatchchar;  gapchar=defgapchar;  misschar=defmisschar;
   numbersep=defnumbersep;  havlabels=defhavlabels;
   nomacfork=defnomacfork;
   numsims=defnumsims;  irrupt=defirrupt;  jseqlen=defjseqlen;
   minlength=defminlength;  minnpoly=defminnpoly;  minscore=defminscore;
   maxoverlap=defmaxoverlap;
   seqtype=defseqtype;  treat_dels=def_treat_dels;
   maxsimpairpval=defmaxsimpairpval;  maxsimglobpval=defmaxsimglobpval;
   havsimpairpval=defhavsimpairpval;  havsimglobpval=defhavsimglobpval;
   maxkapairpval=defmaxkapairpval;  maxkaglobpval=defmaxkaglobpval;
   havkapairpval=defhavkapairpval;  havkaglobpval=defhavkaglobpval; }

static void save_base_vars(void)
 { defcommchars=commchars;
   defmatchchar=matchchar;  defgapchar=gapchar;  defmisschar=misschar;
   defnumbersep=numbersep;  defhavlabels=havlabels;
   defnomacfork=nomacfork;
   defnumsims=numsims;  defirrupt=irrupt;  defjseqlen=jseqlen;
   defminlength=minlength;  defminnpoly=minnpoly;  defminscore=minscore;
   defmaxoverlap=maxoverlap;
   defseqtype=seqtype;  def_treat_dels=treat_dels;
   defmaxsimpairpval=maxsimpairpval;  defmaxsimglobpval=maxsimglobpval;
   defhavsimpairpval=havsimpairpval;  defhavsimglobpval=havsimglobpval;
   defmaxkapairpval=maxkapairpval;  defmaxkaglobpval=maxkaglobpval;
   defhavkapairpval=havkapairpval;  defhavkaglobpval=havkaglobpval; }

#endif


/* Static functions and variables */

/* Functions for reading CONFIG files */
/* Since several CONFIG files may be open simultaneously, */
/*   these functions must be completely reentrant */

typedef struct cfgfile_x
  { const char *cfname;
    char *lobuff, *buffp, *savtok, *cfullname, *fdir;
    struct cfgfile_x  *cfparent;  STRLIST tok;
    FILE *cff;  int ord,nleft,savec,fxdir; }  CFGFILE;

typedef struct { int cnum,cxnum;
                 CFGFILE *cfg; }  CFGLIST;

static CFGLIST cfglist, cfglist_open;

static char *readfirtokcfg(CFGFILE *cf);
static char *readtokcfg(CFGFILE *cf);
static void pushtokcfg(CFGFILE *ct, const char *token);
static void check_cfg_recurse (CFGFILE *cf, const char *fname);

/* Other static functions */

static void parseword(const char *p, CFGFILE *cfpar);
static void parse_verbose(char *verb, CFGFILE *cf, int options_env);
static int flag_yes(const char *type);
static int flag_no(const char *type);
static int do_outfile(const char *type);
static int set_onerange (STRBUFFP ap, const char *type);
static void parsecharflags(char *p);
static char *showcfname(CFGFILE *cf);
static void cfg_err(const char *bleep, const char *bleep2,
      const char *verb, const char *type, CFGFILE *cf);
static void badverb(const char *bleep,
    const char *verb, const char *type, CFGFILE *cf);
static void unspace_permgroup(void);
static void unspace_splist(SEQPAIRLIST seqst);

static void read_config(const char *fn, CFGFILE *cfpar);
static void read_options(const char *type, CFGFILE *cf);
static void read_seqlist(STRLISTP ap,
     const char *verb, const char *type, CFGFILE *cf);
static void read_group(const char *type, CFGFILE *cf);

#ifdef VTEXT
void no_seqf(void);   /* If seqfile==NULL */
#endif

#ifdef VSHOW
/* Two functions used in vswin.c */
int read_fragshow (const char *p);

void read_nconfig(const char *type)
 { read_config(type,NULL); }
#endif


/* (***AAA)  set_options() and friends */

void set_options (int argc, char **argv)
 { int i;
   /* To be safe: */
   save_argc=argc;  save_argv=argv;  commline_shown=FALSE;
   seqfile=outfile=logfile=NULL;  clear_alldumps();
   cl_outfile=saydump=FALSE;  kill_groups();  setdna(FALSE);
   kill_strbuff(&base_range);  kill_strbuff(&base_skip_range);

   /* Read the command line, including CFG files, */
   /*   while preserving the original copies of argv[] */
   for (i=1; i<argc; i++)  parseword(argv[i],NULL);

   /* Save the opening list of CONFIG files */
   cfglist_open=cfglist;  memset(&cfglist,0,sizeof(cfglist));

   /* Make sure that file names are filled in properly */
   set_filenames();

   #ifdef VTEXT
   /* Make sure that an input file name was entered. */
   if (seqfile==NULL || seqfile[0]==0)  no_seqf();
   #endif

   /* Initialize the random-number seed */
   wstartseed=setrand(wseed);

   /* Some final checks on options */
   if (saydump==FALSE && !(dumptab || dumpdif))  dumpfrags=TRUE;

   /* The default is havsimpairpval=havsimglobpval=TRUE */
   /*   and  havkapairpval=havkaglobpval=FALSE */
   /* Maxsimpairpval and maxsimglobpval are never used if numsims==0 */
   /* In that case, turn on ka bounds unless they have been */
   /*   explicitly turned off. */
   if (numsims==0)
    { if (setkapairpval==FALSE)  havkapairpval=TRUE;
      if (setkaglobpval==FALSE)  havkaglobpval=TRUE; }

   #ifdef VSHOW
   /* Make graphics vertical-scale options consistent */
   if (defvertnumstep<1)  defvertnumstep=4;
   if (vertnumstep<1)  vertnumstep=defvertnumstep;
   if (havminscale)
    { if (havmaxscale)
       { havstepscale=TRUE;
         vertstepscale=(vertmaxscale-vertminscale)/vertnumstep; }
      else if (havstepscale && vertnumflag)
       { havmaxscale=TRUE;
         vertmaxscale=vertminscale+vertnumstep*vertstepscale; }}
   #endif /* VSHOW */

   /* Check for contradictory or illegal flag settings */
   checkflags();

   /* If blocsize>0 and grouptype==G_GROUPS, */
   /*   then checkflags sets blocsize=0 */
   /* Otherwise, if blocsize>0, set grouptype=G_BLOCS here */
   if (blocsize>0 && grouptype==G_SIMPLE) grouptype=G_BLOCS;

   /* If set_underscores==TRUE, convert perm_group[], */
   /*   seq_list, seq_skip_list, etc. */
   /* Sequence name comparisons use `striucmp()', which does not */
   /*   distinguish between spaces and underscores. Thus these */
   /*   conversions are only for displays. */
   if (use_underscores)
    { unspace_permgroup();
      unspace_slist2(seq_list);
      unspace_slist2(seq_skip_list);
      unspace_splist(testprlist);
      unspace_splist(listonlylist); }


   if (matchchar==gapchar || matchchar==misschar)  matchchar=0;

   #ifdef VSHOW
   /* Save values of key options if re-entrant */
   save_base_vars();
   #endif
         }



void set_filenames(void)
 { if (logfile==NULL)
    { if (outfile) logfile=strxdup(outfile);
      else if (seqfile) logfile=strxdup(seqfile); }
   if (logfile && checkfname(logfile))
    { char *ps=logfile, *lname="LOGFILE", *cbuff=NULL;
      if (outfile)
       { PARTFN pf=partfn(outfile);
         if (pf.name[0])  lname=cbuff=strxdup(pf.name);  freepf(pf); }
      if (cbuff==NULL && seqfile)
       { PARTFN pf=partfn(seqfile);
         if (pf.name[0])  lname=cbuff=strxdup(pf.name);  freepf(pf); }
      logfile=makefnam(lname,ps);  free(ps);  if (cbuff) free(cbuff); }
   if (outfile==NULL && seqfile)  outfile=strxdup(seqfile);
   if (outfile && checkfname(outfile))
    { char *ps=outfile, *oname="OUTFILE", *cbuff=NULL;
      if (seqfile)
       { PARTFN pf=partfn(seqfile);
         if (pf.name[0])  oname=cbuff=strxdup(pf.name);  freepf(pf); }
      outfile=makefnam(oname,ps);  free(ps);  if (cbuff) free(cbuff); }
   if (seqfile && checkfname(seqfile))
    { char *ps=seqfile;  seqfile=makefnam("SEQFILE",ps);  free(ps); }}


void no_degens(void)
 { nodegenflag=d3flag=simpflag=nosimpflag=mitcodes=FALSE; }

int set_degens(char *dbuff)
 { int n=0;
   if (nodegenflag) dbuff[n++]='A';
   if (d3flag)      dbuff[n++]='3';
   if (simpflag)    dbuff[n++]='S';
   if (nosimpflag)  dbuff[n++]='M';
   if (mitcodes)    dbuff[n++]='V';
   dbuff[n]=0;
   return n; }


/* `Parseword' is called ONLY from the command line, with cfpar==NULL, */
/*   OR ELSE from a configuration file, with cfpar=&(config struct) */

static void parseword(const char *inarg, CFGFILE *cfpar)
 { char *p=strxdup((char *)inarg);  int n;

   if (showcftoks)
     printf ("  PARSEWORD %s (%s)\n", p, showcfname(cfpar));
   while (isspace(*p)) p++;
   /* Configfile comments and distinguished comments will be */
   /*   stripped off in `readtokcfg()' in `read_config()' */
   /*   before getting to this point. */

   /* First, make sure that `inarg' is not too long */
   if ( (n=strlen(p))>MAXTOKSIZE)
    { textprintf (
      "\nWARNING: INPUT WORD (len=%d) TRUNCATED to %d bytes !!\n",
        n, MAXTOKSIZE);  p[MAXTOKSIZE]=0; }

   /* Is `inarg' a verbose-style option? */
   /* The final argument is TRUE from read_options() */
   if (*p == '-')  parse_verbose(p+1,cfpar,FALSE);

   /* Perhaps a string of one or more short-form options? */
   /* Since parsecharflags(p) begins with  switch ( (c=toupper(*++p)), */
   /*   p must not be increased here. */
   else if (*p=='/')  parsecharflags(p);

   /* Is it the name of a config file? */
   else if (isextent(p,"cfg"))  read_config(p,cfpar);

   /* If seqfile==NULL, assume that it is seqfile */
   else if (seqfile==NULL || seqfile[0]==0)  seqfile=strxdup(p);

   /* If seqfile!=NULL, assume that it is the output file, */
   /*   allowing for a previous -outpath=  command. */
   else if (cl_outfile==FALSE)  do_outfile(p);

   /* If outfile!=NULL, assume that it is the log file */
   else if (logfile==NULL || logfile[0]==0)  logfile=strxdup(p);

   /* Exit if four or more file names are entered. */
   else  exitprintf (" FOUR FILE NAMES "
     " <Seqfile>=%.100s\n"
     " <Outfile>=%.100s\n"
     " <Logfile>=%.100s\n"
     " and a fourth file name `%.100s' were entered !\n"
     " The fourth file name is an error.\n"
     " Enter -seqfile=<Seqfile>, -outfile=<Outfile>, or\n"
     " or -logfile=<Logfile> to override previous assignments.",
              seqfile, outfile, logfile, p);

   free(p); }



typedef struct { char *descript;  double *dptr;
                 int *havptr, *setptr; }  DUBSTRUC;

static DUBSTRUC
  simpairpval = { "SimPairPval", &maxsimpairpval, &havsimpairpval },
  simglobpval = { "SimGlobPval", &maxsimglobpval, &havsimglobpval },
  kapairpval = { "KAPairPval", &maxkapairpval,
                       &havkapairpval, &setkapairpval },
  kaglobpval = { "KAGlobPval", &maxkaglobpval,
                       &havkaglobpval, &setkaglobpval },
  minscale = { "MinScale", &vertminscale, &havminscale },
  maxscale = { "MaxScale", &vertmaxscale, &havmaxscale },
  stepscale = { "StepScale", &vertstepscale, &havstepscale },
  DispPvalue = { "DispPval", &minDispPvalue, &havMinDispPvalue };


static void setpvalp (DUBSTRUC a, double pval)
 { if (pval==0.0)
     ErrHandler ("ATTEMPT TO SET Max%s = 0.0 !!  (Did you mean 1.0?)",
       a.descript);
   if (pval<=MAXPVAL) { *(a.dptr)=pval;  *(a.havptr)=TRUE; }
   else *(a.havptr)=FALSE; }


/* Structures and functions for testmatch and listonly */

static int seqpair_match(SEQPAIRLIST seqst, int i1, int i2)
 { if (seqst.list!=NULL && seqst.n>0 && i1>=0)
    { char *sq1=readtag[i1], *sq2=(i2>=0) ? readtag[i2] : NULL;
      int i;
      for (i=0; i<seqst.n; i++)
       { if (!striucmp(sq1,seqst.list[i].sq1))
          { char *tp2=seqst.list[i].sq2;
            if ( (i2<0 && tp2[0]==0) ||
              (i2>=0 && !striucmp(sq2,tp2)) )  return TRUE; }}}
   return FALSE; }


int match_showparms(int i1, int i2)
 { return seqpair_match(testprlist,i1,i2); }

int match_onlylist(int i1, int i2)
 { return seqpair_match(listonlylist,i1,i2); }


int is_onlylist(void)
 { return (listonlylist.list!=NULL && listonlylist.n>0); }



static int show_seqpair(SEQPAIRLIST seqst)
 { if (seqst.list!=NULL && seqst.n>0)
    { int i;
      TOKLIST tokl;  memset(&tokl,0,sizeof(tokl));
      for (i=0; i<seqst.n; i++)
       { char qbuff[MAXNAMETWO];
         if (seqst.list[i].sq2[0])
          { char q1b[MAXNAMELEN], q2b[MAXNAMELEN];
            doquote(q1b,seqst.list[i].sq1);
            doquote(q2b,seqst.list[i].sq2);
            sprintf(qbuff, "%s;%s", q1b,q2b); }
         else
           doquote(qbuff,seqst.list[i].sq1);
         addtokl(&tokl,qbuff,TOK_VOLATILE); }
      ftoklistex(seqst.title, &tokl, 2,NULL,0, TOK_FREE); }
   return seqst.n; }


static int show_testlist(void)
 { return show_seqpair(testprlist); }

static int show_onlylist(void)
 { return show_seqpair(listonlylist); }


static void unspace_splist(SEQPAIRLIST seqst)
 { if (seqst.list!=NULL && seqst.n>0)
    { int i;
      for (i=0; i<seqst.n; i++)
       { unspace_str(seqst.list[i].sq1);
         unspace_str(seqst.list[i].sq2); }}}


/* Warning: DON'T destroy  sp->title !! */

static void kill_seqpairlist (SEQPAIRLIST *sp, int sflag)
 { if (sp->list!=NULL)
    { char *p;  int n=sp->n;  SEQPAIR *list=sp->list;
      if (sflag)  show_seqpair(*sp);
      while (--n>=0)
       { if ( (p=list[n].sq1)!=NULL && p[0])  free(p);
         if ( (p=list[n].sq2)!=NULL && p[0])  free(p); }
      free(sp->list); }
   sp->list=NULL;  sp->n=sp->nx=0; }


void kill_testlist (int sflag)   /* Also called in vswin.c */
 { kill_seqpairlist(&testprlist,sflag); }

void kill_onlylist (int sflag)
 { kill_seqpairlist(&listonlylist,sflag); }


static void read_seqpair(SEQPAIRLISTP sp,
   const char *verb, const char *type, CFGFILE *cf);


static void do_xout(const char *type)
 { if (outfile)
    { char *ps=outfile;  outfile=makefnam(type,ps);  free(ps); }
   else outfile=strxdup(type); }

static int do_outfile(const char *type)
 { if (type==NULL || type[0]==0)  return RET_ERROR;
   do_xout(type);  cl_outfile=TRUE;
   return RET_OK; }

static int do_outpath(const char *type)
 { if (type==NULL || type[0]==0)  return RET_ERROR;
   else
    { char *p, *ntype=strxndup(type,3);
      p=ntype+strlen(ntype);
      while(p>ntype && (*--p=='\\' || *p=='/'))  *p=0;
      *++p=PATHSEP;  *++p=0;
      do_xout(ntype);  free(ntype);
      return RET_OK; }}



static void nomaxpval(void)
 { havsimpairpval=havsimglobpval=FALSE;
   havkapairpval=havkaglobpval=FALSE;
   setkapairpval=setkaglobpval=TRUE; }


static void listbest(int onflag)
 { if (onflag)
    { listglobal=8;  listperpair=3;
      newlglobal=TRUE;  nomaxpval(); }
   else
    { listglobal=LISTGLOBAL;
      if (listperpair<=3)  listperpair=DEFLISTPERPAIR;
      havsimpairpval=havsimglobpval=TRUE;
      setkapairpval=setkaglobpval=FALSE; }}


static void do_fancy(int onflag)
 { if (onflag)
    { expformat=annotate=TRUE;
      #ifdef VSHOW
      fancy_output=TRUE;
      vertminscale=0.0;  vertstepscale=5.0;
      havminscale=havstepscale=skiptext=addhashrules=TRUE;
      vertroundstep=2;  vertskipstep=10;
      bcvertscale=keepvertscale=ylabels_oneside=TRUE;
      #endif
         }
   else
    { expformat=annotate=FALSE;
      #ifdef VSHOW
      fancy_output=FALSE;
      havminscale=havstepscale=skiptext=addhashrules=FALSE;
      bcvertscale=keepvertscale=ylabels_oneside=FALSE;
      #endif
         }}



/* For use in Parse_word(): */
/* RFUNV is a pointer to a function */
/*   that takes no arguments and has no return */
/* RFUNCP is a pointer to a function */
/*   that takes a string pointer and returns an integer. */
/* RFUNCCFP is a pointer to a function */
/*   that takes two arguments and returns an integer. */
/* By C convention, listing the name of a function (or an array) */
/*   in an array or structure lists a pointer to that function or array. */


typedef void (*RFUNV)(void);
typedef int (*RFUNCP) (char *type);
typedef void (*RFUNCCFP) (char *type, CFGFILE *cfpar);



typedef enum { V_NONE, V_FLAG, V_NFLAG,
                 V_CHAR, V_INT, V_INTPOS, V_ULONG,
                 V_STRING, V_STRINGX, V_DBSTRUC, V_PVAL,
               V_NOMAXSIMPVAL, V_NOMAXKAPVAL,
               V_LISTBEST, V_LISTPERPAIR,
               V_NODUMP, V_NO_OUTPUT, V_DUMPALL, V_ALLOUTER, V_NOOUTER,
               V_INDELSKIP,  V_INDELSITE,  V_INDELBLOC,
               V_FANCY,  V_READGROUP, V_READSEQPAIRS,
               V_READFUNV, V_READFUNCP, V_READFUNCCFP,
               V_SEQLIST, V_ADD_SEQLIST, V_SETRANGE,
               V_SETPATH, V_ADD_SETPATH }  V_TYPE;


typedef struct { char *verb;  void *targ;  V_TYPE vtype;
                 int *newflag; }  VERBLIST;


/* Parse_verb() sets a fatal error if a V_INT, V_PVAL, */
/*   or V_DBSTRUC entry is set < 0 */

VERBLIST verblist[] = {
 /* Input/output file names and output file types */
  { "seqfile",           &seqfile,           V_STRINGX },
  { "outfile",           do_outfile,         V_READFUNCP },
  { "outpath",           do_outpath,         V_READFUNCP },
  { "logfile",           &logfile,           V_STRINGX },
  { "logappend",         &logappend,         V_FLAG },
  { "dumptab",           &dumptab,           V_FLAG },
  { "dumpdif",           &dumpdif,           V_FLAG } ,
  { "dumpfrag",          &dumpfrags,         V_FLAG },
  { "dumpjseq",          &dumpjseqs,         V_FLAG },
  { "dumpplot",          &dumpplot,          V_FLAG },
  { "dumpall",           NULL,               V_DUMPALL },
  { "jseqlen",           &jseqlen,           V_INT },
  { "nolog",             &havelog,           V_NFLAG } ,
  { "havelog",           &havelog,           V_FLAG } ,
  { "nodump",            NULL,               V_NODUMP } ,
  { "no_output",         NULL,               V_NO_OUTPUT } ,
 /* Options for input file format */
  { "seqtype",           read_sqtype,        V_READFUNCP },
  { "matchchar",         &matchchar,         V_CHAR },
  { "commchar",          &commchars,         V_STRING },
  { "gapchar",           &gapchar,           V_CHAR },
  { "misschar",          &misschar,          V_CHAR },
  { "nolabels",          &havlabels,         V_NFLAG },
  { "use_iub",           &use_iubnucls,      V_FLAG },
  { "mitcodes",          &mitcodes,          V_FLAG },
  { "phylip",            &phylip,            V_FLAG },
  { "nomac",             &nomacfork,         V_FLAG },
  { "nocomm",            &no_nexus_dcomm,    V_FLAG },
  { "skip_indels",       NULL,               V_INDELSKIP },
  { "no_indels",         NULL,               V_INDELSKIP },
  { "use_individual_indel",  NULL,           V_INDELSITE },
  { "use_indel",         NULL,               V_INDELBLOC },
  { "indel_bloc",        NULL,               V_INDELBLOC },
  { "indel_run",         NULL,               V_INDELBLOC },
 /* Options for controlling fragment lists */
  { "gscale",            &gscale,            V_INT },
  { "circular",          &circular,          V_FLAG },
  { "linear",            &circular,          V_NFLAG },
  { "minlength",         &minlength,         V_INTPOS },
  { "minnpoly",          &minnpoly,          V_INTPOS },
  { "minscore",          &minscore,          V_INTPOS },
  { "maxoverlap",        &maxoverlap,        V_INT },
  { "maxovlap",          &maxoverlap,        V_INT },
  { "listglobal",        &listglobal,        V_INT, &newlglobal },
  { "listperpair",       NULL,               V_LISTPERPAIR },
  { "listpair",          &listpairs,         V_FLAG },
  { "pairlist",          &listpairs,         V_FLAG },
  { "pairwise",          &listpairs,         V_FLAG },
  { "listbest",          NULL,               V_LISTBEST },
  { "nopairs",           &listpairs,         V_NFLAG },
  { "nopairlist",        &listpairs,         V_NFLAG },
  { "nopairwise",        &listpairs,         V_NFLAG },
  { "include_monosite",  &include_monosites, V_FLAG },
 /* Options for which fragment types to analyze */
  { "innerpair",         &innerpairfrags,    V_FLAG },
  { "outerseq",          &outerseqfrags,     V_FLAG },
  { "outerpair",         &outerpairfrags,    V_FLAG },
  { "outergroup",        &outergrpfrags,     V_FLAG },
  { "outergrp",          &outergrpfrags,     V_FLAG },
  { "allouter",          NULL,               V_ALLOUTER },
  { "noouter",           NULL,               V_NOOUTER },
  { "outersims",         &outersims,         V_FLAG },
  { "pairplot",          &pairplot,          V_FLAG },
 /* Options for P-value bounds in fragment lists */
  { "maxsimpwpval",        &simpairpval,       V_PVAL },
  { "maxsimpairpval",      &simpairpval,       V_PVAL },
  { "maxsimpairwisepval",  &simpairpval,       V_PVAL },
  { "maxsimmcpval",        &simglobpval,       V_PVAL },
  { "maxsimglobpval",      &simglobpval,       V_PVAL },
  { "maxsimglobalpval",    &simglobpval,       V_PVAL },
  { "maxkapwpval",         &kapairpval,        V_PVAL },
  { "maxkapairpval",       &kapairpval,        V_PVAL },
  { "maxkapairwisepval",   &kapairpval,        V_PVAL },
  { "maxkamcpval",         &kaglobpval,        V_PVAL },
  { "maxkaglobpval",       &kaglobpval,        V_PVAL },
  { "maxkaglobalpval",     &kaglobpval,        V_PVAL },
  { "nomaxsimpval",        NULL,               V_NOMAXSIMPVAL },
  { "nomaxkapval",         NULL,               V_NOMAXKAPVAL },
  { "nomaxpval",           nomaxpval,          V_READFUNV },
  { "useblast",            &bcsims,            V_NFLAG },
  { "bcsim",               &bcsims,            V_FLAG },
  { "noblast",             &bcsims,            V_FLAG },
 /* Options for permutations */
  { "numsim",            &numsims,           V_INT },
  { "irrupt",            &irrupt,            V_INT },
  { "startseed",         &wseed,             V_ULONG, &newstartseed },
  { "seed",              &wseed,             V_ULONG, &newstartseed },
  { "randomize_site",    &ranflag,           V_FLAG },
  { "ignore_degen",      &nodegenflag,       V_FLAG },
  { "combine_3degen",    &d3flag,            V_FLAG },
  { "simp_poly_only",    &simpflag,          V_FLAG },
  { "mult_poly_only",    &nosimpflag,        V_FLAG },
 /* Options for customizing output file format */
  { "annotate",          &annotate,          V_FLAG },
  { "expformat",         &expformat,         V_FLAG },
  { "widecol",           &widecols,          V_FLAG },
  { "fancy",             NULL,               V_FANCY },
  { "addoffset",         &AddOffset,         V_INT },
  { "showblast",         &showblast,         V_FLAG },
  { "showfragdata",      &showfragdata,      V_FLAG },
  { "showpolys",         &showpolyflag,      V_FLAG },
  { "showbcpwkapval",    &showbcpwkapvals,   V_FLAG },
  { "showlog10pval",     &showlog10pvals,    V_FLAG },
  { "truncate_pvals",    &truncate_pvals,    V_FLAG },
  { "sortname",          &sortnames,         V_FLAG },
  { "sortgfragsbyseq",   &sortgfragsbyseq,   V_FLAG },
  { "sortgfragsboth",    &sortgfragsboth,    V_FLAG },
  { "use_underscore",    &use_underscores,   V_FLAG },
  { "numbersep",         &numbersep,         V_CHAR,   &newnumbersep },
  { "show_maxmeansim",   &show_mmsims,       V_FLAG },
  { "showunal",          &showunal,          V_FLAG },
  { "logfragment",       &logfragments,      V_FLAG },
  { "tok_linesize",      &tok_linesize,      V_INT },
  { "notimer",           &notimer,           V_FLAG },
 /* Options for lists */
  { "seq_list",          &seq_list,          V_SEQLIST },
  { "add_seq_list",      &seq_list,          V_ADD_SEQLIST },
  { "subset",            &seq_list,          V_SEQLIST },
  { "add_subset",        &seq_list,          V_ADD_SEQLIST },
  { "seq_skip",          &seq_skip_list,     V_SEQLIST },
  { "add_seq_skip",      &seq_skip_list,     V_ADD_SEQLIST },
  { "skip_list",         &seq_skip_list,     V_SEQLIST },
  { "add_skip_list",     &seq_skip_list,     V_ADD_SEQLIST },
  { "list_only",         &listonlylist,      V_READSEQPAIRS },
  { "add_list_only",     &listonlylist,      V_READSEQPAIRS },
  { "listonly",          &listonlylist,      V_READSEQPAIRS },
  { "add_listonly",      &listonlylist,      V_READSEQPAIRS },
  { "base_range",        &base_range,        V_SETRANGE },
  { "base_skip",         &base_skip_range,   V_SETRANGE },
  { "filepath",          &filepath,          V_SETPATH },
  { "input",             &filepath,          V_SETPATH },
  { "add_filepath",      &filepath,          V_ADD_SETPATH },
  { "add_input",         &filepath,          V_ADD_SETPATH },
 /* Options for configuration files */
  { "config",            read_config,        V_READFUNCCFP },
  { "option",            read_options,       V_READFUNCCFP },
  { "group",             NULL,               V_READGROUP },
 /* Some options for groups */
  { "bloc",              &blocsize,          V_INT },
  { "homologous",        &homolflag,         V_FLAG },
  { "notfound_nonfatal", &notfound_nonfatal, V_FLAG },
 /* Debug-like display options */
  { "test",              &testprlist,        V_READSEQPAIRS },
  { "showcftok",         &showcftoks,        V_FLAG },
  { "shownexustok",      &ShowNexusToks,     V_FLAG },
  { "showline",          &showlinesflag,     V_FLAG },
  { "showsite",          &showsites,         V_FLAG },
 /* This option keeps the console window temporarily open on a normal */
 /*   termination. This is automatic on a fatal error. */
  { "hold_window",       &hold_window,       V_FLAG },
 /* The following options are used only in VSHOW, but are included */
 /*   here so that the same configuration files can be used */
  { "allowindel",        &allowindelstrat,   V_FLAG },
  { "skiptext",          &skiptext,          V_FLAG },
  { "showtext",          &skiptext,          V_NFLAG },
  { "nocolor",           &nocolor,           V_FLAG },
  { "altcolor",          &altcolor,          V_FLAG },
  { "saveseq",           &saveseqs,          V_FLAG },
  { "dumpexit",          &dumpexit,          V_FLAG },
  #ifdef VSHOW
  { "fragshow",          read_fragshow,      V_READFUNCP },
  #else
  { "fragshow",          NULL,               V_NONE },
  #endif
 /* Options for the vertical scale and axes: */
 /* These are used only in VSHOW. */
 /* If havmin and havstep but not havmax, then round up vertmax */
 /*   to the next multiple of vertround*vertstep */
 /* These are included here so that the same configuration files */
 /*   can be used */
  { "bcvertscale",       &bcvertscale,       V_FLAG },
  { "vertnumstep",       &vertnumstep,       V_INT,    &vertnumflag },
  { "vertroundstep",     &vertroundstep,     V_INT },
  { "vertskipstep",      &vertskipstep,      V_INT },
  { "defvertnumstep",    &defvertnumstep,    V_INT },
  { "vertminscale",      &minscale,          V_DBSTRUC },
  { "vertmaxscale",      &maxscale,          V_DBSTRUC },
  { "vertstepscale",     &stepscale,         V_DBSTRUC },
  { "hashrules",         &addhashrules,      V_FLAG },
  { "addhashrules",      &addhashrules,      V_FLAG },
  { "ylabels_oneside",   &ylabels_oneside,   V_FLAG },
  { "keepvertscale",     &keepvertscale,     V_FLAG },
 /* These options are used only for reading tabfiles in VSHOW */
  { "MinDispPvalue",     &DispPvalue,        V_DBSTRUC },
  { "TabKAPvals",        &TabKAPvals,        V_FLAG },
  { "showtabfrag",       &showtabfrags,      V_FLAG },
 /* End of list */
  { NULL }, { NULL } };



/* The basic parsing function is `parseword()'      */
/* The program comes here if it thinks that the option */
/*    is of the form  -verb=type  */

/* `Verb' either points to a `strxdup'ed local buffer */
/*   or else to a static buffer in readtokcfg(), neither of which */
/*   need to be preserved. */

static void parse_verbose(char *verb, CFGFILE *cf, int options_env)
 { int matched=FALSE, haveq=FALSE;
   char *type, *errmess=NULL;
   VERBLIST *v;

   /* Separate out `type' if `verb' of the form `verb=type' */
   /* Act intelligently if the word reads `verb==type' */
   if ( (type=strchr(verb,'='))!=NULL)
    { haveq=TRUE;  do *type++=0; while (*type=='='); }
   else type="";

   if (showcftoks)
     printf ("  PARSE_VERB %s (=) %s (%s)\n", verb, type, showcfname(cf));
   if (verb[0]==0)  return;  /* Typo at command line? */

   /* Multiple entries for the same function or verb are allowed. */
   /* Matching is case insensitive on minimal acceptable substring */
   for (v=verblist; v->verb; v++)  if (!strmatch(verb,v->verb))
    { char ebuff[288];  matched=TRUE;
      /* If there is an auxilliary flag, set it */
      /*   unless the type is flag set/reset */
      if (v->newflag && v->vtype!=V_FLAG && v->vtype!=V_NFLAG)
          *v->newflag=TRUE;
      switch(v->vtype) {
        default:  if (strlen(verb)<=12)
            sprintf(ebuff, "Command `%s' is unknown.", verb);
          else  sprintf(ebuff, "Command `%.12s...' is unknown.", verb);
          errmess=ebuff;  goto verb_err;
        verb_nonpos: errmess="A number >= 1 is required here.";
                     goto verb_err;
        verb_neg:  errmess="A number >= 0 is required here.";
                     goto verb_err;
        verb_options:
          sprintf(ebuff,"YOU CANNOT USE `%s' in an `options' environment !\n"
            "Enter ``endoptions'' before this command.\n", v->verb);
          errmess=ebuff;  goto verb_err;
        verb_err:  badverb(errmess,verb,type,cf);  break;
        case V_NONE:  break;
      /* Set simple flag options */
        case V_FLAG:
          if (read_flag(type,v->targ,v->newflag))  goto verb_err;
          break;
        case V_NFLAG:
          if (read_nflag(type,v->targ,v->newflag))  goto verb_err;
          break;
        case V_CHAR:
          /* Syntax  -matchchar= OR -matchchar=0  sets  matchchar=0  */
          if (v->targ)  *(char *)v->targ = set_vchar(type);
          break;
        case V_INT:
          if (v->targ)
           { int xx;
             if (flag_no(type))  xx=0;
             else if (sscanf(type,"%d",&xx)<1)  goto verb_err;
             else if (xx<0)  goto verb_neg;
             *(int *)v->targ=xx; }  break;
        case V_INTPOS:
          if (v->targ)
           { int xx;
             if (flag_no(type))  xx=0;
             else if (sscanf(type,"%d",&xx)<1)  goto verb_err;
             if (xx<=0)  goto verb_nonpos;
             *(int *)v->targ=xx; }  break;
        case V_ULONG:
          if (!isdigit(type[0]))  goto verb_err;
          if (v->targ)
           { char c;  unsigned long uu=0L;
             if (!flag_no(type))
              { /* We cannot depend on having sscanf("%lu") nor atoul() */
                if (!isdigit(type[0]))  goto verb_err;
                while ( (c=*type++)!=0 && isdigit(c))
                   uu=10L*uu + (c-'0'); }
             *(unsigned long *)v->targ=uu; }  break;
      /* String, double, and pval variables */
        case V_STRINGX:
          if (type[0]==0)  goto verb_err;
          goto do_string;
        case V_STRING:
        do_string:
          if (haveq==FALSE)  goto verb_err;
          if (v->targ)  *(char **)v->targ
           = strxdup(parse_unquote(type,NULL,NULL));
          break;
        case V_DBSTRUC:
          if (v->targ)
           { DUBSTRUC a=*(DUBSTRUC *)(v->targ);
             double zz;
             if (a.dptr==NULL || a.havptr==NULL)  break;
             if (a.setptr) *(a.setptr)=TRUE;
             if (flag_no(type))  *(a.havptr)=FALSE;
             else if (flag_yes(type))  *(a.havptr)=TRUE;
             else if (sscanf(type,"%lf",&zz)==1)
              { if (zz>=0.0)  { *(a.dptr)=zz;  *(a.havptr)=TRUE; }
                else  goto verb_neg; }
             else  goto verb_err; }  break;
        case V_PVAL:
          if (v->targ)
           { DUBSTRUC a=*(DUBSTRUC *)(v->targ);
             double pval;
             if (a.dptr==NULL || a.havptr==NULL)  break;
             if (a.setptr) *(a.setptr)=TRUE;
             /* Setting >MAXPVAL is one way to clear this option */
             if (flag_no(type))  *(a.havptr)=FALSE;
             else if (flag_yes(type))  *(a.havptr)=TRUE;
             else if (sscanf(type,"%lf",&pval))
              { if (pval>=0.0)  setpvalp(a,pval);
                else  goto verb_neg; }
             else  goto verb_err; }  break;
      /* Various other options */
        case V_NOMAXSIMPVAL:
             havsimpairpval=havsimglobpval=FALSE;  break;
        case V_NOMAXKAPVAL:
             havkapairpval=havkaglobpval=FALSE;
             setkapairpval=setkaglobpval=TRUE;  break;
        case V_LISTBEST:
             if (type[0]==0 || flag_yes(type))  listbest(TRUE);
             else if (flag_no(type))  listbest(FALSE);
             else goto verb_err;  break;
        case V_LISTPERPAIR:
             { int xx;
               if (type[0]==0 || flag_yes(type))  listpairs=TRUE;
               else if (flag_no(type))  listpairs=FALSE;
               else if (sscanf(type,"%d",&xx)==1)
                { if (xx>0)  { listperpair=xx;  listpairs=TRUE; }
                  else if (xx==0)  listpairs=FALSE;
                  else goto verb_neg; }
               break; }
        case V_NODUMP:
             clear_alldumps();  saydump=TRUE;
             if (flag_no(type))  dumpfrags=TRUE;
             break;
        case V_NO_OUTPUT:
             havelog=FALSE;  clear_alldumps();  saydump=TRUE;
             if (flag_no(type))
              { havelog=TRUE;  dumpfrags=TRUE; }
             break;
        case V_DUMPALL:
             saydump=TRUE;
             if (type[0]==0 || flag_yes(type))
               dumpfrags=dumptab=dumpdif=dumpjseqs=dumpplot=TRUE;
             else if (flag_no(type))
              { clear_alldumps();  dumpfrags=TRUE; }
             else goto verb_err;  break;
        case V_ALLOUTER:
             outerseqfrags=TRUE;
             if (type[0]==0 || flag_yes(type))
               outerpairfrags=outergrpfrags=TRUE;
             else if (flag_no(type))
               outerpairfrags=outergrpfrags=FALSE;
             else goto verb_err;  break;
        case V_NOOUTER:
             outerseqfrags=outerpairfrags=outergrpfrags=FALSE;
             if (type[0]==0 || flag_yes(type))  ;
             else if (flag_no(type))  outerseqfrags=TRUE;
             else goto verb_err;  break;
        case V_INDELSKIP:
             if (type[0]==0 || flag_yes(type))  treat_dels=TD_SKIP;
             else if (flag_no(type))  treat_dels=TD_BLOCS;
             else goto verb_err;  break;
        case V_INDELSITE:
             if (type[0]==0 || flag_yes(type))  treat_dels=TD_SITES;
             else if (flag_no(type))  treat_dels=TD_BLOCS;
             else goto verb_err;  break;
        case V_INDELBLOC:
             if (type[0]==0 || flag_yes(type))  treat_dels=TD_BLOCS;
             else if (flag_no(type))  treat_dels=TD_SKIP;
             else goto verb_err;  break;
        case V_FANCY:
             if (type[0]==0 || flag_yes(type))  do_fancy(TRUE);
             else if (flag_no(type))  do_fancy(FALSE);
             else goto verb_err;  break;
        case V_READFUNV:
             if (v->targ) ((RFUNV)(v->targ))();  break;
        case V_READFUNCP:
             if (v->targ && ((RFUNCP)(v->targ))(type)==RET_ERROR)
                goto verb_err;  break;
        case V_READFUNCCFP:
             if (v->targ) ((RFUNCCFP)(v->targ))(type,cf);  break;
        case V_READGROUP:
             if (options_env) goto verb_options;
             read_group(type,cf);  break;
        case V_READSEQPAIRS:
             if (options_env) goto verb_options;
             if (v->targ)
              { if (haveq && type[0]==0)
                   kill_seqpairlist((SEQPAIRLISTP)(v->targ),FALSE);
                else read_seqpair((SEQPAIRLISTP)(v->targ),
                   verb, type, cf); }
             break;
        case V_SEQLIST:
             if (options_env) goto verb_options;
             if (v->targ)
              { kill_strlist((STRLISTP)(v->targ));
                /* You can clear `v->targ' by entering  -seq_list=  */
                goto add_seqlist; }
             break;
        case V_ADD_SEQLIST:
        add_seqlist:
             if (options_env) goto verb_options;
             if (v->targ)
              { if (haveq && type[0]==0)  break;
                read_seqlist((STRLISTP)(v->targ), verb, type, cf); }
             break;
        case V_SETRANGE:
             if (v->targ && set_onerange((STRBUFFP)(v->targ), type))
               goto verb_err;  break;
        case V_SETPATH:
             if (v->targ)
              { kill_strlist((STRLISTP)(v->targ));
                /* Clear by entering  -filepath=  */
                if (haveq && type[0]==0)  break;
                goto add_setpath; }
             break;
        case V_ADD_SETPATH:
        add_setpath:
             if (v->targ)
              { if (haveq && type[0]==0)  break;
                if (parse_addpath((STRLISTP)(v->targ), type))
                  goto verb_err; }
             break;
             }}

   /* Couldn't find a match */
   if (matched==FALSE)
     cfg_err("UNRECOGNIZED COMMAND", NULL, verb, type, cf); }



/* Routine to read single and multivalued flags */

/* Options:  -flag=off  or  -flag=no  or  -flag=false */
/*   means to CLEAR that flag, NOT set it. */
/*   -flag=on, -flag=yes, and -flag=true  are the same as  -flag  */

static int flag_yes(const char *type)
 { return (!strmatch(type,"on")
      || !strmatch(type,"yes") || !strmatch(type,"true")); }

static int flag_no(const char *type)
 { return (!strmatch(type,"off")
      || !strmatch(type,"no") || !strmatch(type,"false")); }

char set_vchar(const char *type)
 { if ( (type[0]=='"' || type[0]=='\'')
     && (type[1]!=0 && !isspace(type[1])) )  type++;
   if (flag_no(type) || type[0]=='0' || isspace(type[0]))  return 0;
   else  return type[0]; }


int read_flag (const char *type, int *vptr, int *newptr)
 { int val;
   if (type[0]==0 || flag_yes(type))  val=TRUE;
   else if (flag_no(type))  val=FALSE;
   else  return RET_ERROR;
   if (vptr) *vptr=val;
   if (newptr && val)  *newptr=TRUE;
   return RET_OK; }


int read_nflag (const char *type, int *vptr, int *newptr)
 { int val;
   if (type[0]==0 || flag_yes(type))  val=FALSE;
   else if (flag_no(type))  val=TRUE;
   else  return RET_ERROR;
   if (vptr) *vptr=val;
   if (newptr && val==FALSE)  *newptr=TRUE;
   return RET_OK; }


int read_sqtype(const char *type)
 { if (!strmatch(type,"auto"))  seqtype=AUTO;
   else if (!strmatch(type,"nucl") || !strmatch(type,"all"))  seqtype=ALL;
   else if (!strmatch(type,"sil"))  seqtype=SILENT;
   else if (!strmatch(type,"prot"))  seqtype=PROTEIN;
   else  return RET_ERROR;
   return RET_OK; }


char *say_sqtype(SEQTYPE type)
 { switch(type) {
     default:      return "UNKNOWN SEQTYPE";
     case NOTYPE:  return "NO TYPE";
     case AUTO:    return "AUTOMATIC";
     case ALL:     return "NUCLEOTIDES (ALL)";
     case SILENT:  return "SILENT SITES";
     case PROTEIN: return "PROTEINS"; }}


char *say_indels(TREAT_DELS tdel)
 { switch(tdel) {
     default:  return "?????";
     case TD_BLOCS:  return "one polymorphism";
     case TD_SKIP:  return "skipped";
     case TD_SITES: return "sites"; }}


static char *showcfname(CFGFILE *cf)
 { return (cf) ? cf->cfullname : "Command line"; }


static void cfg_err(const char *bleep, const char *bleep2,
      const char *verb, const char *type, CFGFILE *cf)
 { char bbuff[800], sbuff[66], q1b[MAXNAMEBUFF], q2b[MAXNAMEBUFF];
   int len;
   if (type==NULL || type[0]==0)
     len=sprintf(bbuff, "%s -%s !", bleep, doquote(q1b,verb));
   else
     len=sprintf(bbuff, "%s -%s=%s !", bleep,
       doquote(q1b,verb), doquote(q2b,type));
   if (bleep2!=NULL && *bleep2!=0)
     sprintf(bbuff+len, "\n%s", bleep2);

   if (cf==NULL)  xformout("%s\n(Command line)", bbuff);
   else xformout("%s\n(Config file `%s')",
     bbuff, shortname(sbuff,cf->cfullname)); }


static void badverb(const char *bleep,
    const char *verb, const char *type, CFGFILE *cf)
 { cfg_err("ILLEGAL COMMAND", bleep, verb, type, cf); }


static int set_onerange (STRBUFFP ap, const char *type)
 { if (ap->s) free(ap->s);  memset(ap,0,sizeof(STRBUFF));
   ap->s=strxdup(type);
   if (sscanf(ap->s, "%d-%d", &(ap->n), &(ap->nx)) < 2
        || ap->nx < ap->n || ap->n<0)
          { kill_strbuff(ap);  return RET_ERROR; }
   return RET_OK; }





static void check_addpair(SEQPAIRLISTP sp, char *tbuff)
 { char *seq2;  int i, nsp=sp->n;
   SEQPAIR tp;  memset(&tp,0,sizeof(tp));
   parse_unquote_sep(tbuff,&seq2,",",NULL);
   /* Don't add it if you have it already */
   for (i=0; i<nsp; i++)
     if (!striucmp(tbuff,sp->list[i].sq1)
       && !striucmp(seq2,sp->list[i].sq2))  return;
   tp.sq1=strxdup(tbuff);
   tp.sq2 = (seq2==NULL) ? "" : strxdup(seq2);
   if (sp->n>=sp->nx)
     sp->list=rexxalloc(sp->list,sizeof(SEQPAIR),
        sp->n, sp->nx+=20);
   sp->list[(sp->n)++]=tp; }




/* On the command line, type is a SEMICOLON-separated list of */
/*   COMMA-SEPARATED  sequence pairs */

static void read_seqpair(SEQPAIRLISTP sp,
   const char *verb, const char *type, CFGFILE *cf)
 { char *p;  int haveseqs=FALSE;
   /* Syntax on command line: -verb=N1,N2;N3;N4,N5   in one word */
   /* The words may contain embedded spaces and embedded punctuation, */
   /*   but the operating system may remove the quotes. */
   if (cf==NULL)
    { if (type!=NULL && type[0]!=0)
       { char *nextp, *org_tbuff, *tbuff;
         tbuff=org_tbuff=strxdup(type);
         do { SEQPAIR tp;  memset(&tp,0,sizeof(tp));
              tbuff=parse_unquote_sep(tbuff,&nextp,";",NULL);
              if (*tbuff==0)  break;
              check_addpair(sp,tbuff);  haveseqs=TRUE; }
         while ( (tbuff=nextp)!=NULL);
         free(org_tbuff); }}
   /* Syntax in config file: -verb N1,N2  N3  N4,N5   */
   /*   ending with a word beginning with - or / or a config file */
   /* Optional alternative: -verb=N1,N2  N3  N4,N4   */
   /*   in which case  type=N1,N2   */
   else
    { if (type && type[0])  pushtokcfg(cf,type);
      while ( (p=readtokcfg(cf))!=NULL)
       { if (showcftoks)
            printf ("  READ_SEQPAIR %s (%s)\n", p, showcfname(cf));
         if (p[0]=='-' || p[0]=='/' || isextent(p,"cfg"))
          { pushtokcfg(cf,p);  break; }
         /* Ignore duplicate sequence names. */
         else  { check_addpair(sp, p);  haveseqs=TRUE; }}}
   if (haveseqs==FALSE)
     badverb ("Empty sequence list !", verb, type, cf); }



static void read_seqlist(STRLISTP ap,
   const char *verb, const char *type, CFGFILE *cf)
 { char *p;  int haveseqs=FALSE;
   /* Syntax on command line: -verb=word1,word2,word3   in one word */
   /* The words may contain embedded spaces and embedded quoted commas, */
   /*   but the operating system may remove the quotes. */
   if (cf==NULL)
    { char *strin, *rs, *nextp;
      if (type!=NULL && type[0]!=0)
       { rs=strin=strxdup(type);
         do { rs=parse_unquote_sep(rs,&nextp,",",NULL);
              if (*rs==0)  break;
              check_addseq(ap,rs);  haveseqs=TRUE; }
         while ( (rs=nextp)!=NULL);
         free(strin); }}
   else
    { /* Syntaxes supported in a config file: */
      /*   -seql=word1 word2 word3  */
      /*   -seql word1 word2 word3  */
      /*   -seql = word1 word2 word3  */
      /*   -seql =word1 word2 word3  */
      /* Not permitted within an options environment */
      /* (parse_verbose() gives a fatal error message.) */
      /* The list is terminated by any word beginning with - or /, */
      /*   a config file, or the end of the configuration file. */
      /* In the first syntax, type==word1 */
      if (type && type[0])  pushtokcfg(cf,type);
      while ( (p=readtokcfg(cf))!=NULL)
       { if (showcftoks)
            printf ("  READ_SEQLIST %s (%s)\n", p, showcfname(cf));
         while (p[0]=='=') p++;  if (p[0]==0)  continue;
         if (p[0]=='-' || p[0]=='/' || isextent(p,"cfg"))
          { pushtokcfg(cf,p);  break; }
         /* Ignore duplicate sequence names. */
         else  { check_addseq(ap, p);  haveseqs=TRUE; }}}
   if (haveseqs==FALSE)
     badverb ("Empty sequence list !", verb, type, cf); }




GROUPLIST *grouplist;
int ngroups;

typedef struct { char *grname;
                 STRLIST seqlist; }  PERM_GROUP;

static PERM_GROUP *perm_group;
static int perm_nxgrp;
int perm_ngrp;


static void unspace_permgroup(void)
 { if (perm_group)
    { int i;
      for (i=0; i<perm_ngrp; i++)
       { unspace_str(perm_group[i].grname);
         unspace_slist2(perm_group[i].seqlist); }}}


void kill_groups(void)
 { if (ngroups>0)
    { int i,j,ns;
      for (i=0; i<ngroups; i++)
       { free(grouplist[i].gname);
         free(grouplist[i].grname);
         if ( (ns=grouplist[i].ns)>0)
          { SEQNAME *sq=grouplist[i].sqn;
            for (j=0; j<ns; j++)
             { free(sq[j].name);  free(sq[j].rname); }}}
      free(grouplist); }
   grouplist=NULL;  ngroups=0; }


void init_grouplist(void)
 { int i,j,ns;
   kill_groups();
   ngroups=perm_ngrp;
   grouplist=cxalloc(ngroups+3, sizeof(grouplist[0]));
   for (i=0; i<ngroups; i++)
    { char *grname=perm_group[i].grname;
      /* For group names, no initial character if all digits */
      grouplist[i].gname=gettag(grname,&grouplist[i].grname,NULL,0);
      /* Leave dostat==FALSE for ``GenPolyOnly'' sequences */
      if (!strmatch(grname,GENPOLYONLY))
         grouplist[i].gen_poly_only=TRUE;
      else  grouplist[i].dostat=TRUE;
      if ( (ns=grouplist[i].ns=perm_group[i].seqlist.n)>0)
       { char **inseq=perm_group[i].seqlist.s;
         SEQNAME *sq=grouplist[i].sqn=cxalloc(ns+3,sizeof(sq[0]));
         /* Initialize at orgoffs==-1, since the sequence may not */
         /*   exist in the Seqfile */
         for (j=0; j<ns; j++)
          { sq[j].name=gettag(inseq[j],&sq[j].rname,NULL,'S');
            sq[j].orgoffs=-1; }}}}




static PERM_GROUP *open_perm_group(const char *rname)
 { int i;  PERM_GROUP rr;
   for (i=0; i<perm_ngrp; i++)
     if (!striucmp(rname,perm_group[i].grname))  return &perm_group[i];
   /* Initiate a new group */
   memset(&rr,0,sizeof(rr));
   rr.grname=strxdup(rname);
   return addlist(&rr,sizeof(PERM_GROUP),
             (void **)&perm_group, &perm_ngrp,&perm_nxgrp); }



static void read_group(const char *type, CFGFILE *cf)
 { char *p,*q;  PERM_GROUP *pgrp=NULL;
   if (cf==NULL)
     xformout("-Group CANNOT BE USED on the command line !\n"
       "(groups MUST BE DEFINED within a configuration file.)");
   /* We must now have grouptype==G_GROUPS */
   grouptype=G_GROUPS;

   /* Syntaxes sypported: */
   /*   -group GNAME=name1 name2 name3 */
   /*   -group GNAME: name1 name2 name3 */
   /*   -group GNAME name1 name2 name3 */
   /* with each list terminated by /, -, or a config file. */
   /* If -group GNAME=name1, then type=name1  */

   if (type && type[0])  pushtokcfg(cf,type);

   while ( (p=readtokcfg(cf))!=NULL)
    { if (showcftoks)
         printf ("  READ_GROUP %s (%s)\n", p, showcfname(cf));
      if (p[0]=='-' || p[0]=='/' || isextent(p,"cfg"))
       { /* Exit the -group environment */
         pushtokcfg(cf,p);  break; }
      else if (pgrp==NULL)
       { /* No groupname defined yet */
         /* Remove trailing colons, as in */
         /*    -group=ALU: Seq1 Seq2 Seq3 /c  */
         /* Since readtokcfg unquotes as it reads, there is no way */
         /*   to distinguish  -group="ALU :": Seq1 Seq2 // */
         q=p+strlen(p);  while(q>p && *--q==':')  *q=0;
         if (*p==0)  badverb("Illegal group name !", "group", type, cf);
         pgrp=open_perm_group(p); }
      /* Store in that group (in parsed form), */
      /*   but ignore duplicate names. */
      else  check_addseq(&pgrp->seqlist, p); }
   if (pgrp==NULL)
     badverb("GROUP COMMAND without a group name !",
       "group", type,cf);
   if (pgrp->seqlist.n==0)
    { char ebuff[44], gbuff[33];
      sprintf(ebuff, "Empty group `%s' !",
         strncpy(gbuff,pgrp->grname,22));
      badverb(ebuff, "group", type,cf);  }}


static void read_options(const char *type, CFGFILE *cf)
 { char *p;
   if (cf==NULL)
     xformout("\nThe command -options CANNOT BE USED "
        "on the command line !\n"
        "(Enter the options directly.)");

   /* In case -options=maxSimPwPval=0.05 .... */
   if (type && type[0])  pushtokcfg(cf,type);

   while ( (p=readtokcfg(cf))!=NULL)
    { if (showcftoks)
         printf ("  READ_OPTIONS %s (%s)\n", p, showcfname(cf));
      /* Ignore redundant initial -s */
      while (*p=='-')  p++;  if (*p==0) continue;
      if (*p=='/')  { parsecharflags(p);  continue; }
      /* Don't assume that ALL words ending with `.cfg' */
      /*   are the names of config files !! */
      /* (It might be `Config=myfile.cfg', which needs to have */
      /*    the `Config=' parsed off in `parse_verbose()'.) */
      else if (strchr(p,'=')==NULL && isextent(p,"cfg"))
       { read_config(p,cf);  continue; }
      else if (!strmatch(p,"endoption"))  break;
      /* Ignore nested -options statements */
      else if (!strmatch(p,"option"))  continue;
      parse_verbose(p,cf,TRUE); }}


/* Collect all the dump options together */
int anydump(void)
 { return (dumpfrags || dumptab || dumpdif || dumpjseqs || dumpplot); }

void clear_alldumps(void)
 { dumpfrags=dumptab=dumpdif=dumpjseqs=dumpplot=FALSE; }




/* Functions for decoding short-form options: */

static void badout (const char *opts, const char *bleep)
 { xformout("\nIMPROPER COMMAND-LINE FORMAT IN /%s !%s",
      opts, bleep); }

static void badval2(char *opts)
 { badout(opts,""); }

static void badval2pos(char *opts)
 { badout(opts,"\nA number >= 1 is required here."); }

static void badval2neg(char *opts)
 { badout(opts,"\nA number >= 0 is required here."); }

static void badval(char c)
 { char opp[3];  opp[0]=c;  opp[1]=0;
   badval2(opp); }

static void badvalneg(char c)
 { char opp[3];  opp[0]=c;  opp[1]=0;
   badval2neg(opp); }


/* Set *iptr=<num>  for  /<char><num>  for various types of <num> */
/* Readint and readintverb  crash if <num> < 0 */
/* Readintverbpos crashes if  <num> <= 0 */

static char *readint(char *p, int *iptr)
 { int n;
   if (sscanf(p+1,"%d",&n)<1)  badval(*p);
   else if (n<0)  badvalneg(*p);
   *iptr=n;
   /* Stop while pointing at the last digit */
   while (isdigit(p[1]))  p++;
   return  p; }


static char *readintverb(char *p, int *iptr, char *verb)
 { int n;
   if (sscanf(p+1,"%d",&n)<1)  badval2(verb);
   else if (n<0)  badval2neg(verb);
   *iptr=n;
   /* Stop while pointing at the last digit */
   while (isdigit(p[1]))  p++;
   return  p; }


static char *readintverbpos(char *p, int *iptr, char *verb)
 { int n;
   if (sscanf(p+1,"%d",&n)<1)  badval2(verb);
   else if (n<=0)  badval2pos(verb);
   *iptr=n;
   /* Stop while pointing at the last digit */
   while (isdigit(p[1]))  p++;
   return  p; }


static char *readulong(char *p, unsigned long *ulptr)
 { char c;  unsigned long uu=0L;
   if (!isdigit(p[1]))  badval(p[0]);
   while ( (c=*++p)!=0 && isdigit(c))
     uu = 10L*uu + (c-'0');
   *ulptr=uu;
   return  --p; }


/* Entered with (for example) p=`c0.05aosp' */

/* Allow only the syntaxes  [n1[.<n2>]] or [n1[.<n2>]][<eE>n3], */
/*   where n1 OR n2 is mandatory,  and n3 is mandatory if <eE> is used */
/*   n1 and n2 must be nonnegative.  n3 can be negative */

static char *readdub(char *p, double *dubptr, char *verb)
 { int pflag,ndig;  char *orgp=p;
   if (p[1]=='-')  badval2neg(verb);
   else if (!isdigit(p[1]) && p[1]!='.')  badval2(verb);
   p++;  ndig=0;  pflag=TRUE;
   /* Find maximal string of digits with at most one . */
   while (isdigit(*p) || (pflag && *p=='.'))
    { if (*p=='.')  pflag=FALSE;  else ndig++;  p++; }
   /* No digits, or no digits before `e' ? */
   if (ndig==0)  badval2(verb);
   if (tolower(*p)=='e')
    { p++;  ndig=0;  if (*p=='-')  p++;
      while (isdigit(*p))  { p++;  ndig++; }
      if (ndig==0)  badval2(verb); } /* No digits found after `e' */
   if (sscanf(orgp+1, "%lf", (double *)dubptr)<1)
      badval2(verb);
   return --p; }


/* Additional checks for a P-value */
/* Entered with (for example) p=`c0.05aosp' */
/* Similar checks are done in `verblist' for the message V_PVAL */

static char *readpval(char *p, DUBSTRUC a, char *verb)
 { if (a.dptr && a.havptr)
    { double pval=12.0;
      if (a.setptr) *(a.setptr)=TRUE;
      p=readdub(p,&pval,verb);
      setpvalp(a, pval); }
   return p; }



/* Decode a string of single-character flags */

/* Initial states are SEQTYPE seqtype=NONE and */
/*    TREAT_DELS treat_dels=TD_BLOCS; */

static void parsecharflags(char *p)
 { char c;
   while ( (c=toupper(*++p))!=0)  switch (c)  {
       default:   xformout("\nUNRECOGNIZED FLAG /%c !", c);
       case '/':  break;  /* Treat  /ao/v2 as /aov2  */
       case 'R':  seqtype = SILENT;   break;
       case 'A':  seqtype = ALL;      break;
       case 'P':  seqtype = PROTEIN;  break;

       case 'C':  circular=TRUE;  break;
       /* `Degeneracy' submenu */
       case 'D':  switch (c=toupper(*++p)) {
                    default: badval('D');  break;
                    case 'A':  nodegenflag = TRUE;  break;
                    case 'C':  nodegenflag = FALSE;  break;
                    case '3':  d3flag = TRUE;  break;
                    case 'S':  simpflag = TRUE;  break;
                    case 'M':  nosimpflag = TRUE;  break;
                    case 'V':  mitcodes = TRUE;  break; }  break;
       /* Customizing output: */
       case 'O':  switch (c=toupper(*++p)) {
                    default: badval('O');  break;
                    case 'S':  outerseqfrags=TRUE;  break;
                    case 'P':  outerpairfrags=TRUE;  break;
                    case 'G':  outergrpfrags=TRUE;  break;
                    case 'A':
                      outerseqfrags=outerpairfrags=outergrpfrags=TRUE;
                      break;
                    case 'I':  outersims=TRUE;  break;
                    case 'N':  switch(c=toupper(*++p))  {
                       default: badval2("ON");  break;
                       case 'S':  outerseqfrags=FALSE;  break;
                       case 'P':  outerpairfrags=FALSE;  break;
                       case 'G':  outergrpfrags=FALSE;  break;
                       case 'A':
                         outerseqfrags=outerpairfrags=outergrpfrags=FALSE;
                         break;
                       case 'I':  outersims=FALSE;  break; }
                          break; }  break;
       case 'S':  switch (c=toupper(*++p)) {
                    default: badval('S');  break;
                    case 'A':  sortgfragsboth=TRUE;  break;
                    case 'B':  showblast=TRUE;  break;
                    case 'D':  showfragdata=TRUE;  break;
                    case 'G':  sortgfragsbyseq=TRUE;  break;
                    case 'N':  sortnames=TRUE;  break;
                    case 'P':  showpolyflag=TRUE;  break;  }  break;
       case 'Q':  switch (c=toupper(*++p)) {
                    default: badval('Q');  break;
                    case 'B':  use_underscores=FALSE;  break;
                    case 'U':  use_underscores=TRUE;
                        break; }  break;
       /* `Fancy' output */
       case 'F':  do_fancy(TRUE);  break;
       /* Set -gscale and other parameters */
       case 'G':  p=readint(p,&gscale);  break;
       case 'N':  p=readint(p,&numsims);  break;
       case 'I':  p=readint(p,&irrupt);  break;
       /* Control how many fragments are listed */
       case 'L':  switch (c=toupper(*++p)) {
                    default: badval('L');  break;
                    case 'B':  listbest(TRUE);  break;
                    case 'G':  p=readintverb(p,&listglobal,"LG");
                               newlglobal=TRUE;  break;
                    case 'P':  if (isdigit(p[1]))
                                { int xx;
                                  p=readintverb(p,&xx,"LP");
                                  if (xx>0)
                                   { listperpair=xx;  listpairs=TRUE; }
                                  else listpairs=FALSE; }
                               else listpairs=TRUE;
                               break; }
                      break;
       /* Maximum (restricting fragments) submenu */
       case 'M':  switch (c=toupper(*++p)) {
                    default: badval('M');  break;
                    case 'I':
                      switch (c=toupper(*++p)) {
                        default: badval2("MI");  break;
                        case 'P':  p=readpval(p,simpairpval,"MIP");  break;
                        case 'G':  p=readpval(p,simglobpval,"MIG");  break;
                        case 'Z':  havsimpairpval=havsimglobpval=FALSE;
                            break; }  break;
                    case 'K':
                      switch (c=toupper(*++p)) {
                        default: badval2("MK");  break;
                        case 'P':  p=readpval(p,kapairpval,"MKP");  break;
                        case 'G':  p=readpval(p,kaglobpval,"MKG");  break;
                        case 'Z':  havkapairpval=havkaglobpval=FALSE;
                              setkapairpval=setkaglobpval=TRUE;  break; }
                              break;
                    case 'L':  p=readintverbpos(p,&minlength,"ML");  break;
                    case 'N':  p=readintverbpos(p,&minnpoly,"MN");  break;
                    case 'S':  p=readintverbpos(p,&minscore,"MS");  break;
                    case 'Z':  nomaxpval();  break; }
                      break;
       case 'B':  if (toupper(p[1])=='H') { homolflag=TRUE;  p++; }
                  /* Allows -b0 or -bh0 to turn off blocs */
                  p=readint(p,&blocsize);  break;
       case 'H':  homolflag=TRUE;  break;

       case 'V':  p=readint(p,&maxoverlap);  break;
       case 'E':  switch (c=toupper(*++p)) {
                     default:  badval('E');
                     case 'B': case 'R':  treat_dels=TD_BLOCS;  break;
                     case 'S':  treat_dels=TD_SKIP;  break; }
                  break;
       case 'X':  hold_window = TRUE;  break;
       case 'Z':  logappend = TRUE;  break;
       case 'W':  p=readulong(p,&wseed);  break;
                  /* This overwrites the buffer at p+1 and */
                  /*   returns a pointer to the NEXT string in p+1 */
       case '$':  commchars=strxdup(parse_unquote(p+1,NULL,NULL));
                  p="A";  break;
       case '!':  if (p[1]) while(strchr("\"'",p[1]))  p++;
                  if (p[1]==0)  badval('!');
                  else if (p[1]=='0')  matchchar=0;
                  else matchchar=p[1];  p++;
                  if (p[1]) while(strchr("\"'",p[1]))  p++;
                  break;
       case '*':  showcftoks=TRUE;  break;
       case '@':  ShowNexusToks=TRUE;  break;
       case '#':  showlinesflag=TRUE;  break;
       case '+':  showtabfrags=TRUE;  break;
                   }}



/* Check the flag settings for contradictory combinations. */

void checkflags(void)
 {
   if (bcsims && numsims==0)
    { WarnHandler (
      "If there are no simulations, you MUST USE BLAST-like scores\n"
      "  for global P-values !!\n");  bcsims=FALSE; }

   if (d3flag && seqtype!=SILENT)
    { warnprintf("THE FLAG `D3' HAS NO EFFECT unless "
     "`R' (SILENT) is set.\n(Flag ignored.)");  d3flag=FALSE; }

   if (dumpjseqs && jseqlen<2)
     xformout (
     "\nJunction sequence length `Jseq_len' (set=%d) must be AT LEAST 2 !\n"
       "  if `dumpjseqs' is set.",  jseqlen);

   if (blocsize==1)
    { warnprintf("Blocsize=1 is the same as NO blocks!\n");
      blocsize=0; }
   if (blocsize>0 && grouptype==G_GROUPS)
    { WarnHandler (
      "WARNING: -group and blocsize=%d both entered!  (BLOCSIZE IGNORED.)\n"
      "ENTER JUST ONE of -group (entries) or -blocs=<num> (/b<num)\n",
         blocsize);  blocsize=0;  homolflag=FALSE; }
   if (blocsize==0 && homolflag)
    { warnprintf("The flag `H' HAS NO EFFECT if there are NO BLOCKS!\n");
      homolflag=FALSE; }
   if (outergrpfrags && grouptype!=G_GROUPS && blocsize==0)
     outergrpfrags=FALSE;

   if (simpflag && nosimpflag)
     xformout (
     "\nYOU CANNOT EXCLUDE BOTH singly AND multiply polymorphic sites !\n"
     "Is that what you intended?");
   if (mitcodes && seqtype!=SILENT)
    { warnprintf("Mitochondrial genetic codes ARE NOT USED"
          " unless `R' (SILENT) is set\n(Flag ignored.)\n");
      mitcodes=FALSE; }

   if (!(innerpairfrags || outerpairfrags
          || outerseqfrags || outergrpfrags))
     xformout (
       "\nNO FRAGMENT TYPES were selected to analyze!\n"
       "Nothing to do!");

   #ifdef VSHOW
   if (havminscale)
    { if (vertminscale<0.0)
        xformout("Can't have vertminscale=%g < 0.0 !\n", vertminscale);
      if (havmaxscale && vertmaxscale<vertminscale+1e-8)
        xformout("Can't have vertminscale=%g  vertmaxscale=%g!\n",
           vertminscale, vertmaxscale);
      if (havstepscale && vertstepscale<1e-8)
        xformout("Can't have vertstepscale=%g!\n", vertstepscale); }
   else if (havmaxscale || havstepscale)
    { WarnHandler ("CAN'T HAVE vertstepscale or vertmaxscale "
         "WITHOUT vertminscale!\n");
      havmaxscale=havstepscale=FALSE; }
   #endif
          }


void gscale_comment(void)
 { if (gscale==0)
     textprintf ("No mismatches in fragments (`gscale'=0)\n");
   else
    { textprintf ("Mismatch penalty: %d * npolys / num_ppen\n", gscale);
      textprintf (
       "  where `num_ppen' is the number of ``potential penalties'':\n");
      textprintf ("    sequence differences for inner fragments,\n");
      textprintf ("    local monomorphisms for outer fragments.\n"); }}




/* (***BBB) Readconfig() and friends */

/* Several CONFIG files may be open simultaneously,  */
/*   since a CONFIG file can call another CONFIG file.  */
/*   Thus the functions                         */
/*                                                           */
/*     read_config(),  rcfget(), readtokcfg(), pushtokcfg(), ...   */
/*                                                           */
/*   must be completely reentrant */

static void read_config(const char *fn, CFGFILE *cfpar)
 { char *rs, *cfmagic=CONFIG_MAGIC;
   CFGFILE cc, *cfp;

   if (showcftoks)
     printf ("  READ_CONFIG filename=%s (%s)\n", fn, showcfname(cfpar));

   /* Add the default extension  .cfg   if necessary, */
   /*   and allocate a new copy of the file name */
   if (!isextent(fn,"cfg"))
    { char *newfn=mxalloc(strlen(fn)+5);
      sprintf(newfn, "%s.cfg", fn);  fn=newfn; }
   else fn=strxdup(fn);

   /* Is this CFG file being called by a recursive copy of itself? */
   /* If so, abort to avoid an infinite loop. */
   err_clear=TRUE;  err_clear_funp=kill_cfgs;
   if (cfpar!=NULL)  check_cfg_recurse (cfpar,fn);

   if (showcftoks)
     printf ("  READ_CONFIG filename=%s (%s)\n", fn, showcfname(cfpar));

   /* Initialize a CFGFILE structure. */
   memset(&cc,0,sizeof(cc));
   /* Open the file in binary mode for DOS, UNIX, or Mac text files, */
   cc.cff=fxopenread(fn,&cc.cfullname,&cc.fdir,&cc.fxdir,FALSE);

   /* Allocate a read buffer, etc. */
   cc.cfname=fn;  cc.lobuff=mxalloc(LOBVFFSIZE+3);
   cc.ord=cfglist.cnum+1;  cc.cfparent=cfpar;
   addstr(&cfgnames,fn);

   /* If the first token is not CONFIG_MAGIC, */
   /*   then this is not a config file after all. */
   if ( (rs=readfirtokcfg(&cc))==NULL || stricmp(rs,cfmagic))
    { char sbuff[66];  free(cc.lobuff);  cc.lobuff=NULL;
      ErrHandler ("%s IS NOT a CONFIG file!!!\n"
        "A CONFIG file MUST BEGIN with `%s'!\n"
        "Initial word found: `%s'\n", shortname(sbuff,fn), cfmagic,
         (rs) ? rs : "<NULL>"); }

   /* It's now safe to add cc to cfglist */
   cfp=addlist(&cc,sizeof(cc),
     (void **)&cfglist.cfg,&cfglist.cnum,&cfglist.cxnum);

   /* Read, save, and parse the tokens */
   /* Note that a token may be another CFG file. */
   while ( (rs=readtokcfg(cfp))!=NULL)  parseword(rs,cfp);

   /* We have now finished reading the file. */
   /* De-allocate the read buffer and close the file. */
   free(cfp->lobuff);  cfp->lobuff=NULL;
   fxclose(cfp->cff,fn);  cfp->cff=NULL;  }


/* Since several CONFIG files may be open simultaneously, */
/*   these two functions must be reentrant */

static int uncfget (int c, CFGFILE *cf)
 { return cf->savec=c; }

static int rcfget(CFGFILE *cf)
 { char ch;
   if (cf->savec>0)
    { ch=cf->savec;  cf->savec=-1;  return ch; }
   /* Cf is read as a binary file. Since strings of white space */
   /*  are taken as separators, including one or more \n or \r or */
   /*  combinations, DOS/UNIX/Mac text files are read correctly. */
   if (--(cf->nleft)<0)
    { cf->buffp=cf->lobuff;
      if ( (cf->nleft=fread(cf->lobuff,1,LOBVFFSIZE,cf->cff))==0)
        return -1;  cf->nleft--; }
   /* If find a \0 in the file, return \1 */
   return ( (ch=*(cf->buffp)++)!=0)  ? ch : 1; }


static char *readfirtokcfg(CFGFILE *cf)
 { static char firtokbuff[CONFIG_MAGIC_BUFFLEN+5];  int c,n;
   /* Skip to the next non-space character */
   while ( (c=rcfget(cf))>=0 && isspace(c))  ;
   if (c<0)
    { if (showcftoks)  printf ("  READFIRTOK=EOF (%s)\n",showcfname(cf));
      return NULL; }
   /* Read into `firtokbuff', without allowing for unquoting */
   /* The first character is guaranteed not to be a space */
   n=0;
   do { if (n<CONFIG_MAGIC_BUFFLEN)  firtokbuff[n++]=c;
        else { printf ("\nWARNING: FIRST CONFIGFILE INPUT WORD\n"
          "  (len<=%d) TRUNCATED to %d bytes.\n", n, CONFIG_MAGIC_BUFFLEN);
               uncfget(c,cf);  break; }}
   while( (c=rcfget(cf))>=0 && !isspace(c));
   firtokbuff[n]=0;
   if (showcftoks)
     printf ("  READFIRTOK %s (%s)\n", firtokbuff,showcfname(cf));
   return firtokbuff; }


static char *readtokcfg(CFGFILE *cf)
 { static char tokbuff[MAXTOKSIZE+8];  int c,n,qc;
   if (cf->savtok)
    { strncpy(tokbuff,cf->savtok,MAXTOKSIZE+3);
      free(cf->savtok);  cf->savtok=NULL;
      if (showcftoks)
        printf ("  READTOK %s (%s,FETCHED)\n", tokbuff,showcfname(cf));
      return tokbuff; }
   /* Skip to the next non-space character */
   while ( (c=rcfget(cf))>=0 && isspace(c))  ;
   while (c=='%')     /* COMMENT: Ignore until EOL */
    { char *p,*q;
      n=0;  tokbuff[n++]=c;  /* Read into tokbuff[] for showcftoks only */
      /* EOL characters are NOT preconverted to \n */
      while ( (c=rcfget(cf))>=0 && c!='\n' && c!='\r')
         if (n<MAXTOKSIZE)  tokbuff[n++]=c;
      tokbuff[n]=0;
      if (showcftoks)
        printf ("  READTOK COMMENT %s (%s)\n",tokbuff,showcfname(cf));
      /* Test for Configfile distinguished comment */
      p=tokbuff;  while(*p=='%') p++;
      if (*p=='!')
       { while(*p=='!') p++;  while(isspace(*p)) p++;
         q=p+strlen(p);  while(q>p && isspace(*--q)) *q=0;
         if (*p!=0)  addstr(&configcomments,p); }
      /* If c>=0, then c=='\n' or '\r': */
      /*   skip to the next non-space character */
      if (c>=0)  while ( (c=rcfget(cf))>=0 && isspace(c))  ;  }
   if (c<0)
    { if (showcftoks)  printf ("  READTOK=EOF (%s)\n",showcfname(cf));
      return NULL; }
   /* Unquote a word or token if necessary, for example */
   /*   to allow for embedded spaces. */
   /* Within quotes, doubled quote-characters count as a single quote. */
   /* Begin with no `quotestate' */
   n=0;  qc=-1;
   do { if (c=='"' || c=='\'')
         { if (c==qc)   /* Perhaps ending the `quotefield' */
            { if ( (c=rcfget(cf))<0)  break;
              else if (c!=qc)  /* NOT a doubled quote character */
               { uncfget(c,cf);  qc=-1;  continue; }}  /* End quotefield */
           /* Fall through if c is a doubled quote character */
           /*   or if it is the other quote character */
           else if (qc==-1) { qc=c;  continue; }}  /* Begin match */
        /* End token at a nonquoted space; otherwise fall through */
        else if (qc==-1)
         { if (isspace(c))  break; }
        /* Within a quote field, end token at EOL */
        else if (c=='\n' || c=='\r')
         { printf ("\nWARNING (token so far=%s)\n"
          "  LINE END in the middle of a token within quotes!!!\n"
          "  Ending the token at this point.\n", tokbuff);  break; }
        /* Add character to the current token */
        if (n<MAXTOKSIZE)  tokbuff[n++]=c;
        else { printf ("\nWARNING: INPUT WORD "
          "(len>=%d) TRUNCATED to %d bytes.\n", n, MAXTOKSIZE);
               uncfget(c,cf);  break; }}
   while( (c=rcfget(cf))>=0);
   tokbuff[n++]=0;  tokbuff[n]=0;
   if (showcftoks)
     printf ("  READTOK %s (%s)\n", tokbuff,showcfname(cf));
   /* Save for later token display */
   addstr(&cf->tok, tokbuff);
   return tokbuff; }


static void pushtokcfg(CFGFILE *cf, const char *token)
 { cf->savtok=strxdup(token); }


void show_cargcs(const char *bleep, const char *spacestr, int fgap)
 { char *bbuff;
   if (bleep==NULL || bleep[0]==0)  /* Default */
     bleep = (commline_sayorg) ? "Original command line" : "Command line";
   bbuff = getstr("%s%s:  %s ", spacestr, bleep, ProgName);
   ftoklistf(bbuff, save_argv+1,save_argc-1, 1,NULL,fgap, TOK_QUOTE);
   free(bbuff);  commline_shown=TRUE; }


/* Use argc,argv, and cfglist_open[] to show the command line */

void show_commline(const char *bleep, int spaceflag)
 { char *spacestr = (spaceflag) ? " " : "";
   int cnum, fgap = (spaceflag) ? 3 : 2;
   show_cargcs(bleep,spacestr,fgap);
   /* Now the contents of any CONFIG files */
   if (cfglist_open.cfg!=NULL && (cnum=cfglist_open.cnum)>0)
    { CFGFILE *cfg=cfglist_open.cfg;
      char *bbuff;  int i,n, mx;
      textprintf ("%sContents of parameter configuration file%s:\n",
        spacestr,  (cnum>1) ? "s (in the order read)" : "");
      /* Find the length of the longest configuration file name */
      for (mx=i=0; i<cnum; i++)
       { int len=(cfg[i].fdir==NULL) ? 0 : strlen(cfg[i].fdir);
         if ( (n=strlen(cfg[i].cfname)+len) > mx)  mx=n; }
      bbuff=mxalloc(mx+50);
      for (i=0; i<cnum; i++)
       { const char *cname=cfg[i].cfname;
         if (cfg[i].fxdir && cfg[i].fdir!=NULL)  cname=cfg[i].fdir;
         if (cfg[i].fxdir || (cfg[i].fdir==NULL && filepath.s==NULL))
           sprintf(bbuff, "%s%s: ", spacestr, cname);
         else
           sprintf(bbuff, "%s%s (%s): ", spacestr, cname,
             (cfg[i].fdir) ? cfg[i].fdir : "default directory");
         ftoklistf(bbuff, cfg[i].tok.s,cfg[i].tok.n,
            2,NULL,fgap, TOK_QUOTE | TOK_GROUPS); }
      free(bbuff); }
   show_testlist();
   show_onlylist(); }





STRBUFF rec_buff;

static char *cfg_recurse_err (CFGFILE *cfanc,
         CFGFILE *cfpar, int n, const char *fn)
 { int i,k,mx,fnlen;  char dbuff[55], abuff[55];
   CFGFILE *cf, *cfg=cfglist.cfg;  int cnum=cfglist.cnum;
   if (showcftoks) nl();
   mx=fnlen=strlen(fn);
   for (i=0; i<cnum; i++)
     if ( (k=strlen(cfg[i].cfname))>mx)  mx=k;
   if (n==0)  strcpy(dbuff,"directly");
   else sprintf(dbuff, "(with %s)", numsay(abuff,n,"intermediary",FALSE));
   rec_buff.n=0;
   addprintf(&rec_buff, 2*fnlen+strlen(cfanc->cfname)+250,
      "THE CONFIGURATION FILE `%s' HAS RECURSIVELY CALLED ITSELF!\n"
      "  CFG#%d: %s  %s called  CFG#%d: %s\n"
      "Config files seen so far (*=open):\n",
         fn,cfanc->ord,cfanc->cfname, dbuff, cnum+1,fn);
   showpath(&rec_buff);
   for (i=0; i<cnum; i++)
    { addprintf(&rec_buff, strlen(cfg[i].cfname)+50,
        "%3d.%c %-*s", i+1, (cfg[i].cff) ? '*' : ' ',
        mx,cfg[i].cfname);
      if ( (cf=cfg[i].cfparent)!=NULL)
        addprintf(&rec_buff, strlen(cf->cfname)+50,
          "  Called by %s (#%d)", cf->cfname, cf->ord);
      addprintf(&rec_buff,0, "\n"); }
   addprintf(&rec_buff, fnlen+strlen(cfpar->cfname)+50,
     "%3d.  %-*s  Called by %s (#%d)\n",
     cnum+1, mx,fn, cfpar->cfname, cfpar->ord);
   if (rec_buff.n > SBUFFSIZE)
    { sprintf(rec_buff.s+SBUFFSIZE, "...");
      rec_buff.n=SBUFFSIZE+3; }
   return rec_buff.s; }


static void check_cfg_recurse (CFGFILE *cfpar, const char *fn)
 { CFGFILE *cf=cfpar;
   if (cf!=NULL)
    { int n=0;
      if (showcftoks)
        printf ("  CHECK_CFG_RECURSE(%s): ", fn);
      do { if (showcftoks)  printf (" %s", cf->cfname);
           if (!stricmp(fn,cf->cfname))
             ErrHandler (cfg_recurse_err(cf,cfpar,n,fn));
           n++; }
      while ( (cf=cf->cfparent)!=NULL);
      if (showcftoks) nl(); }}


void kill_cfgs(void)
 { int i=cfglist.cnum;  CFGFILE *cfg=cfglist.cfg;
   if (rec_buff.s!=NULL) free(rec_buff.s);
   memset(&rec_buff,0,sizeof(rec_buff));
   while (--i>=0)
    { kill_strlist(&cfg[i].tok);
      if (cfg[i].cff)  fxclose(cfg[i].cff,cfg[i].cfname);
      free((char *)cfg[i].cfname);  free(cfg[i].cfullname);
      if (cfg[i].lobuff)  free(cfg[i].lobuff);
      if (cfg[i].savtok)  free(cfg[i].savtok);
      if (cfg[i].fdir)  free(cfg[i].fdir); }
   if (cfg) { free(cfg);  cfglist.cfg=NULL; }
   cfglist.cnum=cfglist.cxnum=0; }


