
/* General utilities for VSHOW and VTEXT
   All rights reserved - Stanley Sawyer

 */

#include "vall.h"
#include <time.h>
#include "vutil.h"
#ifndef UNIX
#include <dir.h>
#endif

char *AllRightsReserved_Vutil =
   "All rights reserved - Stanley Sawyer";

/* Textprintf is initialized at printf() or bprintf() */
/* Dumpx_open() in vdump.c  changes textprintf() to tab output */
/* Dumpx_close() in vdump.c  restores printf() or bprintf() */
/* Set_printf() in vtds.c  changes textprintf() between */
/*   bprintf() and logprintf()  */

PRINTFUNP textprintf = BPRINTF;
void (*textnl)(void) = BNL;
VOID_PRINTFUNP WarnHandler=warnprintf;
VOID_PRINTFUNP ErrHandler=exitprintf;

#define TOK_LINESIZE 72  /* For word-wrap package in vutil.c */

int tok_linesize=TOK_LINESIZE, dumpflag, notimer;


/* Simple string functions */

/* NOTE:  striucmp() does not distinguish spaces from underscores, */
/*   but all other string comparison functions do. */

char *makeuc(char *str)
 { char *p=str;
   if (p) while (*p) { *p = toupper(*p);  p++; }
   return str; }

char *makelc(char *str)
 { char *p=str;
   if (p) while (*p) { *p = tolower(*p);  p++; }
   return str; }

int stricmp(const char *p, const char *q)
 { if (p==NULL) p="";  if (q==NULL) q="";
   while (*p || *q)
    { if (tolower(*p)!=tolower(*q))
         return (tolower(*p)-tolower(*q));  p++; q++; }
   return 0; }

int striucmp(const char *p, const char *q)
 { if (p==NULL) p="";  if (q==NULL) q="";
   while (*p || *q)
    { if (tolower(*p)!=tolower(*q))
       { char c=tolower(*p), d=tolower(*q);
         if (isspace(c)) c='_';  if (isspace(d)) d='_';
         if (c!=d)  return (tolower(*p)-tolower(*q)); }
      p++; q++; }
   return 0; }

int strincmp(const char *p, const char *q, size_t n)
 { if (p==NULL) p="";  if (q==NULL) q="";
   /* n is normally unsigned, so we can't say --n>=0 */
   while (n-- >0 && (*p || *q))
    { if (tolower(*p)!=tolower(*q))
         return (tolower(*p)-tolower(*q));  p++; q++; }
   return 0; }

int strmatch(const char *p, const char *mat)
 { return strincmp(p,mat,strlen(mat)); }

int strlenx(const char *p)
 { return (p!=NULL) ? strlen(p) : 0; }

char *bool(int flag)
 { return (flag) ? "TRUE" : "FALSE"; }

char *boolon(int flag)
 { return (flag) ? "ON" : "OFF"; }


/* Some useful sorts */

int string_sort (const void *p, const void *q)
 { return
    strcmp(*(char **)p, *(char **)q); }

int sortint (const void *p, const void *q)
 { int x;
   if ( (x=*(int *)p - *(int *)q)!=0)  return (x>0) ? 1 : -1;
   return 0; }

int sortdub (const void *p, const void *q)
 { double xx;
   if ( (xx=*(double *)p - *(double *)q)!=0)  return (xx>0.0) ? 1 : -1;
   return 0; }




/* Display and parsing functions */

/* Unquote a word starting at p, parts of which may be within quotes, */
/*   using both " and ' as quote characters.           */
/* Terminate the string at the first space character */
/*   that is NOT within quotes. */

/* Overwrite the first part of the word with the unquoted word */
/*   and return a pointer to the inputted buffer */
/* Store a pointer to the next space-separated string at *nextp */
/*   (*nextp==NULL if this next string is 0) */
/* Return *openfp==TRUE iff the buffer ends with open quotes */

/* Use doubled characters to escape the quote characters themselves */

/* This parses   "$&""_5 7"        as   $&"_5 7        */
/*               "No Way"          as   No Way         */
/*               N"o w"a"y Ata"ll  as   No way Atall   */
/*               '''Z "X=Y"'       as   'Z "X=Y"       */

char *parse_unquote (char *p, char **nextp, int *openfp)
 { int c,qc=-1;  char *r,*oldp;
   if (p==NULL) { if (nextp) *nextp="";  return NULL; }
   /* Space ahead, but string will still be written at oldp */
   r=oldp=p;
   while (isspace(*p)) p++;
   while ( (c=*p++)!=0)
    { if (c=='"' || c=='\'')
       { if (qc==-1) { qc=c;  continue; }  /* Begin quoted text */
         else if (c==qc)  /* Was inside quoted text with this qc */
          { /* After the next command, p points to 2 characters later */
            if ( (c=*p++)==0)  { qc=-1;  break; }
            else if (c==qc) goto double_qc; /* Doubled qcs are escaped */
            else { qc=-1;  p--;  continue; }}}    /* End quoted text */
      /* End the string at a space character that is not */
      /*   in quoted text.  Otherwise, store the character. */
      else if (isspace(c) && qc==-1)  break;
 double_qc:
      *r++=c; }
   *r=0;   /* Terminate the return string */
   if (nextp!=NULL)          /* p points to the next token */
    { /* Set *nextp==NULL if no more strings, or if the buffer */
      /*   ended with open quotes. */
      if (c==0 || qc>=0)  *nextp=NULL;
      else
       { if (*p!=0) while(isspace(*p)) p++;
         *nextp = (*p) ? p : NULL; }}
   if (openfp!=NULL)  *openfp = (qc>=0) ? TRUE : FALSE;
   return oldp; }


/* Same as parse_unquote, but terminates the string at the next */
/*   character in sep[] that is not within quotes. */
/* Spaces are NOT separators if they are not contained in sep[] */

char *parse_unquote_sep (char *p, char **nextp, char *sep, int *openfp)
 { int c,qc=-1;  char *r,*oldp;
   if (p==NULL) { if (nextp) *nextp="";  return NULL; }
   /* Space ahead, but string will still be written at oldp */
   r=oldp=p;
   while (*p!=0 && strchr(sep,*p)!=NULL) p++;
   while ( (c=*p++)!=0)
    { if (c=='"' || c=='\'')
       { if (qc==-1) { qc=c;  continue; }  /* Begin quoted text */
         else if (c==qc)  /* Was inside quoted text with this qc */
          { /* After the next command, p points to 2 characters later */
            if ( (c=*p++)==0)  { qc=-1;  break; }
            else if (c==qc) goto double_qc; /* Doubled qcs are escaped */
            else { qc=-1;  p--;  continue; }}}    /* End quoted text */
      /* End the string at a sep character that is not */
      /*   in quoted text.  Otherwise, store the character. */
      else if (strchr(sep,c)!=NULL && qc==-1)  break;
 double_qc:
      *r++=c; }
   *r=0;   /* Terminate the return string */
   if (nextp!=NULL)          /* p points to the next token */
    { /* Set *nextp==NULL if no more strings, or if the buffer */
      /*   ended with open quotes. */
      if (c==0 || qc>=0)  *nextp=NULL;
      else
       { while (*p!=0 && strchr(sep,*p)!=NULL)  p++;
         *nextp = (*p) ? p : NULL; }}
   if (openfp!=NULL)  *openfp = (qc>=0) ? TRUE : FALSE;
   return oldp; }


/* Put `str' in quotes if it contains internal spaces or semicolons */
/* In addition, truncate the total length to MAXNAMELEN or less. */

char *doquote (char *buff, const char *str)
 { const char *p;  int k, qflag=FALSE;
   char *rsbuff, q1b[MAXNAMEBUFF];
   for (k=0;  str[k]!=0 && k<MAXNAMELEN; k++)
     if (isspace(str[k]) || str[k]==';')
       { qflag=TRUE;  break; }

   rsbuff = (buff==NULL) ? q1b : buff;
   if (qflag)
    { char c,*r=rsbuff;  int n=1;  *r++='"';  p=str;
      while ( (c=*p++)!=0 && ++n<MAXNAMELEN)
       { *r++=c;  if (c=='"') { *r++=c;  n++; }}
      sprintf(r, (n>=MAXNAMELEN) ? "...\"" : "\""); }
   else
    { strncpy(rsbuff,str,MAXNAMELEN+2);
      if (strlen(str)>MAXNAMELEN)
        sprintf(rsbuff+MAXNAMELEN, "..."); }
   return (buff==NULL) ? strxdup(q1b) : buff;
      }


int quotesclen (char *str)
 { int i,m,n, havesp=FALSE;
   for (m=n=i=0; str[i]!=0; i++)
    { n++;  if (str[i]=='"') m++;
      if (isspace(str[i]) || str[i]==';')  havesp=TRUE; }
   return  (havesp) ? m+n+2 : n; }



/* Set up to three names for each sequence: */
/* The calling name `rname' comes directly from either `getnxtoken()'
/*   or `readtokcfg()' or `parse_unquote()', and so is unquoted */

/* Strtag (returned)   - Initial S if all digits, no ctrl chars, */
/*    quoted if it contains internal spaces, truncated at MAXNAMELEN */
/*    (tail replaced by ...)  */
/*    The displayable version of the sequence name. */
/* Readtag (*rdtagptr) - Verbatim as read, although in most cases  */
/*    `gettag()' is called AFTER the name is unquoted.  */
/* Readshowtag (*rdshowptr) - Readable version of Readtag:  */
/*    Exactly the same as Readtag, except that              */
/*      (i) control characters are make printable and       */
/*     (ii) truncated at MAXNAMELEN (tail replaced by ...)  */

/* If use_underscore==TRUE, embedded spaces are changed to underscores */
/*   in all copies of the sequence name.  Sequence names are compared */
/*   using  `striucmp()', which does not distinguish beteen spaces */
/*   and underscores. */

extern int use_underscores, allctrlnum;

char *gettag(char *rname, char **rdtagptr, char **rdshowptr, char numchar)
 { char *strname, *str, *rn, *rdshow, *cbuff, *dbuff=NULL;
   int alldig=TRUE, havctl=FALSE;
   /* By default, convert all spaces to underscores */
   if (use_underscores)  unspace_str(rname);
   /* See if `rname' is all digits or has embedded control characters */
   /* Keep alldig=TRUE if nothing but digits, ctlr chars, and spaces */
   for (rn=rname; *rn; rn++)
    { if (*rn<0x20)  havctl=TRUE;
      else if (!isdigit(*rn) && !isspace(*rn) && *rn!='_')
        alldig=FALSE; }
   /* Open a buffer for the strtag name */
   strname=str=cbuff=mxalloc(strlen(rname)+3);
   /* Precede strname with S if all digits, control chars, and whitespace */
   if (alldig && numchar!=0)  *str++=numchar;
   /*  (i) Copy rname to strname but skip over control characters */
   /*  (ii) The literal rname will be written to `readtag' */
   /*  (iii) Write a separate `rdshow' with the control */
   /*     characters written as `^<char>' */
   if (havctl)
    { char *rsh=rdshow=dbuff=mxalloc(2*strlen(rname)+8);
      for (rn=rname; *rn; rn++)
       { if (*rn<0x20) { *rsh++='^';  *rsh++=*rn+0x40; }
         else *str++=*rsh++=*rn; }}
   else { strcpy(str,rname);  rdshow=rname; }
   /* Truncate rdshow for displays */
   if (strlen(rdshow)>MAXNAMELEN)  sprintf(rdshow+MAXNAMELEN, "...");
   /* If `rname' contains ONLY control characters, */
   /*   then strname[] will be empty.  Otherwise, */
   /*   quote (and possibly truncate) strname. */
   if (strname[0]==0)
    { char sb[44];  sprintf(sb,"ACC%d", ++allctrlnum);
      strname=strxdup(sb); }
   /* Complain if `rname' had strange characters */
   /* Strname has not (yet) been quoted */
   if (alldig || havctl)
     textprintf ("ILLEGAL TAG NAME `%s' changed to `%s'\n",
         rdshow, strname);
   strname=doquote(NULL,strname);  free(cbuff);
   if (rdtagptr)  *rdtagptr=strxdup(rname);
   if (rdshowptr) *rdshowptr=strxdup(rdshow);
   if (dbuff)  free(dbuff);
   return strname; }



/* Print numbers with commas:   */
/*   132,423  instead of  132423  */
/*   But:  1024 and 9836 instead of 1,024 and 9,836 */
/*   and   28000 and 49532  instead of  28,000 and 49,532 */
/* (The comma can be changed by entering, for example, */
/*     -numbersep='.'  at the command line or in a config file */
/*     so that  1323423  displays as   1.323.423 )  */

/* NOTE: commll() with a static buffer may not be safe to use */
/*   with multiple threads, since two threads may try to use the */
/*   same static buffer at the same time. */

extern char numbersep;

char *commll(char *nbuff, unsigned long uu)
 { if (nbuff==NULL)  nbuff=mxalloc(22);
   if (uu<10000)
     sprintf(nbuff,"%lu",uu);
   else
    { char ubuff[22];
      int i,j,m, n=sprintf(ubuff, "%lu", uu);
      m=n+(n-1)/3;  memset(nbuff,0,m+2);
      for (j=i=0; i<n; i++, j++)
       { if (i && (i%3)==0)  nbuff[m-1-j++]=numbersep;
         nbuff[m-1-j] = ubuff[n-1-i]; }}
   return nbuff; }

/* Utilities for displays avoiding awkwardnesses like */
/*   `0 files'  `1 animals'  `3 rallys'  and the like. */
/* Numsay() uses sprintf() to display the number */
/* Numsayc() calls commll() to add commas (or a replacement). */

static char *g_numsay (int sepflag,
    char *nbuff, long n, char *bleep, int capflag)
 { if (n==0 || n==1)
     strcpy (nbuff, (capflag) ? ( (n==1) ? "One" : "No")
         : ( (n==1) ? "one" : "no") );
   else if (sepflag)  commll(nbuff,n);
   else  sprintf (nbuff, "%ld", n);
   if (bleep)
    { sprintf(nbuff+strlen(nbuff), " %s", bleep);
      if (n != 1)
       { char *p=nbuff+strlen(nbuff);
         if (p[-1] == 'y' && strchr("aeiou",tolower(p[-2]))==NULL)
            strcpy (p-1, "ies");
         else  strcpy (p, "s"); }}
   return nbuff; }


char *numsay (char *nbuff, long n, char *bleep, int capflag)
 { return g_numsay(FALSE,nbuff,n,bleep,capflag); }

char *numsayc (char *nbuff, long n, char *bleep, int capflag)
 { return g_numsay(TRUE,nbuff,n,bleep,capflag); }


char *getstr(const char *control, ...)
 { char nbuff[SBUFFSIZE+100];
   va_list list;  va_start(list, control);
   vsprintf(nbuff, control, list);
   va_end(list);
   return strxdup(nbuff); }



/* File-name parsing utilities */

/* All pieces returned are either a malloc'ed string */
/*   or else "".  Thus, to free memory, you should */
/*   free all pieces that are not "". */

PARTFN partfn (const char *source)
 { int n;  char *p;  PARTFN pf;
   memset(&pf, 0, sizeof(pf));
   if (source==NULL || *source==0)
    { pf.drive="";  pf.path="";  pf.name="";  pf.ext="";
      return pf; }
   /* Is there a drive? */
   if (source[0] && source[1]==':')
    { sprintf(pf.drive=mxalloc(3), "%c:", source[0]);
      source += 2; }  else  pf.drive="";
   /* Is there a path? */
   for (n=strlen(source); --n>=0; )
     if (source[n]=='\\' || source[n]=='/')  break;
   if (++n>0)
    { memmove(pf.path = mxalloc(n+1), source, n);
      source += n; }  else  pf.path="";
   /* Is there a file name? */
   if (source[0] && source[0]!='.')
    { if ( (p=strrchr(source,'.'))!=NULL)
       { memmove(pf.name = mxalloc(p-source+1), source, p-source);
         source=p; }
      else { pf.name=strxdup(source);  source=""; }}
   else  pf.name="";
   /* Is there an extension? */
   pf.ext = (source[0]) ? strxdup(source) : "";
   return  pf; }


/* Makefnam ALWAYS returns a malloc'd string */
/*   (through makefmpf() )   */

char *makefnam (const char *rawfn, const char *tplate)
 { char *mname;  PARTFN cpf,tpf, mpf;
   mpf=cpf=partfn(rawfn);  tpf=partfn(tplate);
   if (mpf.drive[0]==0)  mpf.drive=tpf.drive;
   if (mpf.path[0]==0)  mpf.path=tpf.path;
   if (mpf.name[0]==0)  mpf.name=tpf.name;
   if (mpf.ext[0]==0)  mpf.ext=tpf.ext;
   mname=makefmpf(mpf);  freepf(tpf);  freepf(cpf);
   return mname; }


char *makefmpf (PARTFN pf)
 { char *name;
   sprintf (name=mxalloc(strlen(pf.drive)+strlen(pf.path)
                +strlen(pf.name)+strlen(pf.ext)+2),
      "%s%s%s%s", pf.drive, pf.path, pf.name, pf.ext);
   return name; }


int checkfname(const char *filename)
 { PARTFN pf=partfn(filename);  int retflag=RET_OK;
   if (pf.name[0]==0)  retflag=RET_ERROR;
   freepf(pf);  return retflag; }


void freepf (PARTFN pf)
 { if (pf.drive[0])  free(pf.drive);
   if (pf.path[0])  free(pf.path);
   if (pf.name[0])  free(pf.name);
   if (pf.ext[0])  free(pf.ext); }


/* Simple file-oriented utilities */

extern char *seqfile;
extern STRLIST filepath;

static char errwr[] =
  " Possible reasons:\n"
  "    (i) The directory is invalid or does not exist.\n"
  "   (ii) The file exists but is marked Read Only\n"
  "          or is in use by another process.\n"
  "  (iii) The storage device is full or unstable.\n";

static char errcl[] =
  " Possible reason:\n"
  "   (i) the storage device is either full or unstable\n";


static char *err_msg_write (const char *fname, int openflag)
 { char *ermess = (openflag) ? errwr : errcl;
   STRBUFF errb;  memset(&errb,0,sizeof(errb));
   addprintf(&errb, strlen(fname)+100,
     "\n CANNOT %s the %s file  `%s'\n",
     (openflag) ? "OPEN" : "CLOSE", say_type((char *)fname), fname);
   addprintf(&errb, strlen(ermess)+3, ermess);
   return errb.s; }


FILE *fxopenwrite(const char *fname, const char *mode)
 { FILE *fptr;
   if ( (fptr=fopen(fname,mode))==NULL)
     ErrHandler ("%s", err_msg_write(fname,TRUE));
   return fptr; }


int fxclose(FILE *fp, const char *fname)
 { int ret=-1;
   if (fp && (ret=fclose(fp))!=0)
     WarnHandler ("%s", err_msg_write(fname,FALSE));
   return ret; }


#ifndef UNIX
static void showpath_win (STRBUFFP ap);
#endif

static char *err_msg_path (const char *bleep,
    const char *fname, int fixed_path)
 { static STRBUFF errb;  errb.n=0;
   addprintf(&errb,strlen(bleep)+strlen(fname)+100,
      " CANNOT %s the %s file  `%s'\n",
         bleep, say_type((char *)fname), fname);
   if (fixed_path==FALSE)
   #ifndef UNIX
      showpath_win(&errb);
   #else
      showpath(&errb);
   #endif
   return errb.s; }


static char *err_msg_ext (const char *fname, const char *ext)
 { char *errb = mxalloc(strlen(fname)+200);
   sprintf(errb,
    " ILLEGAL EXTENSION (.%s) for SEQUENCE FILE `%s' !!\n"
    " (.%s is a %s output-file extension type.)\n",
        ext, fname, ext, ProgName);
   return errb; }


extern char *outfile_ext_list[];

FILE *fxopenread(const char *fname,
  char **openname, char **seqpath, int *fpp, int tabflag)
 { char *ext, **fpath=filepath.s;
   int npaths=filepath.n, fixed_path;
   char *mode=READMODE;
   FILE *rff;
   /* Check for an illegal output file extension type */

   if (tabflag && isextent(fname,"tab"))  ;
   else if ( (ext=strrchr(fname,'.'))!=NULL)
    { int i;  char **elist=outfile_ext_list;  ext++;
      for (i=0; elist[i]!=NULL; i++)  if (!stricmp(ext,elist[i]))
       { if (openname) *openname=NULL;
         if (seqpath) *seqpath=NULL;
         ErrHandler ("%s", err_msg_ext(fname,ext));
         return NULL; }}
   /* First check the default or indicated directory */
   fixed_path = (strchr(fname,PATHSEP)!=NULL);
   #ifndef UNIX
   /* Check for drive but no path */
   if (strchr(fname,':')!=NULL)
    { PARTFN pf=partfn(fname);  fixed_path=TRUE;
      if (pf.path[0]==0)
       { char pbuff[MAXDIR+8];
         if (!getcurdir(toupper(fname[0])-('A'-1),pbuff+1))
          { pbuff[0]='\\';  if (pbuff[1]!=0) strcat(pbuff,"\\");
            pf.path=strxdup(pbuff);
            fname=makefmpf(pf);  freepf(pf); }}}
   #endif
   if (fpp) *fpp=fixed_path;
   if ( (rff=fopen(fname,mode))!=NULL)
    { if (openname) *openname=strxdup(fname);
      if (fixed_path) *seqpath=strxdup(fname);
      else  *seqpath=NULL;
      return rff; }
   /* Not in default or indicated directory: */
   /* Check `filepath' unless `fname' contains an explicit path */
   if (fpath!=NULL && fixed_path==FALSE)
    { char *p, *nbuff;  int i,x,mxlen;
      for (mxlen=i=0; i<npaths; i++)
        if ( (p=fpath[i])!=NULL && (x=strlen(p))>mxlen)  mxlen=x;
      nbuff=mxalloc(mxlen+strlen(fname)+50);
      /* Parse_addpath() removed trailing PATHSEPs from fpath[i] */
      for (i=0; i<npaths; i++)
       { sprintf(nbuff, "%s%c%s", fpath[i], PATHSEP, fname);
         if ( (rff=fopen(nbuff,mode))!=NULL)  /* Found it !! */
          { if (openname) *openname=strxdup(nbuff);
            if (seqpath)  *seqpath=strxdup(fpath[i]);
            free(nbuff);  return rff; }}
      free(nbuff); }

   /* File not found, so return with error */
   if (openname) *openname=NULL;
   ErrHandler ("%s", err_msg_path("OPEN",fname,fixed_path));
   return NULL; }


int isextent(const char *p, const char *ext)
 { const char *q;
   return (p!=NULL && (q=strrchr(p,'.'))!=NULL && !stricmp(q+1,ext)); }


char *say_type (const char *fname)
 { char *ext;  int i;
   if (fname==NULL || fname[0]==0)  return "(Null)";
   else if (isextent(fname,"tab"))  return "tab (output)";
   else if (isextent(fname,"cfg"))  return "configuration";
   else if (isextent(fname,"sum"))  return "log";
   else for (i=0; (ext=outfile_ext_list[i])!=NULL; i++)
    { if (isextent(fname,ext))  return "output"; }
   return "sequence"; }


int type_msg (char *nbuff, const char *bleep, const char *infile)
 { char sbuff[66];
   if (infile && infile[0])
     return  sprintf (nbuff, "%s (%s file `%s'):\n",
       bleep, say_type(infile), shortname(sbuff,infile));
   else
     return  sprintf (nbuff, "%s (Sequence file=NULL)\n", bleep); }


char *shortname(char *sbuff, const char *fname)
 { if (sbuff==NULL || fname==NULL)  return NULL;
   if (strlen(fname)>50)
    { PARTFN pf=partfn(fname);
      if (strlen(pf.name)>42)  pf.name[42]=0;
      if (strlen(pf.ext)>8)  pf.ext[8]=0;
      sprintf(sbuff, "%s%s", pf.name, pf.ext);
      freepf(pf); }
   else  strcpy(sbuff,fname);
   return sbuff; }


int addprintf (STRBUFFP ap, int maxlen, const char *control, ...)
 { int len;  va_list list;  va_start(list, control);
   if (maxlen>0 && ap->nx - ap->n < maxlen)
     ap->s = rexxalloc(ap->s,1, ap->n, (ap->nx)+=maxlen);
   ap->n += (len=vsprintf(ap->s + ap->n, control, list));
   va_end(list);
   return len; }


static int show_addp (STRBUFFP ap, int col, const char *str)
 { int n;
   if (strpbrk(str," \t")!=NULL)   /* Embedded spaces? */
    { if ( (col+=(n=strlen(str)+4)) > 74)
       { addprintf(ap,2, "\n ");  col=n+4; }
      addprintf(ap, n+4, "  \"%s\"", str); }
   else
    { if ( (col+=(n=strlen(str)+2)) > 74)
       { addprintf(ap,2, "\n ");  col=n+2; }
      addprintf(ap, n+4, "  %s", str); }
   return col; }


void showpath (STRBUFFP ap)
 { if (filepath.s==NULL || filepath.n==0)
     addprintf(ap,30,
       " (No input file path other than default directory.)\n");
   else
    { char *p;  int i, col=addprintf(ap,80,
       " Input file path (besides default directory):");
      for (i=0; i<filepath.n; i++)
        if ( (p=filepath.s[i])!=NULL && p[0]!=0)
          col=show_addp(ap,col,p);
      addprintf (ap,5, " \n"); }}


#ifndef UNIX

static void showpath_win (STRBUFFP ap)
 { int i,col, havp, npaths=filepath.n;
   char *p, **fpath=filepath.s, cbuff[MAXPATH+33];
   havp = (fpath!=NULL && npaths!=0);
   col=addprintf (ap,30, " Looked in director%s:", (havp) ? "ies" : "y");
   if (getcwd(cbuff,MAXPATH+20)==NULL)
     strcpy(cbuff,"Default directory");
   col=show_addp(ap,col,cbuff);
   if (havp)
    { for (i=0; i<npaths; i++)
        if ( (p=fpath[i])!=NULL && p[0]!=0)
          col=show_addp(ap,col,p); }
   addprintf (ap,5, " \n"); }

#endif


/* Argument formatting and specialized print functions */

void nl(void)
 { putchar ('\n'); }


extern int seqf_open, havelog;   /* In vread.c, vsetf.c */
extern int commline_shown;

void warnprintf(const char *control, ...)
 { char nbuff[SBUFFSIZE+150];  int len=0;
   va_list list;  va_start(list, control);
   if (seqf_open)  len=type_msg(nbuff," WARNING",seqfile);
   vsprintf(nbuff+len, control, list);
   va_end(list);
   BPRINTF("\n%s\n", nbuff); }



extern int hold_window;

/* Keep the console window open on a fatal error */
/*   or if asked to keep open on a normal return. */
void presskey(void)
 { printf ("\nPress `Enter' to %s...",
      (hold_window) ? "continue" : "close window");
   fflush(stdin);
   getchar();
   printf ("\n"); }


/* Better behaved than getchar(), but not ANSI:
   printf ("\nPress any key to %s...",
      (hold_window) ? "continue" : "close window");
    while(kbhit())  getch();   getch();

*/


#ifdef VTEXT
void closeout(const char *bleep);   /* In vtds.c */
#endif
void show_cargcs(const char *bleep, const char *spacestr, int fgap);


static void genx_out(const char *nbuff)
 { if (commline_shown==FALSE)
    { BNL();  reset_textprintf();
      show_cargcs(NULL," ",3); }
   #ifdef VTEXT
   closeout(nbuff);
   #else
   BPRINTF("\n%s\n", nbuff);
   #endif
   exit(1); }


void exitprintf(const char *control, ...)
 { char nbuff[SBUFFSIZE+150];  int len=0;
   va_list list;  va_start(list, control);
   if (seqf_open)  len=type_msg(nbuff," ERROR",seqfile);
   vsprintf(nbuff+len, control, list);
   va_end(list);
   genx_out(nbuff);
   exit(1); }


void xformout (const char *control, ...)
 { char nbuff[SBUFFSIZE+200];  int len=0;
   va_list list;  va_start(list, control);
   if (seqf_open)  len=type_msg(nbuff," ERROR",seqfile);
   len+=vsprintf(nbuff+len, control, list);
   va_end(list);
   sprintf (nbuff+len,
     "\nSee the documentation for the correct syntax.");
   genx_out(nbuff);
   exit(1); }



void set_textprintf(PRINTFUNP prfun, void (*prnl)(void))
 { textprintf=prfun;  textnl=prnl; }

void reset_textprintf(void)
 { textprintf=BPRINTF;  textnl=BNL; }


/* Word-wrap (token-printing) package */

/* The principal functions are: */
/*                                                                  */
/*  void ftoklistf (const char *bleep, char **vars,int num,         */
/*     int nrgap,const char *sep,int fgap, int addflags);           */
/*                                                                  */
/*  void toklist (const char *bleep, char **vars,int num);          */
/*                                                                  */
/*  void ftoklistex (const char *bleep, TOKLISTP tokl,              */
/*     int nrgap,const char *sep,int fgap, int addflags);           */
/*                                                                  */
/* These functions write an initial string (``bleep'') followed by  */
/*   a list of strings or tokens (``vars,num''), with line breaks   */
/*   before each token that would otherwise end beyond a line       */
/*   width `tok_linesize'.  */
/*                                                            */
/* The variables ``nrgap,sep,fgap'' define two formatting values */
/*   (see below) and a separator string (``sep'', perhaps ","). */
/*   The final variable `addflags' is a list of ORed flags that */
/*   provide further control.  */
/*                                                               */
/* All three functions write to `textprintf()'.  `Toklist()' is  */
/*   `ftoklistf()' with the settings  nrgap=2, sep=NULL, fgap=0  */
/*   (which means fgap=2), and `addflags'=NONE.  `Ftoklistex()'  */
/*   is essentially `ftoklistf()' except that the tokens are stored */
/*   in a list of TOKSTRUC structures that proved finer control */
/*   of whether tokens are to be quoted or their strings deallocated */
/*   after use.   */
/*                                                              */
/* All three functions call the basic static ``workhorse'' function */
/*   `s_ftokfun()', which is the same as `ftoklistf()' except that */
/*   there is no support for quoting strings.  `Ftoklistex()   */
/*   uses `addtokl()' to build an array of TOKSTRUC structures and */
/*   then calls `s_ftokfun()'.  */

/* `Ftoklistex()' has two helper functions:     */
/*                                                              */
/*  char *addtokl (TOKLISTP tokl, char *str, int addflags);     */
/*  int tokl_string_sort (TOKLIST dd);                          */
/*                                                              */
/*  for either adding to a list of TOKSTRUC structures or for   */
/*  sorting them.                                               */

/* IN MORE DETAIL:  Arguments are: */

/* bleep:         List heading */
/* vars,num:      List of strings and number of items */
/* nrgap:  Add #nrgap spaces BEFORE items #2--#num, EXCEPT before */
/*           the first item or the first item on a line. */
/* sep:    Add AFTER items #1--#(num-1).  (Perhaps a comma.)  */
/* fgap:   Initial gap in succeeding lines (Default if fgap==0: 2) */

/* Bitwise flags for `addflags':  */

/* TOK_NONE:   No flags  (TOK_NONE=0)         */
/* TOK_FREE:   If applied to a token, then that token should be freed if */
/*     `ftoklistex' is called with TOK_FREE */
/*   If applied to `ftoklistex', then all such tokens should be freed */
/*     AND all strings stored in `frstr' pointers should be freed. */
/*     The latter applies to internally-created strings for tokens */
/*     marked TOK_QUOTE or TOK_VOLATILE */
/* TOK_FREE_ALL:  Same as TOK_FREE     */
/* TOK_QUOTE means that `str' should be quoted if it contains spaces */
/*   and/or truncated if it is too long */
/* TOK_VOLATILE means that `str' is in a volatile buffer that may later */
/*   be overwritten, so that `addtokl' should make a copy */
/* TOK_ENDPERIOD means that the list should end with a period */
/* TOK_GROUPS means to space before and after -Group commands */


/* The line width `tok_linesize' is in vsetopts.c */
/* This imposes a minimum line size: */

#define MIN_LINESIZE 50

static char *tok_lbuff;
static int txlinelen;
static int vtok_linesize;

static int tok_groupnames_match(const char *str);

/* Basic ``workhorse'' function: */

static void s_ftokfun (const char *bleep, char **vars,int num,
   int nrgap,const char *sep,int fgap, int addflags)
 { int len, nbgap, nsep, gflag;
   int groupflag = (addflags & TOK_GROUPS);
   /* Assign default values */
   if (bleep==NULL)  bleep="";  if (sep==NULL) sep="";
   if (fgap<=0) fgap=2;   nsep=strlen(sep);
   vtok_linesize=TOK_LINESIZE;
   if (dumpflag)
    { /* `Tok_linesize=off' restores default value. */
      if (tok_linesize<=0)  tok_linesize=TOK_LINESIZE;
      else if (tok_linesize<MIN_LINESIZE)  tok_linesize=MIN_LINESIZE;
      vtok_linesize=tok_linesize; }
   /* Make sure that the line buffer is large enough */
   /* (This also allocates the line buffer the first time through.) */
   if (vtok_linesize+strlen(bleep)+50 > txlinelen)
    { txlinelen=vtok_linesize+strlen(bleep)+100;
      tok_lbuff=rexxalloc(tok_lbuff,1, 0,txlinelen); }
   /* Start with the initial string: */
   /* NOTE: ``len=sprintf(tok_lbuff,bleep);'' may cause */
   /*   unpredicable output if `bleep' contains any `%'s !! */
   len=sprintf(tok_lbuff,"%s",bleep);
   /* The between-item gap `nbgap' is NOT included immediately after */
   /*   the heading AND NOT at the beginning of lines. */
   /*   (`Fgap' applies at the beginning of lines.)    */
   /* If TOK_GROUPS==TRUE, gflag==TRUE within a name list for a group */
   gflag=FALSE;  nbgap=0;
   /* Loop over the items: */
   while (--num>=0)
    { const char *str=(*vars!=NULL)?*vars:"(Null)";
      int itemlen;
      if (groupflag)
       { int gf=FALSE;    /* Start a new line before this item? */
         if (!tok_groupnames_match(str))  gf=gflag=TRUE;
         else if (gflag && (str[0]=='-' || str[0]=='/')
            && str[1]!=0 && strcmp(str,"//")!=0)
          { gf=TRUE;  gflag=FALSE; }
         if (gf)
          { textprintf("%s\n",tok_lbuff);
            len=sprintf(tok_lbuff,"%*s",fgap,"");  nbgap=0; }}
      itemlen=nbgap+strlen(str);
      /* There is no sep AFTER the last item !! */
      if (num>0)  itemlen+=nsep;  else sep="";
      /* Does the current item fit on the line? */
      /* If not, there is no between-item gap at the BEGINNING */
      /*   of the next line, although there is the initial (fgap) */
      /*   line indentation. */
      if (len+itemlen > vtok_linesize)
        { textprintf("%s\n",tok_lbuff);
          len=sprintf(tok_lbuff,"%*s",fgap,"");  nbgap=0; }
      /* Finally, add the item, with a possible between-item gap */
      /*   BEFORE the item and sep AFTER the item. */
      len+=sprintf(tok_lbuff+len, "%*s%s%s", nbgap,"", str, sep);
      vars++;  nbgap=nrgap; }
   /* Perhaps add a final period */
   if (addflags & TOK_ENDPERIOD)
    { tok_lbuff[len++]='.';  tok_lbuff[len]=0; }
   /* Flush the last line and add a final EOL. */
   textprintf("%s\n",tok_lbuff); }



void ftoklistf (const char *bleep, char **vars,int num,
   int nrgap,const char *sep,int fgap, int addflags)
 { if (addflags & TOK_QUOTE)
    { int i;  TOKLIST tokl;  memset(&tokl,0,sizeof(tokl));
      for (i=0; i<num; i++)  addtokl(&tokl, vars[i], addflags);
      ftoklistex(bleep, &tokl, nrgap,sep,fgap, addflags | TOK_FREE); }
   else
    { s_ftokfun(bleep, vars,num, nrgap,sep,fgap, addflags);
      if (addflags & TOK_FREE_ALL)
       { while (--num>=0)  free(vars[num]); }}}


void toklist (const char *bleep, char **vars,int num)
 { s_ftokfun(bleep, vars,num, 2,NULL,0, TOK_NONE); }


void ftoklistex (const char *bleep, TOKLISTP tokl,
   int nrgap,const char *sep,int fgap, int addflags)
 {  char **flist;  int i, n=tokl->n;  TOKSTRUC *tok;
    if (tokl==NULL || tokl->tok==NULL || n==0)
     { textprintf ("%s\n", bleep);  return; }
    flist=cxalloc(n+3,sizeof(char *));  tok=tokl->tok;
    for (i=0; i<n; i++)  flist[i]=tok[i].str;
    s_ftokfun(bleep, flist,n, nrgap,sep,fgap, addflags);
    /* Now clean up */
    free(flist);
    if (addflags & TOK_FREE)
     { while (--i>=0)
        { if (tok[i].fflag)  free(tok[i].str);
          if (tok[i].frstr!=NULL)  free(tok[i].frstr); }
       free(tokl->tok);  memset(tokl,0,sizeof(TOKLIST)); }}


char *addtokl (TOKLISTP tokl, char *str, int addflags)
 { char *p;  int doq=FALSE;
   TOKSTRUC tt;  memset(&tt,0,sizeof(tt));
   if (str==NULL)  return NULL;
   if (addflags & TOK_QUOTE)
    { if (strlen(str)>MAXNAMELEN) doq=TRUE;
      else { for (p=str; *p!=0; p++)
              if (isspace(*p) || *p==';') { doq=TRUE;  break; }}}
   /* TOK_VOLATILE is not necessary if you have to quote it anyway */
   if (doq)
    { tt.str=doquote(NULL,str);  tt.fflag=TRUE;
      if (addflags & TOK_FREE)  tt.frstr=str; }
   else if (addflags & TOK_VOLATILE)
    { tt.str=strxdup(str);  tt.fflag=TRUE;
      if (addflags & TOK_FREE)  tt.frstr=str; }
   else { tt.str=str;  tt.fflag=(addflags & TOK_FREE); }
   if (tokl->n >= tokl->nx)
     tokl->tok = rexxalloc(tokl->tok,sizeof(TOKSTRUC),
          tokl->n, tokl->nx +=50);
   tokl->tok[tokl->n++] = tt;
   return tt.str; }




static int tok_string_sort (const void *p, const void *q)
 { int x;  const TOKSTRUC *p1=p, *q1=q;
   if ( (x = striucmp(p1->str,q1->str))!=0)  return x;
   if ( (x = strcmp(p1->str,q1->str))!=0)  return x;
   if ( (x = p1->fflag - q1->fflag)!=0)  return x;
   return 0; }


int tokl_string_sort (TOKLIST dd)
 { if (dd.tok!=NULL)
     qsort(dd.tok,dd.n, sizeof(dd.tok[0]), tok_string_sort);
   return dd.n; }


static char *tok_grlist[] = {
   "group",          "-group",
   "seq_list",       "-seq_list",
   "add_seq_list",   "-add_seq_list",
   "skip_list",      "-skip_list",
   "add_skip_list",  "-add_skip_list",
   "subset",         "-subset",
   "add_subset",     "-add_subset",
   "seq_skip",       "-seq_skip",
   "add_seq_skip",   "-add_seq_skip",
   "test",           "-test",
   "list_only",      "-list_only",
   "add_list_only",  "-add_list_only",
   "listonly",       "-listonly",
   "add_listonly",   "-add_listonly",
   NULL, NULL };


static int tok_groupnames_match(const char *str)
 { const char *p;  int i=0;
   while ( (p=tok_grlist[i++])!=NULL)
     if (!strmatch(str,p))  return FALSE;
   return TRUE; }



/* Memory allocation functions with error checks */

int numrealloc;
extern int maxsqlinelen;
unsigned long totalloc, totrealloc;

void cant_alloc (unsigned long uu)
 { char cbuff[22];
   ErrHandler (
  "\nCANNOT ALLOCATE (or REALLOCATE) %lu bytes from the system !!\n"
  "Total memory allocated or reallocated so far:  %s bytes\n",
     uu, commll(cbuff,totalloc+totrealloc)); }


void show_alloc (VOID_PRINTFUNP prfun)
 { char nbuff[44], cbuff[22];
   prfun("Total memory allocated:  %s bytes.\n", commll(cbuff,totalloc));
   prfun ("%s", numsayc(nbuff, numrealloc, "reallocation", TRUE));
   if (numrealloc)
      prfun (" of total new size: %s bytes", commll(cbuff,totrealloc));
   prfun (".\n");
   if (maxsqlinelen>0)
     prfun ("Maximum input sequence-file line width: %s.\n",
       numsayc(nbuff,maxsqlinelen, "byte",TRUE)); }

void show_alloc_sofar (VOID_PRINTFUNP prfun)
 { char cbuff[22];
   prfun("Memory allocated so far:  %s bytes.\n",
      commll(cbuff,totalloc)); }

void *mxalloc (size_t size)
 { return cxalloc(1,size); }

void *cxalloc (size_t nitems, size_t size)
 { void *vptr;
   if ( (vptr = calloc(nitems,size)) == NULL)
     cant_alloc ((unsigned long)nitems * size);
   totalloc += (unsigned long)nitems * size;
   return  vptr; }

char *strxdup(const char *p)
 { return (p) ? strcpy (mxalloc(strlen(p)+1), p)
            : mxalloc(1); }


char *strxndup(const char *p, size_t n)
 { return (p) ? strcpy (mxalloc(strlen(p)+1+n), p)
            : mxalloc(1+n); }


void *rexalloc(void *p, size_t newsize)
 { void *voidptr;
   if ( (voidptr = realloc(p,newsize)) == NULL)  cant_alloc(newsize);
   return voidptr; }




void *rexxalloc(void *p, size_t ssize, size_t oldnum, size_t newnum)
 { void *voidptr;  size_t newsize=ssize*(newnum+3);
   if ( (voidptr = realloc(p,newsize)) == NULL)  cant_alloc(newsize);
   if (p!=NULL) { totrealloc += newsize;  numrealloc++; }
   else  totalloc += newsize;
   if (oldnum<newnum)
     memset((char *)voidptr + ssize*oldnum, 0, ssize*(newnum+3-oldnum));
   else
     memset((char *)voidptr + ssize*newnum, 0, ssize*3);
   return voidptr; }


/* Batch size for addlist() is 20 */
/* Batch size for addstr() is 50 */

void *addlist(void *ss, size_t size, void **slist,int *nsp,int *nxsp)
 { char *targ;
   if (*nsp >= *nxsp)
     *slist=rexxalloc(*slist,size, *nsp, (*nxsp)+=20);
   memmove(targ = (char *)*slist + (*nsp)*size, ss, size);
   (*nsp)++;  return targ; }


char *addstr(STRLISTP ap, const char *str)
 { if (ap->n >= ap->nx)
     ap->s=rexxalloc(ap->s,sizeof(char *), ap->n, (ap->nx)+=50);
   return (ap->s)[(ap->n)++]=strxdup(str); }


/* Same as addstr(), but ignore duplicates. */

char *check_addstr(STRLISTP ap, char *str)
 { char **ss=ap->s;  int i, n=ap->n;
   for (i=0; i<n; i++) if (!strcmp(str,ss[i]))  return ss[i];
   return  addstr(ap,str); }


/* Same as addstr(), but ignore duplicates */
/*   AND (if appropriate) convert spaces to underscores */

char *check_addseq(STRLISTP ap, char *str)
 { char **ss=ap->s;  int i, n=ap->n;
   for (i=0; i<n; i++) if (!striucmp(str,ss[i]))  return ss[i];
   return  addstr(ap,str); }


char *unspace_str(char *str)
 { if (str!=NULL)
    { char *p;  for (p=str; *p; p++)  if (isspace(*p))  *p='_'; }
   return str; }

void unspace_slist1(char **str, int n)
 { if (str!=NULL)
    { int i;
      for (i=0; i<n; i++)  unspace_str(str[i]); }}

void unspace_slist2(STRLIST a)
 { unspace_slist1(a.s,a.n); }


void kill_strlist (STRLISTP ap)
 { if (ap!=NULL && ap->s!=NULL)
    { int n=ap->n;
      while(--n>=0) free ((ap->s)[n]);  free(ap->s);
      memset(ap,0,sizeof(STRLIST)); }}


void kill_nstr (char ***listp, int n)
 { if (listp!=NULL && *listp!=NULL)
    { char **str=*listp;
      while (--n>=0) if (str[n]) free(str[n]);
      free(str);  *listp=NULL; }}


void kill_strbuff (STRBUFFP ap)
 { if (ap!=NULL)
    { if (ap->s!=NULL)  free(ap->s);
      memset(ap,0,sizeof(STRBUFF)); }}


int parse_addpath(STRLISTP ap, char *str)
 { char *rs, *nextp, *strin;
   static char sep[3] = { DIRLISTSEP };
   if (ap==NULL || str==NULL)  return RET_ERROR;
   rs=strin=strxdup(str);
   do { char *q, *path=NULL;
        rs=parse_unquote_sep(rs,&nextp,sep,NULL);
        if (*rs==0)  break;
 #ifndef UNIX
   /* If `rs' begins with %<word>%, expand <word> as an */
   /*   environmental variable. Remove if not found.  */
        if (*rs=='%' && (q=strchr(rs+1,'%'))!=NULL && q > rs+1)
          { char *word,*path,*evar;
            int tlen=q-(rs+1);  evar=mxalloc(tlen+2);
            memmove(evar,rs+1,tlen);
            if ( (word=getenv(evar))==NULL)  word="";
            sprintf(path=mxalloc(strlen(word)+strlen(q+1)+3),
               "%s%s", word,q+1);  rs=path;  free(evar); }
 #endif /* UNIX */
        /* Kill any trailing PATHSEP's  (Note can have rs=="") */
        q=rs+strlen(rs);
        while (q>rs && *--q==PATHSEP)  *q=0;
        check_addstr(ap,rs);  if (path!=NULL)  free(path); }
   while ( (rs=nextp)!=NULL);
   free(strin);
   return RET_OK; }




#ifndef UNIX

/* Show time elapsed */
/* (Prototype is at the top of vutil.h along with */
/*   other DOS/UNIX dependancies */

void timer (TIME_VAR tvar, int num, char *bleep)
 { static clock_t  ocloc;
   if (notimer) return;
   switch(tvar) {
     case TIME_INIT:  ocloc=clock();  return;
     case TIME_ELAPSED:
       { char abuff[88];
         double zsec = (double)(clock()-ocloc)/CLK_TCK;
         if (num>2)
           sprintf(abuff, "    Per iteration: %f secs.", zsec/num);
         else abuff[0]=0;
         BPRINTF ("\nElapsed time%s: %.2f secs.%s\n",
            (bleep) ? bleep : "", zsec, abuff); }  break; }}

#endif /* timer() */



/* Package for overlap control */

/* Machinery for restricting within-pair (or within-sequence) */
/*   overlaps to at most  maxoverlap  locally  */

/* Algorithm: Assuming that fragments are sorted by score, */
/*   (a) Store the first fragment in an overlap array */
/*   (b) Return TRUE (OK) for that fragment and the next */
/*     `maxovlap' fragments that intersect it */
/*   (c) Return FALSE (Not OK) for fragments beyond the first */
/*     `maxovlap' that intersect it */
/*   (d) If a fragment does not overlap the first fragment, */
/*     then repeat step (a). */
/* In particular, lower-scoring fragments in the same overlapping */
/*   group do not extend the `dead zone' for that group. */
/*   (Otherwise, the completed list of fragments might show */
/*     mysterious gaps in its coverage.)  */

/* A return of                  */
/*   TRUE means OK to use      */
/*   FALSE means an excess overlap; don't use     */

/* Initializing and reporting steps: */

/* checkoverlap(2,npolys,0) initializes the overlap arrays */
/*   This should be run before any other use */

/* checkoverlap(1,0,0) returns the number of overlaps */
/*   This is for output file comments only. */

/* checkoverlap(0,a,b) checks the fragment (a,b) and */
/*                                                */
/*   returns TRUE if the fragment does not overlap excessively */
/*      and should be added to fragment lists
/*   returns FALSE if it is an excess overlap */
/*      and should be ignored.   */

/* The overlap array structure has a batch size */
/*   (for extending the array) of size  50 */

typedef struct { int start,fin,nov; } VLAP;

int showoverlap;

int checkoverlap(int op, int start, int fin)
 { static VLAP *vlap;
   static int vn,vndropped,vx, npolys;  int i;
   if (op>0)  /* Initializing or reporting step */
    { if (op==2)   /* Initialize */
       { /* Clear overlap arrays and counts */
         if (vlap!=NULL) memset(vlap,0,vx*sizeof(vlap[0]));
         vn=vndropped=0;  npolys=start;  return FALSE; }
      /* Else report how many overlapping fragments were dropped. */
      else  return vndropped; }
   if (showoverlap)  textprintf (" (%d,%d):",start,fin);
   /* Check for overlaps with remembered fragments */
   for (i=0; i<vn; i++)
    { int a=fin-start, b=vlap[i].start-start, c=vlap[i].fin-start;
      /* Compare [0,a) with [b,c) by shifting b to the range */
      /*   0<=b<npolys . */
      while (b<0) { b+=npolys;  c+=npolys; }
      while (b>=npolys) { b-=npolys; c-=npolys; }
      if (showoverlap)  textprintf (" (i=%d,0,%d,%d,%d(%d))",i,a,b,c,npolys);
      /* Break means an overlap, no break means no overlap */
      if (b<a || c>npolys)  break; }
   if (i>=vn)      /* No overlap with any stored fragment: */
    { if (showoverlap) textprintf ("(NEW,i=%d)\n",i);
      /* Add a new fragment range */
      if (vn>=vx)  vlap=rexxalloc(vlap, sizeof(VLAP), vn, vx+=50);
      vlap[vn].start=start;  vlap[vn].fin=fin;  vlap[vn].nov=1;  vn++; }
   else         /* Overlaps with vlap[i] */
    { if (showoverlap)  textprintf ("(i=%d,nov=%d,%s)\n",
             i,vlap[i].nov+1,(vlap[i].nov>=maxoverlap) ? "XX":"PASS");
      if (++vlap[i].nov>maxoverlap)  { vndropped++;  return FALSE; }}
   return TRUE; }  /* OK: Don't censor */



/* Random number package */

/* The `& 0xffff....' in LRAND is so that  nrand()  will work  */
/*   properly if sizeof(long) > 4  (e.g. 64-bit integers)  */

#define MULT 0x19660DL
#define LRAND (rseed = ( (MULT*rseed + 1L) & 0xffffffff) )

static unsigned long  rseed = 0x75BE6ACCL;

int nrand(int m)
 { return  (m > 0)  ?  (int) ((m * (LRAND >> 16)) >> 16) : 0; }


unsigned long setrand(unsigned long initseed)
 { int i;
   if (initseed!=0L)  return  rseed=initseed;
   rseed = time(NULL);
   for (i=0; i<20; i++)  LRAND;
   return  LRAND; }

unsigned long dispseed(void)
 { return  rseed; }




